!!****if* source/Simulation/SimulationMain/magnetoHD/AnisotropicViscosityTest/Simulation_initBlock
!!
!! NAME
!!
!!  Simulation_initBlock
!!
!!
!! SYNOPSIS
!!
!!  Simulation_initBlock(integer(in) :: blockID, 
!!                       integer(in) :: myPE)
!!
!!
!!
!! DESCRIPTION
!!
!!  Initializes fluid data (density, pressure, velocity, etc.) for
!!  a specified block.
!!
!!  Parameters:  blockID      The number of the block to initialize
!!
!! 
!! ARGUMENTS
!!
!!  blockID -          the number of the block to update
!!  myPE   -           my processor number
!!
!! 
!!
!!***

subroutine Simulation_initBlock(blockID, myPE)

  use Simulation_data
  use Grid_interface, ONLY : Grid_getBlkIndexLimits, &
                             Grid_getCellCoords,     &
                             Grid_getDeltas, &
                             Grid_getBlkPtr, &
                             Grid_releaseBlkPtr

  implicit none

#include "constants.h"
#include "Flash.h"

  !! Arguments ------------------------
  integer, intent(in) :: blockID, myPE
  !! ----------------------------------

  integer :: i, j, k, n, istat, sizeX, sizeY, sizeZ
  integer, dimension(2,MDIM) :: blkLimits, blkLimitsGC
  real :: enerZone, ekinZone, eintZone, radius, dx, dy, dz
  real, allocatable,dimension(:) :: xCoord,yCoord,zCoord
  real, dimension(MDIM) :: del
  real, pointer, dimension(:,:,:,:) :: solnData, facexData,faceyData,facezData
  real :: xx,yy,zz, Vphi

  ! dump some output to stdout listing the paramters
  if (myPE == MASTER_PE) then
1    format (1X, 1P, 4(A7, E13.7, :, 1X))
2    format (1X, 1P, 2(A7, E13.7, 1X), A7, I13)
  endif
  
  call Grid_getBlkIndexLimits(blockId,blkLimits,blkLimitsGC)

  sizeX = blkLimitsGC(HIGH,IAXIS)-blkLimitsGC(LOW,IAXIS)+1
  sizeY = blkLimitsGC(HIGH,JAXIS)-blkLimitsGC(LOW,JAXIS)+1
  sizeZ = blkLimitsGC(HIGH,KAXIS)-blkLimitsGC(LOW,KAXIS)+1

  allocate(xCoord(sizeX), stat=istat)
  allocate(yCoord(sizeY), stat=istat)
  allocate(zCoord(sizeZ), stat=istat)

  xCoord  = 0.0
  yCoord  = 0.0
  zCoord  = 0.0

  if (NDIM == 3) then
     call Grid_getCellCoords(KAXIS,blockId,CENTER,sim_gCell,zCoord,sizeZ)
  endif
  if (NDIM >= 2) then
     call Grid_getCellCoords(JAXIS,blockId,CENTER,sim_gCell,yCoord,sizeY)
  endif
  call Grid_getCellCoords(IAXIS,blockId,CENTER,sim_gCell,xCoord,sizeX)

  call Grid_getDeltas(blockID,del)
  dx = del(1)
  dy = del(2)
  dz = del(3)

  call Grid_getBlkPtr(blockID,solnData,CENTER)

#if NFACE_VARS > 0
  if (sim_killdivb) then
     call Grid_getBlkPtr(blockID,facexData,FACEX)
     call Grid_getBlkPtr(blockID,faceyData,FACEY)
     if (NDIM == 3) call Grid_getBlkPtr(blockID,facezData,FACEZ)
  endif
#endif


  !------------------------------------------------------------------------------

  do k = blkLimitsGC(LOW,KAXIS),blkLimitsGC(HIGH,KAXIS)
     do j = blkLimitsGC(LOW,JAXIS),blkLimitsGC(HIGH,JAXIS)
        do i = blkLimitsGC(LOW,IAXIS),blkLimitsGC(HIGH,IAXIS)

           ! define radius of the field loop
           xx = xCoord(i)
           yy = yCoord(j)
           radius = sqrt((xx-sim_xCtr)**2 + (yy-sim_yCtr)**2)

           Vphi = sim_Vphi0*exp(-((radius-sim_Vradius)/sim_Vwidth)**2)
           solnData(VELX_VAR,i,j,k)= -yy/radius*Vphi
           solnData(VELY_VAR,i,j,k)=  xx/radius*Vphi
           solnData(VELZ_VAR,i,j,k)=  sim_Vz0
           
           solnData(DENS_VAR,i,j,k)=  1.
           solnData(PRES_VAR,i,j,k)=  1.
           solnData(TEMP_VAR,i,j,k)=  1.

           ! Compute the gas energy and set the gamma-values needed for the EOS
           ekinZone = 0.5 * dot_product(solnData(VELX_VAR:VELZ_VAR,i,j,k),&
                                        solnData(VELX_VAR:VELZ_VAR,i,j,k))

           ! specific internal energy
           eintZone = solnData(PRES_VAR,i,j,k)/(sim_gamma-1.)/solnData(DENS_VAR,i,j,k)

           ! total specific gas energy
           enerZone = eintZone + ekinZone

           ! Take a limit value
           enerZone = max(enerZone, sim_smallP)

           solnData(ENER_VAR,i,j,k)=enerZone
           solnData(EINT_VAR,i,j,k)=eintZone
           solnData(GAMC_VAR,i,j,k)=sim_gamma
           solnData(GAME_VAR,i,j,k)=sim_gamma

           !Initialize the magnetic field
           solnData(MAGX_VAR,i,j,k) = sim_Bx0
           solnData(MAGY_VAR,i,j,k) = sim_By0
           solnData(MAGZ_VAR,i,j,k) = sim_Bz0
           solnData(DIVB_VAR,i,j,k) = 0.
#if NFACE_VARS > 0
           if (sim_killdivb) then
                 facexData(MAG_FACE_VAR,i,j,k) = sim_Bx0
                 faceyData(MAG_FACE_VAR,i,j,k) = sim_By0
              if (NDIM == 3) then
                 facezData(MAG_FACE_VAR,i,j,k) = sim_Bz0
              endif
           endif
#endif
           ! Update the magnetic pressure
           solnData(MAGP_VAR,i,j,k) = .5*dot_product(solnData(MAGX_VAR:MAGZ_VAR,i,j,k),&
                                                     solnData(MAGX_VAR:MAGZ_VAR,i,j,k))
        enddo
     enddo
  enddo

#ifdef CURJ_VAR
  k=1
  do j = blkLimitsGC(LOW,JAXIS),blkLimitsGC(HIGH,JAXIS)
     do i = blkLimitsGC(LOW,IAXIS),blkLimitsGC(HIGH,IAXIS)
        solnData(CURJ_VAR,i,j,k) =&
               (faceyData(MAG_FACE_VAR,i+1,j,k)-faceyData(MAG_FACE_VAR,i,j,k))/dx &
              -(facexData(MAG_FACE_VAR,i,j+1,k)-facexData(MAG_FACE_VAR,i,j,k))/dy
     enddo
  enddo
#endif

  !------------------------------------------------------------------------------

  ! Release pointer
  call Grid_releaseBlkPtr(blockID,solnData,CENTER)

#if NFACE_VARS > 0
  if (sim_killdivb) then
     call Grid_releaseBlkPtr(blockID,facexData,FACEX)
     call Grid_releaseBlkPtr(blockID,faceyData,FACEY)
     if (NDIM == 3) call Grid_releaseBlkPtr(blockID,facezData,FACEZ)
  endif
#endif

  deallocate(xCoord)
  deallocate(yCoord)
  deallocate(zCoord)

end subroutine Simulation_initBlock



