!!****if* source/physics/Hydro/HydroMain/unsplit/Anisotropic_MHD/hy_uhd_Roe
!!
!! NAME
!!
!!  hy_uhd_Roe
!!
!! SYNOPSIS
!!
!!  hy_uhd_Roe( integer(IN) :: dir,
!!              real(IN)    :: Vm(HY_VARINUM4),
!!              real(IN)    :: Vp(HY_VARINUM4),
!!              real(OUT)   :: Fstar(HY_VARINUM),
!!              real(OUT)   :: vint  )
!!
!! ARGUMENTS
!!
!!  dir    - a spatial direction for which the flux is being considered and computed
!!  Vm     - primitive variable for left state  
!!            (DENS,VELX,VELY,VELZ,PRES,MAGX,MAGY,MAGZ + GAMC,GAME,EINT,TEMP)
!!  Vp     - primitive variable for right state 
!!            (DENS,VELX,VELY,VELZ,PRES,MAGX,MAGY,MAGZ + GAMC,GAME,EINT,TEMP)
!!  Fstar  - computed flux data
!!  vint   - interface velocity
!!
!! DESCRIPTION
!! 
!!   This routine computes high-order Godunov fluxes based on the left and right Riemann states.
!!
!! REFERENCES
!!
!!  * Roe, JCP, 43:357-372, 1981
!!  * Hanawa, Mikami, and Matsumoto, JCP, 227:7952-7976, 2008
!!
!!***

Subroutine hy_uhd_Roe(dir,Vm,Vp,Fstar,vint,ix,iy,iz,blkLimitsGC,LambdaFix)

  use hy_uhd_interface, ONLY : hy_uhd_avgState,       &
                               hy_uhd_eigenParameters,&
                               hy_uhd_eigenValue,     &
                               hy_uhd_eigenVector,    &
                               hy_uhd_prim2con,       &
                               hy_uhd_prim2flx

  use Hydro_data,       ONLY : hy_entropy,hy_shockInstabilityFix

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! Arguments type declaration -----------
  integer, intent(IN) :: dir
  real, dimension(HY_VARINUM4), intent(IN) :: Vm, Vp
  real, dimension(HY_VARINUM),  intent(OUT) :: Fstar
  real, intent(OUT) :: vint
  integer, intent(IN) :: ix,iy,iz
  integer, dimension(LOW:HIGH,MDIM),intent(IN) :: blkLimitsGC
#ifdef FIXEDBLOCKSIZE
  real, dimension(NDIM,2, &
                  GRID_ILO_GC:GRID_IHI_GC, &
                  GRID_JLO_GC:GRID_JHI_GC, &
                  GRID_KLO_GC:GRID_KHI_GC),&
                  intent(IN) :: LambdaFix
#else
  real, dimension(NDIM,2,&
                  blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
                  blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
                  blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS)),&
                  intent(IN) :: LambdaFix
#endif
  !! --------------------------------------

  real,dimension(HY_VARINUM2) :: Vavg !DENS,VELX,VELY,VELZ,MAGX,MAGY,MAGZ,PRES + GAMC,GAME
  real,dimension(HY_VARINUM)   :: Um,Up,sigF,vec,FL,FR
  real,dimension(HY_WAVENUM)   :: lambda,lambdaL,lambdaR
  real,dimension(MDIM)         :: beta
  real,dimension(HY_WAVENUM,HY_VARINUM):: leig
  real,dimension(HY_VARINUM,HY_WAVENUM):: reig
  integer :: k
  logical :: cons
  real :: cs,ca,cf,as,af,bbN,uN
  real :: nu

  ! Second order Godunov flux
  cons=.true.
  call hy_uhd_avgState(dir,Vm(HY_DENS:HY_EINT),Vp(HY_DENS:HY_EINT),Vavg(HY_DENS:HY_GAME))
  call hy_uhd_eigenParameters(Vavg(HY_DENS:HY_GAME),dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
  call hy_uhd_eigenValue(lambda,uN,cf,C_alfn=ca,C_slow=cs)
  call hy_uhd_eigenVector(leig,reig,Vavg(HY_DENS:HY_GAME),dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)


  vint = Vavg(dir+1)

  call hy_uhd_prim2con(Vm(HY_DENS:HY_GAME),Um(HY_DENS:HY_MAGZ))
  call hy_uhd_prim2con(Vp(HY_DENS:HY_GAME),Up(HY_DENS:HY_MAGZ))
  call hy_uhd_prim2flx(dir,Vm(HY_DENS:HY_GAME),FL(F01DENS_FLUX:F08MAGZ_FLUX))
  call hy_uhd_prim2flx(dir,Vp(HY_DENS:HY_GAME),FR(F01DENS_FLUX:F08MAGZ_FLUX))

  if (hy_entropy) then
     cons=.false.
     call hy_uhd_eigenParameters(Vm(HY_DENS:HY_GAME),dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
     call hy_uhd_eigenValue(lambdaL,uN,cf,C_alfn=ca,C_slow=cs)
     call hy_uhd_eigenParameters(Vp(HY_DENS:HY_GAME),dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
     call hy_uhd_eigenValue(lambdaR,uN,cf,C_alfn=ca,C_slow=cs)
     call hy_uhd_entropyFix(lambda,lambdaL,lambdaR)
  endif

  ! Perform H-corrections for Carbuncle instability fix
  ! See Hanawa, Mikami, and Matsumoto, JCP, 227:7952-7976, 2008
  if (NDIM > 1 .and. hy_shockInstabilityFix) then
     do k=HY_SLOWLEFT,HY_SLOWRGHT
        select case (k)
        case(HY_SLOWLEFT,HY_ALFNLEFT,HY_ALFNRGHT,HY_SLOWRGHT)
           lambda(k) = max(LambdaFix(dir,1,ix,iy,iz),abs(lambda(k)))
        case(HY_ENTROPY)
           if (LambdaFix(dir,2,ix,iy,iz) > 0. ) then
              lambda(k) = &
                   abs((lambda(HY_FASTRGHT)+lambda(HY_FASTLEFT))*lambda(HY_ENTROPY)&
                       -2.*lambda(HY_FASTRGHT)*lambda(HY_FASTLEFT))&
                       /(lambda(HY_FASTRGHT)-lambda(HY_FASTLEFT))
           endif
        end select
     enddo
  endif


  sigF =0.
  vec  =0.

  do k=1,HY_WAVENUM

!!$     if (lambda(1)*lambda(HY_WAVENUM) < 0.) then
!!$
!!$        vec(HY_DENS:HY_MAGZ) = &!max(abs(maxval(lambda)),abs(minval(lambda)))
!!$             max(maxval(abs(lambda)),maxval(abs(lambdaL)),maxval(abs(lambdaR))  )&
!!$             *reig(HY_DENS:HY_MAGZ,k)&
!!$             *dot_product(leig(k,HY_DENS:HY_MAGZ),Up(HY_DENS:HY_MAGZ)-Um(HY_DENS:HY_MAGZ))
!!$     else
     vec(HY_DENS:HY_MAGZ)&
          = abs(lambda(k))*reig(HY_DENS:HY_MAGZ,k)&
          *dot_product(leig(k,HY_DENS:HY_MAGZ),Up(HY_DENS:HY_MAGZ)-Um(HY_DENS:HY_MAGZ))
!!$     endif

     sigF(F01DENS_FLUX:F08MAGZ_FLUX) = sigF(F01DENS_FLUX:F08MAGZ_FLUX) + vec(HY_DENS:HY_MAGZ)
  enddo

  ! Godunov Flux
  Fstar(F01DENS_FLUX:F08MAGZ_FLUX) = .5*(FR(F01DENS_FLUX:F08MAGZ_FLUX) &
                                       + FL(F01DENS_FLUX:F08MAGZ_FLUX) &
                                     - sigF(F01DENS_FLUX:F08MAGZ_FLUX))

End Subroutine hy_uhd_Roe
