!!****if* source/physics/Hydro/HydroMain/unsplit/Anisotropic_MHD/hy_uhd_addResistiveFluxes
!!
!! NAME
!!
!!  hy_uhd_addResistiveFluxes
!!
!!
!! SYNOPSIS
!!
!!  hy_uhd_addResistiveFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)
!!
!!  hy_uhd_addResistiveFluxes(integer(IN) :: blockID,
!!                            integer(IN) :: blkLimitsGC(LOW:HIGH,MDIM),
!!                            integer(IN) :: ix,
!!                            integer(IN) :: iy,
!!                            integer(IN) :: iz,
!!                            real(IN)    :: Flux,
!!                            real(IN)    :: TransCoeff,
!!                            integer(IN) :: sweepDir)
!!
!!
!! DESCRIPTION
!!
!!  Adds resistive flux contributions to total MHD fluxes
!!
!!
!! ARGUMENTS
!!
!!  blockID     - a local blockID
!!  blkLimitsGC - an array that holds the lower and upper indices of the section 
!!                of block with the guard cells 
!!  ix,iy,iz    - indices of the line along which the sweep is made
!!  Flux        - array containing MHD fluxes
!!  eta         - magnetic resistivity
!!  sweepDir    - direction of sweep
!!
!!***

!!REORDER(4): U

Subroutine hy_uhd_addResistiveFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)

  use Grid_interface, ONLY : Grid_getBlkPtr, Grid_releaseBlkPtr, Grid_getDeltas
  !use Hydro_data,     ONLY : hy_myPE

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! Argument List ----------------------------------------------------------
  integer, INTENT(IN) :: blockID,ix,iy,iz
  integer, dimension(LOW:HIGH,MDIM),intent(IN) :: blkLimitsGC
  real, dimension(HY_VARINUM), intent(INOUT) :: Flux

#ifdef FIXEDBLOCKSIZE
  real, dimension(HY_TRANSCOEFF_NUM, &
                  GRID_ILO_GC:GRID_IHI_GC, &
                  GRID_JLO_GC:GRID_JHI_GC, &
                  GRID_KLO_GC:GRID_KHI_GC), &
                  intent(IN) :: TransCoeff
#else
  real, dimension(HY_TRANSCOEFF_NUM, &
                  blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
                  blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
                  blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS)), &
                  intent(IN) :: TransCoeff
#endif
  integer, INTENT(IN) :: sweepDir
  !! ----------------------------------------------------------------------

  real    :: jx,jy,jz,idx,idy,idz,eta_loc
  real, dimension(MDIM) :: del
  real, pointer, dimension(:,:,:,:) :: U
  real, dimension(3,3) :: DBDt, Rotate, Working
  real, dimension(3) :: Jvec, Bvec

  real :: Bx, By, Bz, Bmag, Brho
  real :: eta_para, eta_perp, eta_Hall
  real :: DBxDy, DByDx, DBzDx
  real :: DBxDz, DByDz, DBzDy
  real :: c1, s1, c2, s2
  integer :: i, j
  !integer :: debugMsg
  logical :: local_isotropy

  !! Get deltas
  call Grid_getDeltas(blockID,del)

  idx=1./del(DIR_X)
  if (NDIM >= 2) then
     idy=1./del(DIR_Y)
     if (NDIM == 3) then
        idz=1./del(DIR_Z)
     endif
  endif

  !! Get pointer
  call Grid_getBlkPtr(blockID,U,CENTER)

  !--------------------------------------------------------------
  ! Get local values of B-field, resistance, and current
  !--------------------------------------------------------------

  select case(sweepDir)
  case(DIR_X) !Get local parameters for x-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix-1,iy,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix-1,iy,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix-1,iy,iz) )

     !Average cell-centered transport coefficients at face
     eta_para  = 0.5*( TransCoeff(HY_ETA_PARA,ix,iy,iz)+TransCoeff(HY_ETA_PARA,ix-1,iy,iz) )
     eta_perp  = 0.5*( TransCoeff(HY_ETA_PERP,ix,iy,iz)+TransCoeff(HY_ETA_PERP,ix-1,iy,iz) )
     eta_Hall  = 0.5*( TransCoeff(HY_ETA_HALL,ix,iy,iz)+TransCoeff(HY_ETA_HALL,ix-1,iy,iz) )

     !Calculate derivatives at face
     DByDx  =      idx*( U(MAGY_VAR,ix,iy,iz)-U(MAGY_VAR,ix-1,iy,iz) )
     DBzDx  =      idx*( U(MAGZ_VAR,ix,iy,iz)-U(MAGZ_VAR,ix-1,iy,iz) )

#if NDIM >= 2
     DBxDy  = 0.25*idy*( U(MAGX_VAR,ix  ,iy+1,iz)-U(MAGX_VAR,ix  ,iy-1,iz)+&
                         U(MAGX_VAR,ix-1,iy+1,iz)-U(MAGX_VAR,ix-1,iy-1,iz) )
     DBzDy  = 0.25*idy*( U(MAGZ_VAR,ix  ,iy+1,iz)-U(MAGZ_VAR,ix  ,iy-1,iz)+&
                         U(MAGZ_VAR,ix-1,iy+1,iz)-U(MAGZ_VAR,ix-1,iy-1,iz) )
#else
     DBxDy  = 0.
     DBzDy  = 0.
#endif

#if NDIM == 3
     DBxDz  = 0.25*idz*( U(MAGX_VAR,ix  ,iy,iz+1)-U(MAGX_VAR,ix  ,iy,iz-1)+&
                         U(MAGX_VAR,ix-1,iy,iz+1)-U(MAGX_VAR,ix-1,iy,iz-1) )
     DByDz  = 0.25*idz*( U(MAGY_VAR,ix  ,iy,iz+1)-U(MAGY_VAR,ix  ,iy,iz-1)+&
                         U(MAGY_VAR,ix-1,iy,iz+1)-U(MAGY_VAR,ix-1,iy,iz-1) )
#else
     DBxDz  = 0.
     DByDz  = 0.
#endif

  case(DIR_Y) !Get local parameters for y-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy-1,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy-1,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy-1,iz) )

     !Average cell-centered transport coefficients at face
     eta_para  = 0.5*( TransCoeff(HY_ETA_PARA,ix,iy,iz)+TransCoeff(HY_ETA_PARA,ix,iy-1,iz) )
     eta_perp  = 0.5*( TransCoeff(HY_ETA_PERP,ix,iy,iz)+TransCoeff(HY_ETA_PERP,ix,iy-1,iz) )
     eta_Hall  = 0.5*( TransCoeff(HY_ETA_HALL,ix,iy,iz)+TransCoeff(HY_ETA_HALL,ix,iy-1,iz) )

     !Calculate derivatives at face
     DByDx  = 0.25*idx*( U(MAGY_VAR,ix+1,iy  ,iz)-U(MAGY_VAR,ix-1,iy  ,iz)+&
                         U(MAGY_VAR,ix+1,iy-1,iz)-U(MAGY_VAR,ix-1,iy-1,iz) )
     DBzDx  = 0.25*idx*( U(MAGZ_VAR,ix+1,iy  ,iz)-U(MAGZ_VAR,ix-1,iy  ,iz)+&
                         U(MAGZ_VAR,ix+1,iy-1,iz)-U(MAGZ_VAR,ix-1,iy-1,iz) )
     DBxDy  =      idy*( U(MAGX_VAR,ix,iy,iz)-U(MAGX_VAR,ix,iy-1,iz) )
     DBzDy  =      idy*( U(MAGZ_VAR,ix,iy,iz)-U(MAGZ_VAR,ix,iy-1,iz) )

#if NDIM == 3
     DBxDz  = 0.25*idz*( U(MAGX_VAR,ix,iy  ,iz+1)-U(MAGX_VAR,ix,iy  ,iz-1)+&
                         U(MAGX_VAR,ix,iy-1,iz+1)-U(MAGX_VAR,ix,iy-1,iz-1) )
     DByDz  = 0.25*idz*( U(MAGY_VAR,ix,iy  ,iz+1)-U(MAGY_VAR,ix,iy  ,iz-1)+&
                         U(MAGY_VAR,ix,iy-1,iz+1)-U(MAGY_VAR,ix,iy-1,iz-1) )
#else
     DBxDz  = 0.
     DByDz  = 0.
#endif

  case(DIR_Z) !Get local parameters for z-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy,iz-1) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy,iz-1) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy,iz-1) )

     !Average cell-centered transport coefficients at face
     eta_para  = 0.5*( TransCoeff(HY_ETA_PARA,ix,iy,iz)+TransCoeff(HY_ETA_PARA,ix,iy,iz-1) )
     eta_perp  = 0.5*( TransCoeff(HY_ETA_PERP,ix,iy,iz)+TransCoeff(HY_ETA_PERP,ix,iy,iz-1) )
     eta_Hall  = 0.5*( TransCoeff(HY_ETA_HALL,ix,iy,iz)+TransCoeff(HY_ETA_HALL,ix,iy,iz-1) )

     !Calculate derivatives at face
     DByDx  = 0.25*idx*( U(MAGY_VAR,ix+1,iy,iz  )-U(MAGY_VAR,ix-1,iy,iz  )+&
                         U(MAGY_VAR,ix+1,iy,iz-1)-U(MAGY_VAR,ix-1,iy,iz-1) )
     DBzDx  = 0.25*idx*( U(MAGZ_VAR,ix+1,iy,iz  )-U(MAGZ_VAR,ix-1,iy,iz  )+&
                         U(MAGZ_VAR,ix+1,iy,iz-1)-U(MAGZ_VAR,ix-1,iy,iz-1) )
     DBxDy  = 0.25*idy*( U(MAGX_VAR,ix,iy+1,iz  )-U(MAGX_VAR,ix,iy-1,iz  )+&
                         U(MAGX_VAR,ix,iy+1,iz-1)-U(MAGX_VAR,ix,iy-1,iz-1) )
     DBzDy  = 0.25*idy*( U(MAGZ_VAR,ix,iy+1,iz  )-U(MAGZ_VAR,ix,iy-1,iz  )+&
                         U(MAGZ_VAR,ix,iy+1,iz-1)-U(MAGZ_VAR,ix,iy-1,iz-1) )
     DBxDz  =      idz*( U(MAGX_VAR,ix,iy,iz)-U(MAGX_VAR,ix,iy,iz-1) )
     DByDz  =      idz*( U(MAGY_VAR,ix,iy,iz)-U(MAGY_VAR,ix,iy,iz-1) )
     
  end select

  !--------------------------------------------------------------
  ! Determine resistive fluxes
  !--------------------------------------------------------------

  !Calculate magnitude and orientation of magnetic field
  Bmag = Bx**2 + By**2
  Brho = sqrt(Bmag)
  Bmag = sqrt(Bmag + Bz**2)

  !Calculate current
  Jvec(1) = DBzDy - DByDz
  Jvec(2) = DBxDz - DBzDx
  Jvec(3) = DByDx - DBxDy

  !debugMsg = 0
  !if( debugMsg == 1 ) then
  
  !   Bvec(1) = Bx
  !   Bvec(2) = By
  !   Bvec(3) = Bz

  !   if (hy_myPE .EQ. MASTER_PE .and. sweepDir == DIR_X) then
  !      if( Bx > 1e-15 .or. By > 1e-15 .or. Bz > 1e-15  ) then
  !         if( Jvec(1) > 1e-15 .or. Jvec(2) > 1e-15 .or. Jvec(3) > 1e-15  ) then
  !            print *, "Coords=", ix, iy, iz
  !            print *, "B0=", Bvec
  !            print *, "J0=", Jvec
  !            print *, "eta=", eta_para, eta_perp, eta_Hall
  !            debugMsg = 2
  !         endif
  !      endif
  !   endif
  !endif

  !check for isotropic resistivity
  if( eta_para == eta_perp .and. eta_Hall == 0. ) then
     local_isotropy = .true.
  else
     local_isotropy = .false.
  endif

  if( local_isotropy .or. Bmag == 0 ) then

     !Isotropic magnetic field flux
     DBDt(1,1) =  0.
     DBDt(1,2) = -eta_para*Jvec(3)
     DBDt(1,3) =  eta_para*Jvec(2)
     DBDt(2,1) =  eta_para*Jvec(3)
     DBDt(2,2) =  0.
     DBDt(2,3) = -eta_para*Jvec(1)
     DBDt(3,1) = -eta_para*Jvec(2)
     DBDt(3,2) =  eta_para*Jvec(1)
     DBDt(3,3) =  0.

  else

     !Align current to B-field if necessary
     if( Brho /= 0 ) then
        !Parameters for the rotation matrix
        c1 =  By/Brho
        s1 =  Bx/Brho
        c2 =  Bz/Bmag
        s2 =  Brho/Bmag

        !Setup rotation matrix
        Rotate(1,1) = c1
        Rotate(1,2) = -s1
        Rotate(1,3) = 0.
        Rotate(2,1) = s1*c2
        Rotate(2,2) = c1*c2
        Rotate(2,3) = -s2
        Rotate(3,1) = s1*s2
        Rotate(3,2) = c1*s2
        Rotate(3,3) = c2

        !get R*J, Jvec(1) = perp, Jvec(2) = Hall, Jvec(3) = para
        Jvec = MATMUL(Rotate, Jvec)
        !if( debugMsg == 2 ) then
        !   Bvec = MATMUL(Rotate, Bvec)
        !endif
     endif

     !Anisotropic magnetic field flux
     DBDt(1,1) =  0.
     DBDt(1,2) = -eta_para*Jvec(3)
     DBDt(1,3) =  eta_Hall*Jvec(1) + eta_perp*Jvec(2)
     DBDt(2,1) =  eta_para*Jvec(3)
     DBDt(2,2) =  0.
     DBDt(2,3) = -eta_perp*Jvec(1) + eta_Hall*Jvec(2)
     DBDt(3,1) = -eta_Hall*Jvec(1) - eta_perp*Jvec(2)
     DBDt(3,2) =  eta_perp*Jvec(1) - eta_Hall*Jvec(2)
     DBDt(3,3) =  0.
     
     !Rotate magnetic field flux back to lab frame
     if( Brho /= 0 ) then
        !first step of (R^T*P)*R, get R^T*P
        DO I = 1, 3
          DO J = 1, 3
            Working(I, J) = SUM( Rotate(:,I)*DBDt(:,J) )
          END DO
        END DO
        !second step of (R^T*P)*R
        DBDt = MATMUL(Working, Rotate)
     endif

  endif !Bmag /= 0

  !if (debugMsg == 2) then
  !   print *, "B_rot=", Bvec
  !   print *, "J_rot=", Jvec
  !   print *, "Rotate="
  !   print *, Rotate(1,:)
  !   print *, Rotate(2,:)
  !   print *, Rotate(3,:)
  !   print *, "DBDt="
  !   print *, DBDt(1,:)
  !   print *, DBDt(2,:)
  !   print *, DBDt(3,:)
  !endif

  !Compute resistive parts and add them to flux components
  select case(sweepDir)
  case(DIR_X)
     Flux(F07MAGY_FLUX) = Flux(F07MAGY_FLUX) + DBDt(1,2)
     Flux(F08MAGZ_FLUX) = Flux(F08MAGZ_FLUX) + DBDt(1,3)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) + Bz*DBDt(1,2) + By*DBDt(1,3)
     
  case(DIR_Y)
     Flux(F08MAGZ_FLUX) = Flux(F08MAGZ_FLUX) + DBDt(2,3)
     Flux(F06MAGX_FLUX) = Flux(F06MAGX_FLUX) + DBDt(2,1)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) + Bx*DBDt(2,3) + Bz*DBDt(2,1)
     
  case(DIR_Z)
     Flux(F06MAGX_FLUX) = Flux(F06MAGX_FLUX) + DBDt(3,1)
     Flux(F07MAGY_FLUX) = Flux(F07MAGY_FLUX) + DBDt(3,2)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) + By*DBDt(3,1) + Bx*DBDt(3,2)
  end select

  !--------------------------------------------------------------

  !! Release pointer
  call Grid_releaseBlkPtr(blockID,U,CENTER)

End Subroutine hy_uhd_addResistiveFluxes
