!!****if* source/physics/Hydro/HydroMain/unsplit/hy_uhd_addThermalFluxes
!!
!! NAME
!!
!!  hy_uhd_addThermalFluxes
!!
!!
!! SYNOPSIS
!!
!!  hy_uhd_addThermalFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)
!!
!!  hy_uhd_addThermalFluxes(integer(IN) :: blockID,
!!                          integer(IN) :: blkLimitsGC(LOW:HIGH,MDIM),
!!                          integer(IN) :: ix,
!!                          integer(IN) :: iy,
!!                          integer(IN) :: iz,
!!                          real(IN)    :: Flux,
!!                          real(IN)    :: kappa,
!!                          integer(IN) :: sweepDir)
!!
!!
!! DESCRIPTION
!!
!!  Adds thermal flux contributions to total fluxes
!!
!!
!! ARGUMENTS
!!
!!  blockID     - a local blockID
!!  blkLimitsGC - an array that holds the lower and upper indices of the section 
!!                of block with the guard cells 
!!  ix,iy,iz    - indices of the line along which the sweep is made
!!  Flux        - array containing MHD fluxes
!!  kappa       - thermal conductivity
!!  sweepDir    - direction of sweep
!!
!!***

!!REORDER(4): U

Subroutine hy_uhd_addThermalFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)

  use Hydro_data,     ONLY : hy_qref
  use Grid_interface, ONLY : Grid_getBlkPtr, Grid_releaseBlkPtr, Grid_getDeltas

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! Argument List ----------------------------------------------------------
  integer, INTENT(IN) :: blockID,ix,iy,iz
  integer, dimension(LOW:HIGH,MDIM),intent(IN) :: blkLimitsGC 
  real, dimension(HY_VARINUM), intent(INOUT) :: Flux

#ifdef FIXEDBLOCKSIZE
  real, dimension(HY_TRANSCOEFF_NUM, &
                  GRID_ILO_GC:GRID_IHI_GC, &
                  GRID_JLO_GC:GRID_JHI_GC, &
                  GRID_KLO_GC:GRID_KHI_GC), &
                  intent(IN) :: TransCoeff
#else
  real, dimension(HY_TRANSCOEFF_NUM, &
                  blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
                  blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
                  blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS)), &
                  intent(IN) :: TransCoeff
#endif
  integer, INTENT(IN) :: sweepDir
  !! ----------------------------------------------------------------------

  real    :: idx,idy,idz,kappa_loc
  real, dimension(MDIM) :: del
  real, pointer, dimension(:,:,:,:) :: U

  real :: Bx, By, Bz, Bmag
  real :: K_para, K_perp, K_caret
  real :: DTDx, DTDy, DTDz
  !real :: limDTDx, limDTDy, limDTDz
  real, dimension(NDIM,NDIM) :: ConductionK
  !real, parameter :: lim = 0.005

  !! Get deltas
  call Grid_getDeltas(blockID,del)

  idx=1./del(DIR_X)
  if (NDIM >= 2) then
     idy=1./del(DIR_Y)
     if (NDIM == 3) then
        idz=1./del(DIR_Z)
     endif
  endif

  !! Get pointer
  call Grid_getBlkPtr(blockID,U,CENTER)

  select case(sweepDir)
  case(DIR_X) !Get local parameters for x-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix-1,iy,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix-1,iy,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix-1,iy,iz) )
     
     !Average cell-centered transport coefficients at face
     K_para  = 0.5*(TransCoeff(HY_K_PARA ,ix,iy,iz)+TransCoeff(HY_K_PARA ,ix-1,iy,iz))
     K_perp  = 0.5*(TransCoeff(HY_K_PERP ,ix,iy,iz)+TransCoeff(HY_K_PERP ,ix-1,iy,iz))
     K_caret = 0.5*(TransCoeff(HY_K_CARET,ix,iy,iz)+TransCoeff(HY_K_CARET,ix-1,iy,iz))

     !Calculate derivatives at face
     DTDx    = idx*( U(TEMP_VAR,ix,iy,iz)-U(TEMP_VAR,ix-1,iy,iz) )
     !limDTDx = 0.5*idx*lim*( U(TEMP_VAR,ix,iy,iz)+U(TEMP_VAR,ix-1,iy,iz) )
     !DTDx    = sign( min( abs(DTDx), limDTDx ), DTDx )
        
#if NDIM >= 2
     DTDy    = 0.25*idy*( U(TEMP_VAR,ix  ,iy+1,iz)-U(TEMP_VAR,ix  ,iy-1,iz)+&
                          U(TEMP_VAR,ix-1,iy+1,iz)-U(TEMP_VAR,ix-1,iy-1,iz) )
     !limDTDy = 0.25*idy*lim*( U(TEMP_VAR,ix  ,iy+1,iz)+U(TEMP_VAR,ix  ,iy-1,iz)+&
     !                         U(TEMP_VAR,ix-1,iy+1,iz)+U(TEMP_VAR,ix-1,iy-1,iz) )
     !DTDy    = sign( min( abs(DTDy), limDTDy ), DTDy )

#if NDIM == 3
     DTDz    = 0.25*idz*( U(TEMP_VAR,ix  ,iy,iz+1)-U(TEMP_VAR,ix  ,iy,iz-1)+&
                          U(TEMP_VAR,ix-1,iy,iz+1)-U(TEMP_VAR,ix-1,iy,iz-1) )
     !limDTDz = 0.25*idz*lim*( U(TEMP_VAR,ix  ,iy,iz+1)+U(TEMP_VAR,ix  ,iy,iz-1)+&
     !                         U(TEMP_VAR,ix-1,iy,iz+1)+U(TEMP_VAR,ix-1,iy,iz-1) )
     !DTDz    = sign( min( abs(DTDz), limDTDz ), DTDz )
#endif !NDIM == 3

  case(DIR_Y) !Get local parameters for y-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy-1,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy-1,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy-1,iz) )
     
     !Average cell-centered transport coefficients at face
     K_para  = 0.5*(TransCoeff(HY_K_PARA ,ix,iy,iz)+TransCoeff(HY_K_PARA ,ix,iy-1,iz))
     K_perp  = 0.5*(TransCoeff(HY_K_PERP ,ix,iy,iz)+TransCoeff(HY_K_PERP ,ix,iy-1,iz))
     K_caret = 0.5*(TransCoeff(HY_K_CARET,ix,iy,iz)+TransCoeff(HY_K_CARET,ix,iy-1,iz))

     !Calculate derivatives at face
     DTDx    = 0.25*idx*( U(TEMP_VAR,ix+1,iy  ,iz)-U(TEMP_VAR,ix-1,iy  ,iz)+&
                          U(TEMP_VAR,ix+1,iy-1,iz)-U(TEMP_VAR,ix-1,iy-1,iz) )
     !limDTDx = 0.25*idx*lim*( U(TEMP_VAR,ix+1,iy  ,iz)+U(TEMP_VAR,ix-1,iy  ,iz)+&
     !                         U(TEMP_VAR,ix+1,iy-1,iz)+U(TEMP_VAR,ix-1,iy-1,iz) )
     !DTDx    = sign( min( abs(DTDx), limDTDx ), DTDx )

     DTDy    = idy*( U(TEMP_VAR,ix,iy,iz)-U(TEMP_VAR,ix,iy-1,iz) )
     !limDTDy = 0.5*idy*lim*( U(TEMP_VAR,ix,iy,iz)+U(TEMP_VAR,ix,iy-1,iz) )
     !DTDy    = sign( min( abs(DTDy), limDTDy ), DTDy )

#if NDIM == 3
     DTDz    = 0.25*idz*( U(TEMP_VAR,ix,iy  ,iz+1)-U(TEMP_VAR,ix,iy  ,iz-1)+&
                          U(TEMP_VAR,ix,iy-1,iz+1)-U(TEMP_VAR,ix,iy-1,iz-1) )
     !limDTDz = 0.25*idz*lim*( U(TEMP_VAR,ix,iy  ,iz+1)+U(TEMP_VAR,ix,iy  ,iz-1)+&
     !                         U(TEMP_VAR,ix,iy-1,iz+1)+U(TEMP_VAR,ix,iy-1,iz-1) )
     !DTDz    = sign( min( abs(DTDz), limDTDz ), DTDz )

  case(DIR_Z) !Get local parameters for z-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy,iz-1) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy,iz-1) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy,iz-1) )
     
     !Average cell-centered transport coefficients at face
     K_para  = 0.5*(TransCoeff(HY_K_PARA ,ix,iy,iz)+TransCoeff(HY_K_PARA ,ix,iy,iz-1))
     K_perp  = 0.5*(TransCoeff(HY_K_PERP ,ix,iy,iz)+TransCoeff(HY_K_PERP ,ix,iy,iz-1))
     K_caret = 0.5*(TransCoeff(HY_K_CARET,ix,iy,iz)+TransCoeff(HY_K_CARET,ix,iy,iz-1))

     !Calculate derivatives at face
     DTDx    = 0.25*idx*( U(TEMP_VAR,ix+1,iy,iz  )-U(TEMP_VAR,ix-1,iy,iz  )+&
                          U(TEMP_VAR,ix+1,iy,iz-1)-U(TEMP_VAR,ix-1,iy,iz-1) )
     !limDTDx = 0.25*idx*lim*( U(TEMP_VAR,ix+1,iy,iz  )+U(TEMP_VAR,ix-1,iy,iz  )+&
     !                         U(TEMP_VAR,ix+1,iy,iz-1)+U(TEMP_VAR,ix-1,iy,iz-1) )
     !DTDx    = sign( min( abs(DTDx), limDTDx ), DTDx )


     DTDy    = 0.25*idy*( U(TEMP_VAR,ix,iy+1,iz  )-U(TEMP_VAR,ix,iy-1,iz  )+&
                          U(TEMP_VAR,ix,iy+1,iz-1)-U(TEMP_VAR,ix,iy-1,iz-1) )
     !limDTDy = 0.25*idy*lim*( U(TEMP_VAR,ix,iy+1,iz  )+U(TEMP_VAR,ix,iy-1,iz  )+&
     !                         U(TEMP_VAR,ix,iy+1,iz-1)+U(TEMP_VAR,ix,iy-1,iz-1) )
     !DTDy    = sign( min( abs(DTDy), limDTDy ), DTDy )

     DTDz    = idz*( U(TEMP_VAR,ix,iy,iz)-U(TEMP_VAR,ix,iy,iz-1) )
     !limDTDz = 0.5*idz*lim*( U(TEMP_VAR,ix,iy,iz)+U(TEMP_VAR,ix,iy-1,iz) )
     !DTDz    = sign( min( abs(DTDz), limDTDz ), DTDz )

#endif !NDIM == 3
#endif !NDIM >= 2
  end select

  !normalize magnetic field
  Bmag = sqrt(Bx**2 + By**2 + Bz**2)
  if( Bmag == 0.0 ) then
     K_perp = K_para
     K_caret = 0.
     Bx = 1.
     By = 1.
     Bz = 1.
  else
     Bx = Bx/Bmag
     By = By/Bmag
     Bz = Bz/Bmag
  endif

  !if( ix==10 .and. iy==10 ) then
  !   WRITE(6,*) "addThermalFluxes: Bx=", Bx, "By=", By, "Bz=", Bz, "|B|=", Bmag
  !   WRITE(6,*) " sweep=", sweepDir, "K||=", K_para, "K_perp=", K_perp, "K^=", K_caret
  !end if

  !if( hy_myPE == MASTER_PE ) then
  !   if( DTDx > 1. .or. DTDy > 1. .or. DTDz > 1. ) then
  !      print *, "aTF: Coords=", ix, iy, iz, sweepDir
  !      print *, "aTF: B=", Bx, By, Bz
  !      print *, "aTF: DTDx=", DTDx, DTDy, DTDz
  !      print *, "aTF: K_comp=", K_para, K_perp, K_caret
  !      print *, "aTF: flux0=", Flux(F05ENER_FLUX)
  !   endif
  !endif

  select case(sweepDir)
  case(DIR_X) !Calculate thermal energy flux for x-sweep

     !Calculate conduction coefficients and fluxes
     ConductionK(DIR_X,DIR_X) = K_para*Bx**2 + K_perp*(1.-Bx**2)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_X,DIR_X)*DTDx
#if NDIM >= 2
     ConductionK(DIR_X,DIR_Y) = (K_para-K_perp)*Bx*By - K_caret*Bz
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_X,DIR_Y)*DTDy
#if NDIM == 3
     ConductionK(DIR_X,DIR_Z) = (K_para-K_perp)*Bx*Bz + K_caret*By
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_X,DIR_Z)*DTDz
#endif !NDIM == 3

  case(DIR_Y) !Calculate thermal energy flux for y-sweep

     !Calculate conduction coefficients and fluxes
     ConductionK(DIR_Y,DIR_X) = (K_para-K_perp)*By*Bx + K_caret*Bz
     ConductionK(DIR_Y,DIR_Y) = K_para*By**2 + K_perp*(1.-By**2)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_Y,DIR_X)*DTDx &
                          -ConductionK(DIR_Y,DIR_Y)*DTDy
#if NDIM == 3
     ConductionK(DIR_Y,DIR_Z) = (K_para-K_perp)*By*Bz - K_caret*Bx
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_Y,DIR_Z)*DTDz
     
  case(DIR_Z) !Calculate thermal energy flux for z-sweep

     ConductionK(DIR_Z,DIR_X) = (K_para-K_perp)*Bz*Bx - K_caret*By
     ConductionK(DIR_Z,DIR_Y) = (K_para-K_perp)*Bz*By + K_caret*Bx
     ConductionK(DIR_Z,DIR_Z) = K_para*Bz**2 + K_perp*(1.-Bz**2)
     Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) &
                          -ConductionK(DIR_Z,DIR_X)*DTDx &
                          -ConductionK(DIR_Z,DIR_Y)*DTDy &
                          -ConductionK(DIR_Z,DIR_Z)*DTDz
#endif !NDIM == 3
#endif !NDIM >= 2
  end select

  !if( hy_myPE == MASTER_PE ) then
  !   if( DTDx > 1. .or. DTDy > 1. .or. DTDz > 1. ) then
  !      print *, "aTF: flux =", Flux(F05ENER_FLUX)
  !      print *, "aTF: K_mtrx=", ConductionK(sweepDir,:)
  !   endif
  !endif

  call Grid_releaseBlkPtr(blockID,U,CENTER)

End Subroutine hy_uhd_addThermalFluxes
