!!****if* source/physics/Hydro/HydroMain/unsplit/hy_uhd_addViscousFluxes
!!
!! NAME
!!
!!  hy_uhd_addViscousFluxes
!!
!!
!! SYNOPSIS
!!
!!  hy_uhd_addViscousFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)
!!
!!  hy_uhd_addViscousFluxes(integer(IN) :: blockID,
!!                          integer(IN) :: blkLimitsGC(LOW:HIGH,MDIM),
!!                          integer(IN) :: ix,
!!                          integer(IN) :: iy,
!!                          integer(IN) :: iz,
!!                          real(IN)    :: Flux,
!!                          real(IN)    :: mu,
!!                          integer(IN) :: sweepDir)
!!
!!
!! DESCRIPTION
!!
!!  Adds viscous flux contributions to total fluxes
!!
!!
!! ARGUMENTS
!!
!!  blockID     - a local blockID
!!  blkLimitsGC - an array that holds the lower and upper indices of the section 
!!                of block with the guard cells 
!!  ix,iy,iz    - indices of the line along which the sweep is made
!!  Flux        - array containing MHD fluxes
!!  mu          - dynamic viscosity
!!  sweepDir    - direction of sweep
!!
!!***

!!REORDER(4): U

Subroutine hy_uhd_addViscousFluxes(blockID,blkLimitsGC,ix,iy,iz,Flux,TransCoeff,sweepDir)

  use Grid_interface, ONLY : Grid_getBlkPtr, Grid_releaseBlkPtr, Grid_getDeltas
  !use Hydro_data,     ONLY : hy_myPE

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! Argument List ----------------------------------------------------------
  integer, INTENT(IN) :: blockID,ix,iy,iz
  integer, dimension(LOW:HIGH,MDIM),intent(IN) :: blkLimitsGC 
  real, dimension(HY_VARINUM), intent(INOUT) :: Flux

#ifdef FIXEDBLOCKSIZE
  real, dimension(HY_TRANSCOEFF_NUM, &
                  GRID_ILO_GC:GRID_IHI_GC, &
                  GRID_JLO_GC:GRID_JHI_GC, &
                  GRID_KLO_GC:GRID_KHI_GC), &
                  intent(IN) :: TransCoeff
#else
  real, dimension(HY_TRANSCOEFF_NUM, &
                  blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
                  blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
                  blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS)), &
                  intent(IN) :: TransCoeff
#endif
  integer, INTENT(IN) :: sweepDir
  !! ----------------------------------------------------------------------

  real :: idx,idy,idz,mu_loc
  real, dimension(MDIM) :: del
  real, pointer, dimension(:,:,:,:) :: U
  real, dimension(3,3) :: DvDx, StressP, Rotate, Working
  real, dimension(3) :: Bvec, vvec

  real :: Bx, By, Bz, Brho, Bmag
  real :: pi0, pi1, pi2, pi3, pi4
  real :: Wxx, Wxy, Wxz, Wyy, Wyz, Wzz
  real :: c1, s1, c2, s2
  real :: vx, vy, vz
  integer :: i, j
  !integer :: debugMsg
  logical :: local_isotropy

  !! Get deltas
  call Grid_getDeltas(blockID,del)

  idx=1./del(DIR_X)
  if (NDIM >= 2) then
     idy=1./del(DIR_Y)
     if (NDIM == 3) then
        idz=1./del(DIR_Z)
     endif
  endif

  !! Get pointer
  call Grid_getBlkPtr(blockID,U,CENTER)

  !--------------------------------------------------------------
  ! Get local values of B-field, velocity, and velocity Jacobian
  !--------------------------------------------------------------

  select case(sweepDir)
  case(DIR_X) !Get local parameters for x-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix-1,iy,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix-1,iy,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix-1,iy,iz) )

     !Average cell-centered velocity at face
     vx = 0.5*( U(VELX_VAR,ix,iy,iz)+U(VELX_VAR,ix-1,iy,iz) )
     vy = 0.5*( U(VELY_VAR,ix,iy,iz)+U(VELY_VAR,ix-1,iy,iz) )
     vz = 0.5*( U(VELZ_VAR,ix,iy,iz)+U(VELZ_VAR,ix-1,iy,iz) )

     !Average cell-centered transport coefficients at face
     pi0 = 0.5*( TransCoeff(HY_PI0,ix,iy,iz)+TransCoeff(HY_PI0,ix-1,iy,iz) )
     pi1 = 0.5*( TransCoeff(HY_PI1,ix,iy,iz)+TransCoeff(HY_PI1,ix-1,iy,iz) )
     pi2 = 0.5*( TransCoeff(HY_PI2,ix,iy,iz)+TransCoeff(HY_PI2,ix-1,iy,iz) )
     pi3 = 0.5*( TransCoeff(HY_PI3,ix,iy,iz)+TransCoeff(HY_PI3,ix-1,iy,iz) )
     pi4 = 0.5*( TransCoeff(HY_PI4,ix,iy,iz)+TransCoeff(HY_PI4,ix-1,iy,iz) )


     !Calculate derivatives at face
     DvDx(1,1) =      idx*( U(VELX_VAR,ix,iy,iz)-U(VELX_VAR,ix-1,iy,iz) )
     DvDx(2,1) =      idx*( U(VELY_VAR,ix,iy,iz)-U(VELY_VAR,ix-1,iy,iz) )
     DvDx(3,1) =      idx*( U(VELZ_VAR,ix,iy,iz)-U(VELZ_VAR,ix-1,iy,iz) )

#if NDIM >= 2
     DvDx(1,2) = 0.25*idy*( U(VELX_VAR,ix  ,iy+1,iz)-U(VELX_VAR,ix  ,iy-1,iz)+&
                            U(VELX_VAR,ix-1,iy+1,iz)-U(VELX_VAR,ix-1,iy-1,iz) )
     DvDx(2,2) = 0.25*idy*( U(VELY_VAR,ix  ,iy+1,iz)-U(VELY_VAR,ix  ,iy-1,iz)+&
                            U(VELY_VAR,ix-1,iy+1,iz)-U(VELY_VAR,ix-1,iy-1,iz) )
     DvDx(3,2) = 0.25*idy*( U(VELZ_VAR,ix  ,iy+1,iz)-U(VELZ_VAR,ix  ,iy-1,iz)+&
                            U(VELZ_VAR,ix-1,iy+1,iz)-U(VELZ_VAR,ix-1,iy-1,iz) )
#else
     DvDx(1,2) = 0.
     DvDx(2,2) = 0.
     DvDx(3,2) = 0.
#endif

#if NDIM == 3
     DvDx(1,3) = 0.25*idz*( U(VELX_VAR,ix  ,iy,iz+1)-U(VELX_VAR,ix  ,iy,iz-1)+&
                            U(VELX_VAR,ix-1,iy,iz+1)-U(VELX_VAR,ix-1,iy,iz-1) )
     DvDx(2,3) = 0.25*idz*( U(VELY_VAR,ix  ,iy,iz+1)-U(VELY_VAR,ix  ,iy,iz-1)+&
                            U(VELY_VAR,ix-1,iy,iz+1)-U(VELY_VAR,ix-1,iy,iz-1) )
     DvDx(3,3) = 0.25*idz*( U(VELZ_VAR,ix  ,iy,iz+1)-U(VELZ_VAR,ix  ,iy,iz-1)+&
                            U(VELZ_VAR,ix-1,iy,iz+1)-U(VELZ_VAR,ix-1,iy,iz-1) )
#else
     DvDx(1,3) = 0.
     DvDx(2,3) = 0.
     DvDx(3,3) = 0.
#endif

  case(DIR_Y) !Get local parameters for y-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy-1,iz) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy-1,iz) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy-1,iz) )

     !Average cell-centered velocity at face
     vx = 0.5*( U(VELX_VAR,ix,iy,iz)+U(VELX_VAR,ix,iy-1,iz) )
     vy = 0.5*( U(VELY_VAR,ix,iy,iz)+U(VELY_VAR,ix,iy-1,iz) )
     vz = 0.5*( U(VELZ_VAR,ix,iy,iz)+U(VELZ_VAR,ix,iy-1,iz) )

     !Average cell-centered transport coefficients at face
     pi0 = 0.5*( TransCoeff(HY_PI0,ix,iy,iz)+TransCoeff(HY_PI0,ix,iy-1,iz) )
     pi1 = 0.5*( TransCoeff(HY_PI1,ix,iy,iz)+TransCoeff(HY_PI1,ix,iy-1,iz) )
     pi2 = 0.5*( TransCoeff(HY_PI2,ix,iy,iz)+TransCoeff(HY_PI2,ix,iy-1,iz) )
     pi3 = 0.5*( TransCoeff(HY_PI3,ix,iy,iz)+TransCoeff(HY_PI3,ix,iy-1,iz) )
     pi4 = 0.5*( TransCoeff(HY_PI4,ix,iy,iz)+TransCoeff(HY_PI4,ix,iy-1,iz) )

     !Calculate derivatives at face
     DvDx(1,1) = 0.25*idx*( U(VELX_VAR,ix+1,iy  ,iz)-U(VELX_VAR,ix-1,iy  ,iz)+&
                            U(VELX_VAR,ix+1,iy-1,iz)-U(VELX_VAR,ix-1,iy-1,iz) )
     DvDx(2,1) = 0.25*idx*( U(VELY_VAR,ix+1,iy  ,iz)-U(VELY_VAR,ix-1,iy  ,iz)+&
                            U(VELY_VAR,ix+1,iy-1,iz)-U(VELY_VAR,ix-1,iy-1,iz) )
     DvDx(3,1) = 0.25*idx*( U(VELZ_VAR,ix+1,iy  ,iz)-U(VELZ_VAR,ix-1,iy  ,iz)+&
                            U(VELZ_VAR,ix+1,iy-1,iz)-U(VELZ_VAR,ix-1,iy-1,iz) )
     DvDx(1,2) =      idy*( U(VELX_VAR,ix,iy,iz)-U(VELX_VAR,ix,iy-1,iz) )
     DvDx(2,2) =      idy*( U(VELY_VAR,ix,iy,iz)-U(VELY_VAR,ix,iy-1,iz) )
     DvDx(3,2) =      idy*( U(VELZ_VAR,ix,iy,iz)-U(VELZ_VAR,ix,iy-1,iz) )

#if NDIM == 3
     DvDx(1,3) = 0.25*idz*( U(VELX_VAR,ix,iy  ,iz+1)-U(VELX_VAR,ix,iy  ,iz-1)+&
                            U(VELX_VAR,ix,iy-1,iz+1)-U(VELX_VAR,ix,iy-1,iz-1) )
     DvDx(2,3) = 0.25*idz*( U(VELY_VAR,ix,iy  ,iz+1)-U(VELY_VAR,ix,iy  ,iz-1)+&
                            U(VELY_VAR,ix,iy-1,iz+1)-U(VELY_VAR,ix,iy-1,iz-1) )
     DvDx(3,3) = 0.25*idz*( U(VELZ_VAR,ix,iy  ,iz+1)-U(VELZ_VAR,ix,iy  ,iz-1)+&
                            U(VELZ_VAR,ix,iy-1,iz+1)-U(VELZ_VAR,ix,iy-1,iz-1) )
#else
     DvDx(1,3) = 0.
     DvDx(2,3) = 0.
     DvDx(3,3) = 0.
#endif

  case(DIR_Z) !Get local parameters for z-sweep

     !Average cell-centered magnetic field at face
     Bx = 0.5*( U(MAGX_VAR,ix,iy,iz)+U(MAGX_VAR,ix,iy,iz-1) )
     By = 0.5*( U(MAGY_VAR,ix,iy,iz)+U(MAGY_VAR,ix,iy,iz-1) )
     Bz = 0.5*( U(MAGZ_VAR,ix,iy,iz)+U(MAGZ_VAR,ix,iy,iz-1) )

     !Average cell-centered velocity at face
     vx = 0.5*( U(VELX_VAR,ix,iy,iz)+U(VELX_VAR,ix,iy,iz-1) )
     vy = 0.5*( U(VELY_VAR,ix,iy,iz)+U(VELY_VAR,ix,iy,iz-1) )
     vz = 0.5*( U(VELZ_VAR,ix,iy,iz)+U(VELZ_VAR,ix,iy,iz-1) )

     !Average cell-centered transport coefficients at face
     pi0 = 0.5*( TransCoeff(HY_PI0,ix,iy,iz)+TransCoeff(HY_PI0,ix,iy,iz-1) )
     pi1 = 0.5*( TransCoeff(HY_PI1,ix,iy,iz)+TransCoeff(HY_PI1,ix,iy,iz-1) )
     pi2 = 0.5*( TransCoeff(HY_PI2,ix,iy,iz)+TransCoeff(HY_PI2,ix,iy,iz-1) )
     pi3 = 0.5*( TransCoeff(HY_PI3,ix,iy,iz)+TransCoeff(HY_PI3,ix,iy,iz-1) )
     pi4 = 0.5*( TransCoeff(HY_PI4,ix,iy,iz)+TransCoeff(HY_PI4,ix,iy,iz-1) )

     !Calculate derivatives at face
     DvDx(1,1)  = 0.25*idx*( U(VELX_VAR,ix+1,iy,iz  )-U(VELX_VAR,ix-1,iy,iz  )+&
                             U(VELX_VAR,ix+1,iy,iz-1)-U(VELX_VAR,ix-1,iy,iz-1) )
     DvDx(2,1)  = 0.25*idx*( U(VELY_VAR,ix+1,iy,iz  )-U(VELY_VAR,ix-1,iy,iz  )+&
                             U(VELY_VAR,ix+1,iy,iz-1)-U(VELY_VAR,ix-1,iy,iz-1) )
     DvDx(3,1)  = 0.25*idx*( U(VELZ_VAR,ix+1,iy,iz  )-U(VELZ_VAR,ix-1,iy,iz  )+&
                             U(VELZ_VAR,ix+1,iy,iz-1)-U(VELZ_VAR,ix-1,iy,iz-1) )
     DvDx(1,2)  = 0.25*idy*( U(VELX_VAR,ix,iy+1,iz  )-U(VELX_VAR,ix,iy-1,iz  )+&
                             U(VELX_VAR,ix,iy+1,iz-1)-U(VELX_VAR,ix,iy-1,iz-1) )
     DvDx(2,2)  = 0.25*idy*( U(VELY_VAR,ix,iy+1,iz  )-U(VELY_VAR,ix,iy-1,iz  )+&
                             U(VELY_VAR,ix,iy+1,iz-1)-U(VELY_VAR,ix,iy-1,iz-1) )
     DvDx(3,2)  = 0.25*idy*( U(VELZ_VAR,ix,iy+1,iz  )-U(VELZ_VAR,ix,iy-1,iz  )+&
                             U(VELZ_VAR,ix,iy+1,iz-1)-U(VELZ_VAR,ix,iy-1,iz-1) )
     DvDx(1,3)  =      idz*( U(VELX_VAR,ix,iy,iz)-U(VELX_VAR,ix,iy,iz-1) )
     DvDx(2,3)  =      idz*( U(VELY_VAR,ix,iy,iz)-U(VELY_VAR,ix,iy,iz-1) )
     DvDx(3,3)  =      idz*( U(VELZ_VAR,ix,iy,iz)-U(VELZ_VAR,ix,iy,iz-1) )
     
  end select

  !--------------------------------------------------------------
  ! Determine viscous fluxes
  !--------------------------------------------------------------

  !Calculate magnitude and orientation of magnetic field
  Bmag = Bx**2 + By**2
  Brho = sqrt(Bmag)
  Bmag = sqrt(Bmag + Bz**2)

  !debugMsg = 0
  !if( debugMsg == 1 ) then

  !   Bvec(1) = Bx
  !   Bvec(2) = By
  !   Bvec(3) = Bz
  !   vvec(1) = vx
  !   vvec(2) = vy
  !   vvec(3) = vz

  !   if (hy_myPE .EQ. MASTER_PE .and. sweepDir == DIR_X) then
  !      if( Bx > 1e-15 .or. By > 1e-15 .or. Bz > 1e-15  ) then
  !         if( vvec(1) > 1e-15 .or. vvec(2) > 1e-15 .or. vvec(3) > 1e-15  ) then
  !            print *, "Coords=", ix, iy, iz
  !            print *, "B0=", Bvec
  !            print *, "v0=", vvec
  !            print *, "pi=", pi0, pi1, pi2, pi3, pi4
  !            debugMsg = 2
  !         endif
  !      endif
  !   endif
  !endif

  !check for isotropic viscosity
  if( pi0 == pi1 .and. pi0 == pi2 .and. pi3 == 0. .and. pi4 == 0. ) then
     local_isotropy = .true.
  else
     local_isotropy = .false.
  endif

  if( local_isotropy .or. Bmag == 0 ) then
     !Calculate the rate of strain tensor
     Wxx = (2./3.)*(2.*DvDx(1,1) - DvDx(2,2) - DvDx(3,3))
     Wxy = DvDx(1,2) + DvDx(2,1)
     Wxz = DvDx(1,3) + DvDx(3,1)
     Wyy = (2./3.)*(2.*DvDx(2,2) - DvDx(3,3) - DvDx(1,1))
     Wyz = DvDx(2,3) + DvDx(3,2)
     Wzz = (2./3.)*(2.*DvDx(3,3) - DvDx(1,1) - DvDx(2,2))

     !Isotropic stress tensor
     StressP(1,1) = -pi0*Wxx
     StressP(1,2) = -pi0*Wxy
     StressP(1,3) = -pi0*Wxz
     StressP(2,1) = StressP(1,2)
     StressP(2,2) = -pi0*Wyy
     StressP(2,3) = -pi0*Wyz
     StressP(3,1) = StressP(1,3)
     StressP(3,2) = StressP(2,3)
     StressP(3,3) = -pi0*Wzz

  else

     !Align velocity Jacobian to B-field if necessary
     if( Brho /= 0 ) then
        !Parameters for the rotation matrix
        c1 =  By/Brho
        s1 =  Bx/Brho
        c2 =  Bz/Bmag
        s2 =  Brho/Bmag

        !Setup rotation matrix
        Rotate(1,1) = c1
        Rotate(1,2) = -s1
        Rotate(1,3) = 0.
        Rotate(2,1) = s1*c2
        Rotate(2,2) = c1*c2
        Rotate(2,3) = -s2
        Rotate(3,1) = s1*s2
        Rotate(3,2) = c1*s2
        Rotate(3,3) = c2

        !first step of (R*J)*R^T, get R*J
        Working = MATMUL(Rotate, DvDx) 
        !second step of (R*J)*R^T
        DO I = 1, 3
          DO J = 1, 3
            DvDx(I, J) = SUM( Working(I,:)*Rotate(J,:) )
          END DO
        END DO

        !if (debugMsg == 2) then
        !   vvec = MATMUL(Rotate, vvec)
        !   Bvec = MATMUL(Rotate, Bvec)
        !endif
     endif

     !Calculate the rate of strain tensor
     Wxx = (2./3.)*(2.*DvDx(1,1) - DvDx(2,2) - DvDx(3,3))
     Wyy = (2./3.)*(2.*DvDx(2,2) - DvDx(3,3) - DvDx(1,1))
     Wzz = (2./3.)*(2.*DvDx(3,3) - DvDx(1,1) - DvDx(2,2))
     Wxy = DvDx(1,2) + DvDx(2,1)
     Wxz = DvDx(1,3) + DvDx(3,1)
     Wyz = DvDx(2,3) + DvDx(3,2)
     
     !Anisotropic stress tensor, given by NRL plasma formulary, pg 37
     StressP(1,1) = -0.5*pi0*(Wxx+Wyy) - 0.5*pi1*(Wxx-Wyy) - pi3*Wxy
     StressP(1,2) = -pi1*Wxy + 0.5*pi3*(Wxx-Wyy)
     StressP(1,3) = -pi2*Wxz - pi4*Wyz
     StressP(2,1) = StressP(1,2)
     StressP(2,2) = -0.5*pi0*(Wxx+Wyy) + 0.5*pi1*(Wxx-Wyy) + pi3*Wxy
     StressP(2,3) = -pi2*Wyz + pi4*Wxz
     StressP(3,1) = StressP(1,3)
     StressP(3,2) = StressP(2,3)
     StressP(3,3) = -pi0*Wzz
     
     !Rotate stress back to lab frame
     if( Brho /= 0 ) then
        !first step of (R^T*P)*R, get R^T*P
        DO I = 1, 3
          DO J = 1, 3
            Working(I, J) = SUM( Rotate(:,I)*StressP(:,J) )
          END DO
        END DO
        !second step of (R^T*P)*R
        StressP = MATMUL(Working, Rotate)
     endif

  endif !Bmag /= 0

  !if (debugMsg == 2) then
  !   print *, "B_rot=", Bvec
  !   print *, "v_rot=", vvec
  !   print *, "Rotate="
  !   print *, Rotate(1,:)
  !   print *, Rotate(2,:)
  !   print *, Rotate(3,:)
  !   print *, "StressP="
  !   print *, StressP(1,:)
  !   print *, StressP(2,:)
  !   print *, StressP(3,:)
  !endif

  !Add results to fluxes
  Flux(F02XMOM_FLUX) = Flux(F02XMOM_FLUX) + StressP(sweepDir,1)
  Flux(F03YMOM_FLUX) = Flux(F03YMOM_FLUX) + StressP(sweepDir,2)
  Flux(F04ZMOM_FLUX) = Flux(F04ZMOM_FLUX) + StressP(sweepDir,3)
  Flux(F05ENER_FLUX) = Flux(F05ENER_FLUX) + vx*StressP(sweepDir,1) &
                                          + vy*StressP(sweepDir,2) &
                                          + vz*StressP(sweepDir,3)
  
  !--------------------------------------------------------------

  !! Release pointer
  call Grid_releaseBlkPtr(blockID,U,CENTER)

End Subroutine hy_uhd_addViscousFluxes
