!!****if* source/physics/Hydro/HydroMain/unsplit/Anisotropic_MHD/hy_uhd_dataReconstOneStep
!!
!! NAME
!!
!!  hy_uhd_dataReconstOneStep
!!
!! SYNOPSIS
!!
!!  hy_uhd_dataReconstOneStep(integer(IN) :: blockID,
!!                            integer(IN) :: blkLimitsGC(:,:),
!!                            integer(IN) :: ix,
!!                            integer(IN) :: iy,
!!                            integer(IN) :: iz,
!!                            real(IN)    :: dt,
!!                            real(IN)    :: del(MDIM),
!!                            real(IN)    :: ogravX(:,:,:),
!!                            real(IN)    :: ogravY(:,:,:),
!!                            real(IN)    :: ogravZ(:,:,:),
!!                            real(IN)    :: V0 (HY_VARINUM2),
!!                            real(IN)    :: Vxp(HY_VARINUM2),
!!                            real(IN)    :: Vxn(HY_VARINUM2),
!!                            real(IN)    :: Vyp(HY_VARINUM2),
!!                            real(IN)    :: Vyn(HY_VARINUM2),
!!                            real(IN)    :: Vzp(HY_VARINUM2),
!!                            real(IN)    :: Vzn(HY_VARINUM2),
!!                            real(IN)    :: Vxpp(HY_VARINUM2),
!!                            real(IN)    :: Vxnn(HY_VARINUM2),
!!                            real(IN)    :: Vypp(HY_VARINUM2),
!!                            real(IN)    :: Vynn(HY_VARINUM2),
!!                            real(IN)    :: Vzpp(HY_VARINUM2),
!!                            real(IN)    :: Vznn(HY_VARINUM2),
!!                            real(IN)    :: Vxppp(HY_VARINUM2),
!!                            real(IN)    :: Vxnnn(HY_VARINUM2),
!!                            real(IN)    :: Vyppp(HY_VARINUM2),
!!                            real(IN)    :: Vynnn(HY_VARINUM2),
!!                            real(IN)    :: Vzppp(HY_VARINUM2),
!!                            real(IN)    :: Vznnn(HY_VARINUM2),
!!                            real(IN)    :: FlatCoeff(:,:,:,:),
!!                            real(OUT)   :: lambda(NDIM,HY_WAVENUM),
!!                            real(OUT)   :: Wxp(HY_VARINUM3),
!!                            real(OUT)   :: Wxn(HY_VARINUM3),
!!                            real(OUT)   :: Wyp(HY_VARINUM3),
!!                            real(OUT)   :: Wyn(HY_VARINUM3),
!!                            real(OUT)   :: Wzp(HY_VARINUM3),
!!                            real(OUT)   :: Wzn(HY_VARINUM3))
!!
!! ARGUMENTS
!!
!!  blockID     - local block ID
!!  blkLimitsGC - block limits including guardcells
!!  ix,iy,iz    - local indices
!!  dt          - timestep
!!  del         - deltas in each {x,y,z} direction
!!  ogravX,ogravY,ogravZ - gravity components at n step in each direction
!!  V0          - array containing primitive variables + gamc + game at cell center
!!  Vxp,Vxn,Vyp,Vyn,Vzp,Vzn,Vxpp,Vxnn,Vypp,Vynn,Vzpp,Vznn - data in neighboring cells
!!  Vxppp,Vxnnn,Vyppp,Vynnn,Vzppp,Vznnn - data in neighboring cells
!!  FlatCoeff   - flattening parameters
!!  lambda      - eigenvalue is passed out and is stored for shock instability fix if needed
!!  Wxp,Wxn,Wyp,Wyn,Wzp,Wzn - left(m) and right(p) states
!!
!! DESCRIPTION
!!
!!  This onestep data reconstruction routine evolves the cell centered 
!!  values by dt/2 time step at each cell interface using characteristic 
!!  tracing method based on the second (MUSCL-Hancock) or third (PPM) order
!!  multidimensional scheme.
!!  The cell interface values are reconstructed in multidimensional
!!  fashion with inclusions of MHD multidimensional terms that are
!!  proportional to divB for the unsplit MHD implementation.
!!
!!
!! REFERENCES
!!
!!  * Toro, Riemann Solvers and Numerical Methods for Fluid Dynamics, Springer, 1997
!!  * Lee, D., Ph.D. Dissertation, Univ. of MD, 2006
!!  * Lee, D. and Deane, A., "An Unsplit Staggered Mesh Scheme for Multidimensional
!!                            Magnetohydrodynamics", 228 (2009), 952-975, JCP
!!  * Stone, Gardiner, Teuben, Hawley, Simon, "Athena: A new code for astrophysical MHD"
!!    arXiv:0804.0402v1 [astro-ph] 2 Apr 2008
!!  * Colella and Woodward, 54, 174 (1984), JCP
!!  * Colella, 87, 171-200 (1990), JCP
!!
!!***


!!REORDER(4): B[xyz]

Subroutine hy_uhd_dataReconstOneStep(blockID,blkLimitsGC,ix,iy,iz, &
                                     dt,del,ogravX,ogravY,ogravZ,V0,  &
                                     Vxp,  Vxn,  Vyp,  Vyn,  Vzp,  Vzn,  &
                                     Vxpp, Vxnn, Vypp, Vynn, Vzpp, Vznn, &
                                     Vxppp,Vxnnn,Vyppp,Vynnn,Vzppp,Vznnn,&
                                     FlatCoeff, lambda,                  &
                                     Wxp, Wxn, Wyp, Wyn, Wzp, Wzn )

  use Hydro_data,        ONLY : hy_eswitch, hy_killdivb, hy_forceHydroLimit
  use Grid_interface,    ONLY : Grid_getBlkPtr, Grid_releaseBlkPtr

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"


  !!-----Arguments---------------------------------------------------------
  integer,intent(IN) :: blockID
  integer,intent(IN),dimension(LOW:HIGH,MDIM):: blkLimitsGC
  integer,intent(IN) :: ix,iy,iz
  real,   intent(IN) :: dt
  real,   intent(IN), dimension(MDIM) :: del
#ifdef FIXEDBLOCKSIZE
  real, dimension(GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC), intent(IN) :: ogravX,ogravY,ogravZ
  real, dimension(NDIM,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC), intent(IN) :: FlatCoeff
#else
  real, dimension(blkLimitsGC(HIGH,IAXIS),blkLimitsGC(HIGH,JAXIS),blkLimitsGC(HIGH,KAXIS)), &
       intent(IN) :: ogravX,ogravY,ogravZ
  real, dimension(NDIM,blkLimitsGC(HIGH,IAXIS),blkLimitsGC(HIGH,JAXIS),blkLimitsGC(HIGH,KAXIS)), &
       intent(IN) :: FlatCoeff
#endif
  real, intent(INOUT),  dimension(HY_VARINUM3) :: V0, Vxp,  Vxn,  Vyp,  Vyn,  Vzp,  Vzn, &
                                                    Vxpp, Vxnn, Vypp, Vynn, Vzpp, Vznn,&
                                                    Vxppp,Vxnnn,Vyppp,Vynnn,Vzppp,Vznnn
  real, intent(OUT), dimension(NDIM,HY_WAVENUM) :: lambda
  real, intent(OUT), dimension(HY_VARINUM3)    :: Wxp, Wxn, Wyp, Wyn, Wzp, Wzn
  !!------------------------------------------------------------------------

  real :: dx,dy,dz,hdtx,hdty,hdtz,hdt,uB,dnBx,dnBy,dnBz
  real :: tinyD,tinyP
  real, dimension(HY_VARINUM3)  :: sigX,sigY,sigZ
  real, dimension(HY_VARINUM)   :: aBx,aBy,aBz
  integer :: hyEnd


#if NDIM > 1
  real, dimension(:,:,:,:), pointer :: Bx, By, Bz
#if NDIM < 3
  !!NAG complains if this isn't here, since Bz is uninitialized.
  nullify(Bz)
#endif !NDIM < 3 
#endif !NDIM > 1


#ifndef GRAVITY
  hyEnd = HY_GAME
#else
  hyEnd = HY_GRAV
#endif


  hdt=.5*dt

  dx  =del(DIR_X)
  hdtx=0.5*dt/dx
#if NDIM >= 2
  dy  =del(DIR_Y)
  hdty=0.5*dt/dy
#if NDIM == 3
  dz  =del(DIR_Z)
  hdtz=0.5*dt/dz
#endif
#endif

  Wxp = 0.
  Wxn = 0.
  Wyp = 0.
  Wyn = 0.
  Wzp = 0.
  Wzn = 0.

#if NFACE_VARS > 0
#if NDIM >=2
  call Grid_getBlkPtr(blockID,Bx,FACEX)
  call Grid_getBlkPtr(blockID,By,FACEY)
#if NDIM == 3
  call Grid_getBlkPtr(blockID,Bz,FACEZ)
#endif
#endif
#endif

  !dot product of U & B fields
  uB = dot_product(V0(HY_VELX:HY_VELZ),V0(HY_MAGX:HY_MAGZ))

  !Multidimensional MHD terms
  if (.not. hy_forceHydroLimit) then
     aBx= (/0.                     , &
           -V0(HY_MAGX)/V0(HY_DENS), &
           -V0(HY_MAGY)/V0(HY_DENS), &
           -V0(HY_MAGZ)/V0(HY_DENS), &
           (V0(HY_GAME)-1.)*uB     , &     
            0.                     , &
           -V0(HY_VELY)            , &
           -V0(HY_VELZ)             /)

     aBy=aBx
     aBy(HY_MAGX)=-V0(HY_VELX)
     aBy(HY_MAGY)=0.

     aBz=aBx
     aBz(HY_MAGX)=-V0(HY_VELX)
     aBz(HY_MAGZ)=0.
  else
     aBx = 0.
     aBy = 0.
     aBz = 0.
  endif


  !! ***********************************************************
  !! (I) Perform a characteristic tracing in x-direction       *
  !! ***********************************************************
  !! Define the derivative of the normal B field
  dnBx = 0.
  dnBy = 0.
  dnBz = 0.

#if NFACE_VARS > 0
#if NDIM > 1
  if (hy_killdivb) then
     dnBx = Bx(MAG_FACE_VAR,ix+1,iy,iz)-Bx(MAG_FACE_VAR,ix,iy,iz)
     if (NDIM >= 2) then
        dnBy = By(MAG_FACE_VAR,ix,iy+1,iz)-By(MAG_FACE_VAR,ix,iy,iz)
        if (NDIM == 3) then
           dnBz = Bz(MAG_FACE_VAR,ix,iy,iz+1)-Bz(MAG_FACE_VAR,ix,iy,iz)
        endif
     endif
  else
     dnBx = .5*(Vxp(HY_MAGX)-Vxn(HY_MAGX))
     if (NDIM >= 2) then
        dnBy = .5*(Vyp(HY_MAGY)-Vyn(HY_MAGY))
        if (NDIM == 3) then
           dnBz = .5*(Vzp(HY_MAGZ)-Vzn(HY_MAGZ))
        endif
     endif
  endif
#endif
#endif

  !! Initialize data at the most distant neighboring location i+3 and i-3, which
  !! are only available for NGUARD > 4

#if NGUARD > 4
  Vxppp(HY_GRAV) = ogravX(ix+3,iy,iz)
  Vxnnn(HY_GRAV) = ogravX(ix-3,iy,iz)
#endif

  V0  (HY_GRAV) = ogravX(ix,  iy,iz)
  Vxp (HY_GRAV) = ogravX(ix+1,iy,iz)
  Vxn (HY_GRAV) = ogravX(ix-1,iy,iz)
  Vxpp(HY_GRAV) = ogravX(ix+2,iy,iz)
  Vxnn(HY_GRAV) = ogravX(ix-2,iy,iz)
  call DataReconstructNormalDir(DIR_X,dt,dx,V0,Vxp,Vxn,Vxpp,Vxnn,Vxppp,Vxnnn,&
                                dnBx,aBx,&
                                FlatCoeff(DIR_X,ix,iy,iz), &
                                lambda(DIR_X,1:HY_WAVENUM),&
                                Wxp,Wxn,sigX)
#if NDIM >= 2
  if (NDIM>=2) then
#if NGUARD > 4
     Vyppp(HY_GRAV) = ogravY(ix,iy+3,iz)
     Vynnn(HY_GRAV) = ogravY(ix,iy-3,iz)
#endif
     V0  (HY_GRAV) = ogravY(ix,iy,  iz)
     Vyp (HY_GRAV) = ogravY(ix,iy+1,iz)
     Vyn (HY_GRAV) = ogravY(ix,iy-1,iz)
     Vypp(HY_GRAV) = ogravY(ix,iy+2,iz)
     Vynn(HY_GRAV) = ogravY(ix,iy-2,iz)
     call DataReconstructNormalDir(DIR_Y,dt,dy,V0,Vyp,Vyn,Vypp,Vynn,Vyppp,Vynnn,&
                                   dnBy,aBy,&
                                   FlatCoeff(DIR_Y,ix,iy,iz), &
                                   lambda(DIR_Y,1:HY_WAVENUM),&
                                   Wyp,Wyn,sigY)
#if NDIM == 3
     if (NDIM == 3) then
#if NGUARD > 4
        Vzppp(HY_GRAV) = ogravZ(ix,iy,iz+3)
        Vznnn(HY_GRAV) = ogravZ(ix,iy,iz-3)
#endif
        V0  (HY_GRAV) = ogravZ(ix,iy,iz  )
        Vzp (HY_GRAV) = ogravZ(ix,iy,iz+1)
        Vzn (HY_GRAV) = ogravZ(ix,iy,iz-1)
        Vzpp(HY_GRAV) = ogravZ(ix,iy,iz+2)
        Vznn(HY_GRAV) = ogravZ(ix,iy,iz-2)
        !!NAG complains about the array bound for FlatCoeff.
        !!since this is not executed except in 3d this should be ok
        !!for now. --PMR
        call DataReconstructNormalDir(DIR_Z,dt,dz,V0,Vzp,Vzn,Vzpp,Vznn,Vzppp,Vznnn,&
                                      dnBz,aBz,&
                                      FlatCoeff(DIR_Z,ix,iy,iz), &
                                      lambda(DIR_Z,1:HY_WAVENUM),&
                                      Wzp,Wzn,sigZ)
    endif
#endif  !NDIM == 3
  endif
#endif !NDIM >= 2


#if NDIM >= 2
  if (NDIM > 1) then
     !! *************************************************************
     !! (II) Perform to make contributions from transverse fluxes   *
     !! *************************************************************
     !! This is needed only for 2D & 3D

     !! For 2D
     Wxn(HY_DENS:HY_MAGZ) = Wxn(HY_DENS:HY_MAGZ) - hdty*sigY(HY_DENS:HY_MAGZ)     &
                                                 - hdty*aBy(HY_DENS:HY_MAGZ)*dnBy
     Wxp(HY_DENS:HY_MAGZ) = Wxp(HY_DENS:HY_MAGZ) - hdty*sigY(HY_DENS:HY_MAGZ)     &
                                                 - hdty*aBy(HY_DENS:HY_MAGZ)*dnBy
     ! gamc, game, and gravity
     Wxn(HY_GAMC:hyEnd)   = Wxn(HY_GAMC:hyEnd)   - hdty*sigY(HY_GAMC:hyEnd)
     Wxp(HY_GAMC:hyEnd)   = Wxp(HY_GAMC:hyEnd)   - hdty*sigY(HY_GAMC:hyEnd)


     Wyn(HY_DENS:HY_MAGZ) = Wyn(HY_DENS:HY_MAGZ) - hdtx*sigX(HY_DENS:HY_MAGZ)     &
                                                 - hdtx*aBx(HY_DENS:HY_MAGZ)*dnBx
     Wyp(HY_DENS:HY_MAGZ) = Wyp(HY_DENS:HY_MAGZ) - hdtx*sigX(HY_DENS:HY_MAGZ)     &
                                                 - hdtx*aBx(HY_DENS:HY_MAGZ)*dnBx
     ! gamc, game, and gravity
     Wyn(HY_GAMC:hyEnd)   = Wyn(HY_GAMC:hyEnd)   - hdtx*sigX(HY_GAMC:hyEnd)
     Wyp(HY_GAMC:hyEnd)   = Wyp(HY_GAMC:hyEnd)   - hdtx*sigX(HY_GAMC:hyEnd)

#if NDIM == 3
     Wxn(HY_DENS:HY_MAGZ) = Wxn(HY_DENS:HY_MAGZ) - hdtz*sigZ(HY_DENS:HY_MAGZ)     &
                                                 - hdtz*aBz(HY_DENS:HY_MAGZ)*dnBz
     Wxp(HY_DENS:HY_MAGZ) = Wxp(HY_DENS:HY_MAGZ) - hdtz*sigZ(HY_DENS:HY_MAGZ)     &
                                                 - hdtz*aBz(HY_DENS:HY_MAGZ)*dnBz
     ! gamc, game, and gravity
     Wxn(HY_GAMC:hyEnd)   = Wxn(HY_GAMC:hyEnd)   - hdtz*sigZ(HY_GAMC:hyEnd)
     Wxp(HY_GAMC:hyEnd)   = Wxp(HY_GAMC:hyEnd)   - hdtz*sigZ(HY_GAMC:hyEnd)


     Wyn(HY_DENS:HY_MAGZ) = Wyn(HY_DENS:HY_MAGZ) - hdtz*sigZ(HY_DENS:HY_MAGZ)     &
                                                 - hdtz*aBz(HY_DENS:HY_MAGZ)*dnBz
     Wyp(HY_DENS:HY_MAGZ) = Wyp(HY_DENS:HY_MAGZ) - hdtz*sigZ(HY_DENS:HY_MAGZ)     &
                                                 - hdtz*aBz(HY_DENS:HY_MAGZ)*dnBz
     ! gamc, game, and gravity
     Wyn(HY_GAMC:hyEnd)   = Wyn(HY_GAMC:hyEnd)   - hdtz*sigZ(HY_GAMC:hyEnd)
     Wyp(HY_GAMC:hyEnd)   = Wyp(HY_GAMC:hyEnd)   - hdtz*sigZ(HY_GAMC:hyEnd)


     Wzn(HY_DENS:HY_MAGZ) = Wzn(HY_DENS:HY_MAGZ) - hdtx*sigX(HY_DENS:HY_MAGZ)     &
                                                 - hdtx*aBx(HY_DENS:HY_MAGZ)*dnBx &
                                                 - hdty*sigY(HY_DENS:HY_MAGZ)     &
                                                 - hdty*aBy(HY_DENS:HY_MAGZ)*dnBy

     Wzp(HY_DENS:HY_MAGZ) = Wzp(HY_DENS:HY_MAGZ) - hdtx*sigX(HY_DENS:HY_MAGZ)     &
                                                 - hdtx*aBx(HY_DENS:HY_MAGZ)*dnBx &
                                                 - hdty*sigY(HY_DENS:HY_MAGZ)     &
                                                 - hdty*aBy(HY_DENS:HY_MAGZ)*dnBy

     ! gamc, game, and gravity
     Wzn(HY_GAMC:hyEnd)   = Wzn(HY_GAMC:hyEnd)   - hdtx*sigX(HY_GAMC:hyEnd) &
                                                 - hdty*sigY(HY_GAMC:hyEnd)

     Wzp(HY_GAMC:hyEnd)   = Wzp(HY_GAMC:hyEnd)   - hdtx*sigX(HY_GAMC:hyEnd) &
                                                 - hdty*sigY(HY_GAMC:hyEnd)
#endif
  endif
#endif



  !! ***************************************************************
  !! (III) Use the continuous normal fields at each cell interface *
  !! ***************************************************************

#if NFACE_VARS > 0
  ! 2D and 3D MHD runs
  if (hy_killdivb) then
     Wxn(HY_MAGX)   = Bx(MAG_FACE_VAR,ix,   iy,  iz  )
     Wxp(HY_MAGX)   = Bx(MAG_FACE_VAR,ix+1, iy,  iz  )

     Wyn(HY_MAGY)   = By(MAG_FACE_VAR,ix,  iy,   iz  )
     Wyp(HY_MAGY)   = By(MAG_FACE_VAR,ix,  iy+1, iz  )

#if NDIM == 3
     Wzn(HY_MAGZ)   = Bz(MAG_FACE_VAR,ix,  iy,   iz  )
     Wzp(HY_MAGZ)   = Bz(MAG_FACE_VAR,ix,  iy,   iz+1)
#endif

  endif
#else
  ! 1D MHD or pure hydro mode
  Wxn(HY_MAGX) = V0(HY_MAGX)
  Wxp(HY_MAGX) = V0(HY_MAGX)
  Wyn(HY_MAGY) = V0(HY_MAGY)
  Wyp(HY_MAGY) = V0(HY_MAGY)
  Wzn(HY_MAGZ) = V0(HY_MAGZ)
  Wzp(HY_MAGZ) = V0(HY_MAGZ)
#endif


  !! *************************************************************
  !! (IV) Avoid any possible negative states:                    *
  !!   - Use first-order Godunov scheme if they occur            *
  !! *************************************************************
  tinyD=hy_eswitch*V0(HY_DENS)
  tinyP=hy_eswitch*V0(HY_PRES)

  if (Wxn(HY_DENS) < tinyD .or. Wxp(HY_DENS) < tinyD .or. &
      Wxn(HY_PRES) < tinyP .or. Wxp(HY_PRES) < tinyP ) then
     Wxn(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
     Wxp(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
  endif
#if NDIM >= 2
  if (Wyn(HY_DENS) < tinyD .or. Wyp(HY_DENS) < tinyD .or. &
      Wyn(HY_PRES) < tinyP .or. Wyp(HY_PRES) < tinyP ) then
     Wyn(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
     Wyp(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
  endif
#if NDIM == 3
  if (Wzn(HY_DENS) < tinyD .or. Wzp(HY_DENS) < tinyD .or. &
      Wzn(HY_PRES) < tinyP .or. Wzp(HY_PRES) < tinyP ) then
     Wzn(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
     Wzp(HY_DENS:HY_MAGZ) = V0(HY_DENS:HY_MAGZ)
  endif
#endif
#endif

  !! Release pointers for facevars
#if NFACE_VARS > 0
#if NDIM >= 2
  call Grid_releaseBlkPtr(blockID,Bx,FACEX)
  call Grid_releaseBlkPtr(blockID,By,FACEY)
#if NDIM == 3
  call Grid_releaseBlkPtr(blockID,Bz,FACEZ)
#endif
#endif
#endif

End Subroutine hy_uhd_dataReconstOneStep




Subroutine DataReconstructNormalDir&
     (dir,dt,delta,Vc,Vp,Vm,Vpp,Vmm,Vppp,Vmmm,        &
      dnBn,aBn,FlatCoeff,lambda,Wp,Wm,sig)

  use Hydro_data,        ONLY : hy_charLimiting,   &
                                hy_eswitch,        &
                                hy_fullyLimit,     &
                                hy_RiemannSolver,  &
                                hy_tiny,           &
                                hy_order,          &
                                hy_entropy,        &
                                hy_flattening,     &
                                hy_transOrder,     &
                                hy_ContactSteepening,&
                                hy_upwindTVD

  use hy_uhd_interface,     ONLY : hy_uhd_TVDslope,        &
                                   hy_uhd_TVDslopeUpwind,  &
                                   hy_uhd_eigenParameters, &
                                   hy_uhd_eigenValue,      &
                                   hy_uhd_eigenVector

  use hy_uhd_slopeLimiters, ONLY : checkMedian

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !!-----Arguments---------------------------------------------------------
  integer,intent(IN) :: dir
  real,intent(IN) :: dt,delta
  real,intent(IN), dimension(HY_VARINUM3):: Vc,Vp,Vm,Vpp,Vmm,Vppp,Vmmm
  real,intent(IN) :: dnBn
  real,intent(IN), dimension(HY_VARINUM)  :: aBn
  real,intent(IN)  :: FlatCoeff
  real,intent(OUT),dimension(HY_WAVENUM)  :: lambda
  real,intent(OUT),dimension(HY_VARINUM3):: Wp,Wm
  real,intent(OUT),dimension(HY_VARINUM3):: sig
  !!------------------------------------------------------------------------
  integer :: n
  real    :: dtn,hdtn,factor
  real    :: constA,constB,constC,constD,lambdaMax,lambdaMin
  real    :: temp1, temp2, temp3, eta_steep, del2rhoR, del2rhoL
  real, dimension(HY_VARINUM)   :: vec,sigL,sigR
  real, dimension(HY_VARINUM3) :: vecL,vecR,delW,W6,delbar,delbarP,delbarN
  real, PARAMETER :: eta1=20.E0, eta2=0.05E0,epsln=0.01E0,K0=0.1E0

  ! EIG SYSTEM
  logical :: cons=.false.
  real    :: cs,ca,cf,as,af,uN
  real, dimension(MDIM) :: beta
  real, dimension(HY_WAVENUM)   :: lambdaP,lambdaN,lambdaPP,lambdaNN
  real, dimension(HY_WAVENUM,HY_VARINUM) :: leig,leigP,leigN,leigPP,leigNN
  real, dimension(HY_VARINUM,HY_WAVENUM) :: reig,reigP,reigN,reigPP,reigNN

  ! WENO parameters
  integer :: NVAR_DO
  real, dimension(HY_VARINUM3) :: Vc0,Vp0,Vpp0,Vm0,Vmm0,Vppp0,Vmmm0,& ! primitive variables
                                   Wc0,Wp0,Wpp0,Wm0,Wmm0,Wppp0,Wmmm0,& ! characteristic variables
                                   vecL_temp, vecR_temp, &
                                   vecL_temp1,vecR_temp1,&
                                   vecL_temp2,vecR_temp2
  real, dimension(3,3) :: coeff1
  real, dimension(3):: coeff2
  real, dimension(3):: W5p,W5m,IS5p,IS5m,Alpha5p,Alpha5m,omega,omegaBar,V5p,V5m
  real, dimension(6):: Intp
  real :: sumAlpha
  real :: errorCheck,errorCheck1
!!$  real :: theta
  real :: Flattening
  real :: Mach
  real,dimension(HY_VARINUM3) :: TransFlux
  integer :: TVDorder = 1
  integer :: hyEnd

#ifndef GRAVITY
  hyEnd = HY_GAME
#else
  hyEnd = HY_GRAV
#endif


  dtn=dt/delta
  hdtn=0.5*dtn

  vecL = 0.
  vecR = 0.
  vec  = 0.
  sigL = 0.
  sigR = 0.
  sig  = 0.


  !! Initialize in x-direction
  !! Cell center value
  Vc0(HY_DENS:HY_GAME) = Vc(HY_DENS:HY_GAME)

  if (dir==DIR_X) then
     Vc0(HY_MAGX)   = 0.
  elseif (dir==DIR_Y) then
     Vc0(HY_MAGY)   = 0.
  elseif (dir==DIR_Z) then
     Vc0(HY_MAGZ)   = 0.
  endif

  !! Prepare for 3rd order PPM or 5th order WENO
  if (hy_order >= 3) then
     Vp0 (HY_DENS:HY_GAME) = Vp (HY_DENS:HY_GAME)
     Vpp0(HY_DENS:HY_GAME) = Vpp(HY_DENS:HY_GAME)
     Vm0 (HY_DENS:HY_GAME) = Vm (HY_DENS:HY_GAME)
     Vmm0(HY_DENS:HY_GAME) = Vmm(HY_DENS:HY_GAME)
     if (hy_order == 5) then
        Vmmm0(HY_DENS:HY_GAME) = Vmmm(HY_DENS:HY_GAME)
        Vppp0(HY_DENS:HY_GAME) = Vppp(HY_DENS:HY_GAME)
     endif

     if (dir==DIR_X) then
        Vp0 (HY_MAGX) = 0.
        Vpp0(HY_MAGX) = 0.
        Vm0 (HY_MAGX) = 0.
        Vmm0(HY_MAGX) = 0.
        if (hy_order == 5) then
           Vmmm0(HY_MAGX) = 0.
           Vppp0(HY_MAGX) = 0.
        endif
     elseif (dir==DIR_Y) then
        Vp0 (HY_MAGY) = 0.
        Vpp0(HY_MAGY) = 0.
        Vm0 (HY_MAGY) = 0.
        Vmm0(HY_MAGY) = 0.
        if (hy_order == 5) then
           Vmmm0(HY_MAGY) = 0.
           Vmmm0(HY_MAGY) = 0.
        endif
     elseif (dir==DIR_Z) then
        Vp0 (HY_MAGZ) = 0.
        Vpp0(HY_MAGZ) = 0.
        Vm0 (HY_MAGZ) = 0.
        Vmm0(HY_MAGZ) = 0.
        if (hy_order == 5) then
           Vmmm0(HY_MAGZ) = 0.
           Vppp0(HY_MAGZ) = 0.
        endif
     endif

  endif

  !! -------------------------------------------------------------------------------------!
  !! Calculate eigensystem  --------------------------------------------------------------!
  !! -------------------------------------------------------------------------------------!
  call hy_uhd_eigenParameters(Vc,dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
  call hy_uhd_eigenValue(lambda,uN,cf,C_alfn=ca,C_slow=cs)
  call hy_uhd_eigenVector(leig,reig,Vc,dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)

  if ((hy_order > 2) .or. (hy_entropy)) then
     call hy_uhd_eigenParameters(Vp,dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
     call hy_uhd_eigenValue(lambdaP,uN,cf,C_alfn=ca,C_slow=cs)
     call hy_uhd_eigenVector(leigP,reigP,Vp,dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)

     call hy_uhd_eigenParameters(Vm,dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
     call hy_uhd_eigenValue(lambdaN,uN,cf,C_alfn=ca,C_slow=cs)
     call hy_uhd_eigenVector(leigN,reigN,Vm,dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)

     ! Entropy fix for low density region - this treatment is not needed for the data reconstruction
     ! in general, but users can turn it on when needed.
     call hy_uhd_entropyFix(lambda,lambdaN,lambdaP)
  endif


  !! -------------------------------------------------------------------------------------!
  !! (1) Apply TVD slope limiter for normal gradients ------------------------------------!
  !! -------------------------------------------------------------------------------------!
  if (NGUARD <= 4) hy_upwindTVD = .false.
  if (.not. hy_upwindTVD) then
     !! Original
        call hy_uhd_TVDslope&
             (dir,TVDorder,Vmm,Vm,Vc,Vp,Vpp,lambdaN,lambda,lambdaP,leig,delbar)

        if (hy_order >= 3) then
           call hy_uhd_TVDslope&
                (dir,TVDorder,Vm,Vc,Vp,Vpp,Vppp,lambda,lambdaP,lambdaPP,leigP,delbarP)

           call hy_uhd_TVDslope&
                (dir,TVDorder,Vmmm,Vmm,Vm,Vc,Vp,lambdaNN,lambdaN,lambda,leigN,delbarN)
        endif
  else

     !! Upwinded TVD begins here ----
     !! Check local Alfven Mach number
!!$     Mach = abs(lambda(HY_ENTROPY))/max(abs(lambda(HY_FASTRGHT)-lambda(HY_ENTROPY)),1.e-20)
!!$     if (Mach <= 1.) then
        if ((lambdaN(HY_FASTRGHT) > 0. .and. lambdaP(HY_FASTLEFT) < 0.) .or. &
            (lambdaN(HY_FASTRGHT) > 0. .and. lambda (HY_FASTLEFT) < 0.) .or. &
            (lambda (HY_FASTRGHT) > 0. .and. lambdaP(HY_FASTLEFT) < 0.)) then
           call hy_uhd_TVDslopeUpwind&
                (dir,Vmm,Vm,Vc,Vp,Vpp,lambdaN,lambda,lambdaP,leig,delbar)
!!$        else
!!$           call hy_uhd_TVDslope&
!!$                (dir,TVDorder,Vmm,Vm,Vc,Vp,Vpp,lambdaN,lambda,lambdaP,leig,delbar)
!!$        endif
     else
        call hy_uhd_TVDslope&
             (dir,TVDorder,Vmm,Vm,Vc,Vp,Vpp,lambdaN,lambda,lambdaP,leig,delbar)
     endif

     if (hy_order >= 3) then

        call hy_uhd_eigenParameters(Vpp,dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
        call hy_uhd_eigenValue(lambdaPP,uN,cf,C_alfn=ca,C_slow=cs)
!!$        call hy_uhd_eigenVector(leigPP,reigPP,Vpp,dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
!!$
        call hy_uhd_eigenParameters(Vmm,dir,cons,uN,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)
        call hy_uhd_eigenValue(lambdaNN,uN,cf,C_alfn=ca,C_slow=cs)
!!$        call hy_uhd_eigenVector(leigNN,reigNN,Vmm,dir,cons,cf,C_alfn=ca,C_slow=cs,A_f=af,A_s=as,B_beta=beta)

        if (hy_entropy) then
           call hy_uhd_entropyFix(lambdaP,lambda,lambdaPP)
           call hy_uhd_entropyFix(lambdaN,lambdaNN,lambda)
        endif

        !! Check local Alfven Mach number
!!$        Mach = abs(lambdaP(HY_ENTROPY))/max(abs(lambdaP(HY_FASTRGHT)-lambdaP(HY_ENTROPY)),1.e-20)
!!$        if (Mach <= 1.) then
           if ((lambda (HY_FASTRGHT) > 0. .and. lambdaPP(HY_FASTLEFT) < 0.) .or. &
               (lambda (HY_FASTRGHT) > 0. .and. lambdaP (HY_FASTLEFT) < 0.) .or. &
               (lambdaP(HY_FASTRGHT) > 0. .and. lambdaPP(HY_FASTLEFT) < 0.)) then
              call hy_uhd_TVDslopeUpwind&
                   (dir,Vm,Vc,Vp,Vpp,Vppp,lambda,lambdaP,lambdaPP,leigP,delbarP)
!!$           else
!!$              call hy_uhd_TVDslope&
!!$                   (dir,TVDorder,Vm,Vc,Vp,Vpp,Vppp,lambda,lambdaP,lambdaPP,leigP,delbarP)
!!$           endif
        else
           call hy_uhd_TVDslope&
                (dir,TVDorder,Vm,Vc,Vp,Vpp,Vppp,lambda,lambdaP,lambdaPP,leigP,delbarP)
        endif

        !! Check local Alfven Mach number
!!$        Mach = abs(lambdaN(HY_ENTROPY))/max(abs(lambdaN(HY_FASTRGHT)-lambdaN(HY_ENTROPY)),1.e-20)
!!$        if (Mach <= 1.) then
           if ((lambdaNN(HY_FASTRGHT) > 0. .and. lambda (HY_FASTLEFT) < 0.) .or. &
               (lambdaNN(HY_FASTRGHT) > 0. .and. lambdaN(HY_FASTLEFT) < 0.) .or. &
               (lambdaN (HY_FASTRGHT) > 0. .and. lambda (HY_FASTLEFT) < 0.)) then
              call hy_uhd_TVDslopeUpwind&
                   (dir,Vmmm,Vmm,Vm,Vc,Vp,lambdaNN,lambdaN,lambda,leigN,delbarN)
!!$           else
!!$              call hy_uhd_TVDslope&
!!$                   (dir,TVDorder,Vmmm,Vmm,Vm,Vc,Vp,lambdaNN,lambdaN,lambda,leigN,delbarN)
!!$           endif
        else
           call hy_uhd_TVDslope&
                (dir,TVDorder,Vmmm,Vmm,Vm,Vc,Vp,lambdaNN,lambdaN,lambda,leigN,delbarN)
        endif
     endif
  endif



  !! -------------------------------------------------------------------------------------!
  !! (2) Begin high-order polynomial interpolation in normal direction -------------------!
  !! -------------------------------------------------------------------------------------!
  !! Calculate interface values at i+1/2 and i-1/2 using high-order polynomial

  !! First initialize flattening coefficients
  if (hy_flattening) then
     Flattening = FlatCoeff
  else
     Flattening = 0.
  endif


  if (hy_order == 2) then
     Wp(HY_GAMC:HY_GAME)=Vc(HY_GAMC:HY_GAME)&
          +(.5-max(lambda(HY_ENTROPY),0.)*hdtn)*delbar(HY_GAMC:HY_GAME)*(1.-Flattening)

     Wm(HY_GAMC:HY_GAME)=Vc(HY_GAMC:HY_GAME)&
          -(.5+min(lambda(HY_ENTROPY),0.)*hdtn)*delbar(HY_GAMC:HY_GAME)*(1.-Flattening)

#ifdef GRAVITY
!!$     Wp(HY_GRAV)=Vc(HY_GRAV)+(.5-max(lambda(HY_ENTROPY),0.)*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
!!$     Wm(HY_GRAV)=Vc(HY_GRAV)-(.5+min(lambda(HY_ENTROPY),0.)*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
     !! Right states ----------------
     lambdaMax =max(lambda(HY_FASTRGHT),0.)
     temp1 = Vc(HY_GRAV)+(.5-lambdaMax*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
     if (lambda(HY_FASTRGHT) <= 0.) temp1 = 0.

     lambdaMax = max(lambda(HY_FASTLEFT),0.)
     temp2 = Vc(HY_GRAV)+(.5-lambdaMax*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
     if (lambda(HY_FASTLEFT) <= 0.) temp2 = 0.

     if (lambda(HY_FASTLEFT) > 0. .and. lambda(HY_FASTRGHT) > 0.) then
        factor = 0.5
     else
        factor = 1.
     endif
        Wp(HY_GRAV) = factor*(temp2 + temp1)

     !! Left states ----------------
        lambdaMin = -min(lambda(HY_FASTLEFT),0.)
        temp1 = Vc(HY_GRAV)-(.5-lambdaMin*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
        if (lambda(HY_FASTLEFT) >= 0.) temp1 = 0.

        lambdaMin = -min(lambda(HY_FASTRGHT),0.)
        temp1 = Vc(HY_GRAV)-(.5-lambdaMin*hdtn)*delbar(HY_GRAV)*(1.-Flattening)
        if (lambda(HY_FASTRGHT) >= 0.) temp2 = 0.

        if (lambda(HY_FASTLEFT) < 0. .and. lambda(HY_FASTRGHT) < 0.) then
           factor = 0.5
        else
           factor = 1.
        endif
        Wm(HY_GRAV) = factor*(temp1 + temp2)
#endif

     ! Ensure that the interpolated values lie between the cell-centered values
     do n=1,HY_VARINUM+2
        Wm(n) = max(min(Vc(n),Wm(n)),min(max(Vc(n),Wm(n)),Wm(n)))
        Wp(n) = max(min(Vc(n),Wp(n)),min(max(Vc(n),Wp(n)),Wp(n)))
     enddo
#ifdef GRAVITY
     Wm(HY_GRAV) = max(min(Vc(HY_GRAV),Vm(HY_GRAV)),min(max(Vc(HY_GRAV),Vm(HY_GRAV)),Wm(HY_GRAV)))
     Wp(HY_GRAV) = max(min(Vc(HY_GRAV),Vp(HY_GRAV)),min(max(Vc(HY_GRAV),Vp(HY_GRAV)),Wp(HY_GRAV)))
#endif

  elseif (hy_order >= 3) then
     ! delbar contains characteristic variables that are limited.
     ! Project delbar to primitive variables now.
     if (hy_charLimiting) then
        delbar(HY_DENS:HY_MAGZ)  = reig(HY_DENS:HY_MAGZ,HY_FASTLEFT)*delbar(HY_FASTLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNLEFT)*delbar(HY_ALFNLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWLEFT)*delbar(HY_SLOWLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ENTROPY )*delbar(HY_ENTROPY )+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWRGHT)*delbar(HY_SLOWRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNRGHT)*delbar(HY_ALFNRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_FASTRGHT)*delbar(HY_FASTRGHT)

        delbarP(HY_DENS:HY_MAGZ) = reigP(HY_DENS:HY_MAGZ,HY_FASTLEFT)*delbarP(HY_FASTLEFT)+&
                                   reigP(HY_DENS:HY_MAGZ,HY_ALFNLEFT)*delbarP(HY_ALFNLEFT)+&
                                   reigP(HY_DENS:HY_MAGZ,HY_SLOWLEFT)*delbarP(HY_SLOWLEFT)+&
                                   reigP(HY_DENS:HY_MAGZ,HY_ENTROPY )*delbarP(HY_ENTROPY )+&
                                   reigP(HY_DENS:HY_MAGZ,HY_SLOWRGHT)*delbarP(HY_SLOWRGHT)+&
                                   reigP(HY_DENS:HY_MAGZ,HY_ALFNRGHT)*delbarP(HY_ALFNRGHT)+&
                                   reigP(HY_DENS:HY_MAGZ,HY_FASTRGHT)*delbarP(HY_FASTRGHT)

        delbarN(HY_DENS:HY_MAGZ) = reigN(HY_DENS:HY_MAGZ,HY_FASTLEFT)*delbarN(HY_FASTLEFT)+&
                                   reigN(HY_DENS:HY_MAGZ,HY_ALFNLEFT)*delbarN(HY_ALFNLEFT)+&
                                   reigN(HY_DENS:HY_MAGZ,HY_SLOWLEFT)*delbarN(HY_SLOWLEFT)+&
                                   reigN(HY_DENS:HY_MAGZ,HY_ENTROPY )*delbarN(HY_ENTROPY )+&
                                   reigN(HY_DENS:HY_MAGZ,HY_SLOWRGHT)*delbarN(HY_SLOWRGHT)+&
                                   reigN(HY_DENS:HY_MAGZ,HY_ALFNRGHT)*delbarN(HY_ALFNRGHT)+&
                                   reigN(HY_DENS:HY_MAGZ,HY_FASTRGHT)*delbarN(HY_FASTRGHT)
     endif


     if (hy_order == 3) then
        !! -------------------------------------------------------------------------------!
        !! (2)-a: Begining of polynomial interpolation for PPM ---------------------------!
        !! -------------------------------------------------------------------------------!
        !! Parabolic interpolation at the left and right cell interfaces
        !! Colella-Woodward Eqn 1.9, Sekora-Colella Eqn 7, Stone et al Eqn 46
        vecL(HY_DENS:HY_GAME) = 0.5*(Vc0(HY_DENS:HY_GAME)+Vm0(HY_DENS:HY_GAME)) &
             - (delbar(HY_DENS:HY_GAME)-delbarN(HY_DENS:HY_GAME))/6.

        vecR(HY_DENS:HY_GAME) = 0.5*(Vc0(HY_DENS:HY_GAME)+Vp0(HY_DENS:HY_GAME)) &
             - (delbarP(HY_DENS:HY_GAME)-delbar(HY_DENS:HY_GAME))/6.

#ifdef GRAVITY
        vecL(HY_GRAV) = 0.5*(Vc(HY_GRAV)+Vm(HY_GRAV)) - (delbar(HY_GRAV)-delbarN(HY_GRAV))/6.
        vecR(HY_GRAV) = 0.5*(Vc(HY_GRAV)+Vp(HY_GRAV)) - (delbarP(HY_GRAV)-delbar(HY_GRAV))/6.
#endif
        !! End of polynomial interpolation for PPM


     elseif (hy_order == 5) then
        !! -------------------------------------------------------------------------------!
        !! (2)-b: Begining of polynomial interpolation for WENO --------------------------!
        !! -------------------------------------------------------------------------------!
        if (hy_charLimiting) then
           ! Initialize characteristic variables W*
           Wc0  = 0.
           Wp0  = 0.
           Wm0  = 0.
           Wpp0 = 0.
           Wmm0 = 0.
           Wppp0= 0.
           Wmmm0= 0.
           NVAR_DO = HY_WAVENUM
        else
           Wc0  = Vc0
           Wp0  = Vp0
           Wm0  = Vm0
           Wpp0 = Vpp0
           Wmm0 = Vmm0
           Wppp0= Vppp0
           Wmmm0= Vmmm0
           NVAR_DO = HY_VARINUM+2
        endif


        do n=1,NVAR_DO

           if (hy_charLimiting) then
              ! Note: It is important to use eigenvectors of the cell Vc at which the reconstruction of
              !       the two left and right states, Wm and Wp, are considered.
              Wc0(n)   = dot_product(leig(n,HY_DENS:HY_MAGZ), Vc0  (HY_DENS:HY_MAGZ))

              Wp0(n)   = dot_product(leig(n,HY_DENS:HY_MAGZ), Vp0  (HY_DENS:HY_MAGZ))
              Wpp0(n)  = dot_product(leig(n,HY_DENS:HY_MAGZ), Vpp0 (HY_DENS:HY_MAGZ))
              Wppp0(n) = dot_product(leig(n,HY_DENS:HY_MAGZ), Vppp0(HY_DENS:HY_MAGZ))

              Wm0(n)   = dot_product(leig(n,HY_DENS:HY_MAGZ), Vm0  (HY_DENS:HY_MAGZ))
              Wmm0(n)  = dot_product(leig(n,HY_DENS:HY_MAGZ), Vmm0 (HY_DENS:HY_MAGZ))
              Wmmm0(n) = dot_product(leig(n,HY_DENS:HY_MAGZ), Vmmm0(HY_DENS:HY_MAGZ))
           endif

           if (lambda(HY_ENTROPY) > 0.) then

              coeff1(1,1:3) = (/ 2.,  5., -1./)
              coeff1(2,1:3) = (/-1.,  5.,  2./)
              coeff1(3,1:3) = (/ 2., -7., 11./)
              coeff1 = coeff1/6.

              coeff2(1:3)   = (/0.3, 0.6, 0.1/)

              !! Define interpolants
              Intp(1:6) = (/Wmmm0(n),Wmm0(n),Wm0(n),Wc0(n),Wp0(n),Wpp0(n)/)


           else

              coeff1(1,1:3) = (/ 11.,-7.,  2./)
              coeff1(2,1:3) = (/ 2.,  5., -1./)
              coeff1(3,1:3) = (/-1.,  5.,  2./)
              coeff1 = coeff1/6.

              coeff2(1:3)   = (/0.1, 0.6, 0.3/)

              !! Define interpolants
              Intp(1:6) = (/Wmm0(n),Wm0(n),Wc0(n),Wp0(n),Wpp0(n),Wppp0(n)/)

           endif


           !! Calculate interface values at i+1/2
           W5p(1) = dot_product(coeff1(1,1:3),Intp(4:6))
           W5p(2) = dot_product(coeff1(2,1:3),Intp(3:5))
           W5p(3) = dot_product(coeff1(3,1:3),Intp(2:4))

           !! Calculate interface values at i-1/2
           W5m(1) = dot_product(coeff1(1,1:3),Intp(3:5))
           W5m(2) = dot_product(coeff1(2,1:3),Intp(2:4))
           W5m(3) = dot_product(coeff1(3,1:3),Intp(1:3))

           !! Calculate smoothness indicators at i+1/2
           IS5p(1) = 13./12.*(Intp(4)-2.*Intp(5)+Intp(6))**2 + 0.25*(3.*Intp(4)-4.*Intp(5)+Intp(6))**2
           IS5p(2) = 13./12.*(Intp(3)-2.*Intp(4)+Intp(5))**2 + 0.25*(Intp(3)-Intp(5))**2
           IS5p(3) = 13./12.*(Intp(2)-2.*Intp(3)+Intp(4))**2 + 0.25*(Intp(2)-4.*Intp(3)+3.*Intp(4))**2

           !! Calculate smoothness indicators at i-1/2
           IS5m(1) = 13./12.*(Intp(3)-2.*Intp(4)+Intp(5))**2 + 0.25*(3.*Intp(3)-4.*Intp(4)+Intp(5))**2
           IS5m(2) = 13./12.*(Intp(2)-2.*Intp(3)+Intp(4))**2 + 0.25*(Intp(2)-Intp(4))**2
           IS5m(3) = 13./12.*(Intp(1)-2.*Intp(2)+Intp(3))**2 + 0.25*(Intp(1)-4.*Intp(2)+3.*Intp(3))**2

           !! Calculate weights at i+1/2
           Alpha5p(1) = coeff2(1)/(hy_tiny+IS5p(1))!**2
           Alpha5p(2) = coeff2(2)/(hy_tiny+IS5p(2))!**2
           Alpha5p(3) = coeff2(3)/(hy_tiny+IS5p(3))!**2

           !! Normalize weights at i+1/2
           sumAlpha = Alpha5p(1)+Alpha5p(2)+Alpha5p(3)
           omega(1) = Alpha5p(1)/sumAlpha
           omega(2) = Alpha5p(2)/sumAlpha
           omega(3) = Alpha5p(3)/sumAlpha

!!$           ! optimize weights
!!$           Alpha5p(1:3) = (Alpha5p(1:3)*(omega(1:3)+coeff2(1:3))**2&
!!$                            -3.*coeff2(1:3)*Alpha5p(1:3)+Alpha5p(1:3)**2)/&
!!$                            (coeff2(1:3)**2+Alpha5p(1:3)*(1.-2.*coeff2(1:3)))
!!$
!!$           ! normalize again
!!$           sumAlpha=Alpha5p(1)+Alpha5p(2)+Alpha5p(3)
!!$           omega(1)=Alpha5p(1)/sumAlpha
!!$           omega(2)=Alpha5p(2)/sumAlpha
!!$           omega(3)=Alpha5p(3)/sumAlpha

           !! Compute interface value at i+1/2
           vecR(n) = dot_product(omega(1:3), W5p(1:3))

           !! Calculate weights at i-1/2
           Alpha5m(1) = coeff2(1)/(hy_tiny+IS5m(1))!**2
           Alpha5m(2) = coeff2(2)/(hy_tiny+IS5m(2))!**2
           Alpha5m(3) = coeff2(3)/(hy_tiny+IS5m(3))!**2

           !! Normalize weights at i-1/2
           sumAlpha = Alpha5m(1)+Alpha5m(2)+Alpha5m(3)
           omega(1) = Alpha5m(1)/sumAlpha
           omega(2) = Alpha5m(2)/sumAlpha
           omega(3) = Alpha5m(3)/sumAlpha

!!$           ! optimize weights
!!$           Alpha5m(1:3) = (Alpha5m(1:3)*(omega(1:3)+coeff2(1:3))**2&
!!$                            -3.*coeff2(1:3)*Alpha5m(1:3)+Alpha5m(1:3)**2)/&
!!$                            (coeff2(1:3)**2+Alpha5m(1:3)*(1.-2.*coeff2(1:3)))
!!$
!!$           ! normalize again
!!$           sumAlpha=Alpha5m(1)+Alpha5m(2)+Alpha5m(3)
!!$           omega(1)=Alpha5p(1)/sumAlpha
!!$           omega(2)=Alpha5p(2)/sumAlpha
!!$           omega(3)=Alpha5p(3)/sumAlpha

           !! Compute interface value at i+1/2
           vecL(n) = dot_product(omega(1:3), W5m(1:3))
        enddo


        if (hy_charLimiting) then
           vecR(HY_DENS:HY_MAGZ) = reig(HY_DENS:HY_MAGZ,HY_FASTLEFT)*vecR(HY_FASTLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNLEFT)*vecR(HY_ALFNLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWLEFT)*vecR(HY_SLOWLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ENTROPY )*vecR(HY_ENTROPY )+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWRGHT)*vecR(HY_SLOWRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNRGHT)*vecR(HY_ALFNRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_FASTRGHT)*vecR(HY_FASTRGHT)

           vecL(HY_DENS:HY_MAGZ) = reig(HY_DENS:HY_MAGZ,HY_FASTLEFT)*vecL(HY_FASTLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNLEFT)*vecL(HY_ALFNLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWLEFT)*vecL(HY_SLOWLEFT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ENTROPY )*vecL(HY_ENTROPY )+&
                                   reig(HY_DENS:HY_MAGZ,HY_SLOWRGHT)*vecL(HY_SLOWRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_ALFNRGHT)*vecL(HY_ALFNRGHT)+&
                                   reig(HY_DENS:HY_MAGZ,HY_FASTRGHT)*vecL(HY_FASTRGHT)

           ! gamc and game
           vecL(HY_GAMC:HY_GAME) = 0.5*(Vc0(HY_GAMC:HY_GAME)+Vm0(HY_GAMC:HY_GAME)) &
                - (delbar(HY_GAMC:HY_GAME)-delbarN(HY_GAMC:HY_GAME))/6.

           vecR(HY_GAMC:HY_GAME) = 0.5*(Vc0(HY_GAMC:HY_GAME)+Vp0(HY_GAMC:HY_GAME)) &
                - (delbarP(HY_GAMC:HY_GAME)-delbar(HY_GAMC:HY_GAME))/6.

#ifdef GRAVITY
           vecL(HY_GRAV) = 0.5*(Vc(HY_GRAV)+Vm(HY_GRAV)) - (delbar(HY_GRAV)-delbarN(HY_GRAV))/6.
           vecR(HY_GRAV) = 0.5*(Vc(HY_GRAV)+Vp(HY_GRAV)) - (delbarP(HY_GRAV)-delbar(HY_GRAV))/6.
#endif
        endif
        !! End of polynomial interpolation for WENO

     endif


     !! -------------------------------------------------------------------------------!
     !! (2)-c: Contact steepening for PPM or WENO -------------------------------------!
     !! -------------------------------------------------------------------------------!
     temp1 = Vp0(HY_DENS) - Vm0(HY_DENS)
     if (hy_ContactSteepening .and. abs(temp1) > hy_tiny) then
        ! Eqn 1.17 : Second derivatives
        del2rhoR = (Vpp0(HY_DENS)-2.*Vp0(HY_DENS)+ Vc0(HY_DENS))/(6.*delta*delta)
        del2rhoL = ( Vc0(HY_DENS)-2.*Vm0(HY_DENS)+Vmm0(HY_DENS))/(6.*delta*delta)

        ! Third derivative
        eta_steep = (del2rhoL-del2rhoR)*delta**2/temp1
        if (del2rhoR*del2rhoL >= 0.) then
           eta_steep = 0.
        endif
        if (epsln*min(Vp0(HY_DENS),Vm0(HY_DENS))-abs(Vp0(HY_DENS)-Vm0(HY_DENS)) >= 0.) then
           eta_steep = 0.
        endif

        ! Eqn 1.16
        eta_steep = max(0., min(1., eta1*(eta_steep - eta2)))

        ! Eqn 3.2
        temp2 = abs(Vp0(HY_PRES)-Vm0(HY_PRES))/min(Vp0(HY_PRES),Vm0(HY_PRES))
        temp3 = abs(Vp0(HY_DENS)-Vm0(HY_DENS))/min(Vp0(HY_DENS),Vm0(HY_DENS))

        if (Vc(HY_GAME)*K0*temp3-temp2 < 0.0) then
           eta_steep = 0.
        endif

        ! Eqn 1.15
        vecL(HY_DENS) = vecL(HY_DENS)*(1.-eta_steep) + (Vm0(HY_DENS)+0.5*delbarN(HY_DENS))*eta_steep
        vecR(HY_DENS) = vecR(HY_DENS)*(1.-eta_steep) + (Vp0(HY_DENS)-0.5*delbarP(HY_DENS))*eta_steep
     endif

     ! Rotational discontinuity steepening - applying an equivalent contact steepening for magnetic fields
     do n=HY_MAGX,HY_MAGZ
        temp1 = Vp0(n) - Vm0(n)
        if (hy_ContactSteepening .and. abs(temp1) > hy_tiny) then
           ! Eqn 1.17 : Second derivatives
           del2rhoR = (Vpp0(n)-2.*Vp0(n)+ Vc0(n))/(6.*delta*delta)
           del2rhoL = ( Vc0(n)-2.*Vm0(n)+Vmm0(n))/(6.*delta*delta)

           ! Third derivative
           eta_steep = (del2rhoL-del2rhoR)*delta**2/temp1
           if (del2rhoR*del2rhoL >= 0.) then
              eta_steep = 0.
           endif
           if (epsln*min(Vp0(n),Vm0(n))-abs(Vp0(n)-Vm0(n)) >= 0.) then
              eta_steep = 0.
           endif

           ! Eqn 1.16
           eta_steep = max(0., min(1., eta1*(eta_steep - eta2)))

           ! Eqn 3.2 or equivalent
           ! Note: at rotational discontinuity, both density and pressure do not change
           temp2 = abs(Vp0(HY_PRES)-Vm0(HY_PRES))/min(Vp0(HY_PRES),Vm0(HY_PRES))
           temp3 = abs(Vp0(HY_DENS)-Vm0(HY_DENS))/min(Vp0(HY_DENS),Vm0(HY_DENS))
!!$           temp3 = abs(Vp0(n)-Vm0(n))/max(min(Vp0(n),Vm0(n)),1.e-20)

           if (Vc(HY_GAME)*K0*temp3-temp2 < 0.0) then
              eta_steep = 0.
           endif

           ! Eqn 1.15
           vecL(n) = vecL(n)*(1.-eta_steep) + (Vm0(n)+0.5*delbarN(n))*eta_steep
           vecR(n) = vecR(n)*(1.-eta_steep) + (Vp0(n)-0.5*delbarP(n))*eta_steep
        endif
     enddo
     !! End of Contact steepening for PPM or WENO


     !! -------------------------------------------------------------------------------!
     !! (2)-d: Flattening for PPM or WENO ---------------------------------------------!
     !! -------------------------------------------------------------------------------!
     do n=1,HY_VARINUM+2
!!$        ! Flattening
!!$        select case(n)
!!$        case (HY_DENS,HY_MAGX,HY_MAGY,HY_MAGZ)
!!$           Flattening = 0.
!!$        case (HY_PRES)
!!$           Flattening = 1.
!!$        end select
        vecL(n) = Flattening*Vc0(n) + (1.0-Flattening)*vecL(n)
        vecR(n) = Flattening*Vc0(n) + (1.0-Flattening)*vecR(n)
     enddo

     !! -------------------------------------------------------------------------------!
     !! (2)-e: Monotonicity constraint for PPM or WENO --------------------------------!
     !! -------------------------------------------------------------------------------!
     errorCheck = 0.
     do n=1,HY_VARINUM+2
        vecL_temp(n) = checkMedian(vecL(n),Vc0(n),Vm0(n))
        vecR_temp(n) = checkMedian(vecR(n),Vc0(n),Vp0(n))

        vecL_temp1(n) = checkMedian(vecL_temp(n),Vc0(n),3.*Vc0(n)-2.*vecR_temp(n))
        vecR_temp1(n) = checkMedian(vecR_temp(n),Vc0(n),3.*Vc0(n)-2.*vecL_temp(n))

        errorCheck = errorCheck+(vecL(n) - vecL_temp1(n))*(vecR(n)-vecR_temp1(n))**2
     enddo

     if (errorCheck >= 1.E-12) then
        ! if two states get modified then we check two possibilities:
        ! (1) existence of local extremum (i.e., vecL = vecR = Vc)
        errorCheck1 = 0.
        do n=1,HY_VARINUM+2
           ! check if vecL = vecR = Vc
           errorCheck1 = errorCheck1 + abs(vecL(n) + vecR(n) - 2.*Vc0(n))
        enddo
        if (errorCheck1 < 1.E-12) then
           ! we know there is a local extremum in this case and bound WENO values
           vecL_temp2(n) = checkMedian(vecL_temp1(n),Vc0(n),3.*Vc0(n)-2.*vecR_temp(n))
           vecR_temp2(n) = checkMedian(vecR_temp1(n),Vc0(n),3.*Vc0(n)-2.*vecL_temp(n))

           vecL(n) = checkMedian(vecL_temp(n),vecL_temp1(n),vecL_temp2(n))
           vecR(n) = checkMedian(vecR_temp(n),vecR_temp1(n),vecR_temp2(n))
        endif
     endif


     ! Ensure that the interpolated values lie between the cell-centered values
     do n=1,HY_VARINUM+2
        vecL(n) = max(min(Vc0(n),Vm0(n)),min(max(Vc0(n),Vm0(n)),vecL(n)))
        vecR(n) = max(min(Vc0(n),Vp0(n)),min(max(Vc0(n),Vp0(n)),vecR(n)))
     enddo
#ifdef GRAVITY
     vecL(HY_GRAV) = max(min(Vc(HY_GRAV),Vm(HY_GRAV)),min(max(Vc(HY_GRAV),Vm(HY_GRAV)),vecL(HY_GRAV)))
     vecR(HY_GRAV) = max(min(Vc(HY_GRAV),Vp(HY_GRAV)),min(max(Vc(HY_GRAV),Vp(HY_GRAV)),vecR(HY_GRAV)))
#endif


     ! Limit according to Colella-Woodward Eqn 1.10
     do n=1,HY_VARINUM+2
        if ( (vecR(n) - Vc0(n))*(Vc0(n)-vecL(n)) <= 0.) then
           vecL(n) = Vc0(n)
           vecR(n) = Vc0(n)
        endif
        if ( 6.*(vecR(n)-vecL(n))*(Vc0(n)-0.5*(vecL(n)+vecR(n))) > (vecR(n) - vecL(n))**2  ) then
           vecL(n) = 3.*Vc0(n) - 2.*vecR(n)
        endif
        if ( 6.*(vecR(n)-vecL(n))*(Vc0(n)-0.5*(vecL(n)+vecR(n))) < -(vecR(n) - vecL(n))**2  ) then
           vecR(n) = 3.*Vc0(n) - 2.*vecL(n)
        endif
     enddo
#ifdef GRAVITY
     if ( (vecR(HY_GRAV) - Vc(HY_GRAV))*(Vc(HY_GRAV)-vecL(HY_GRAV)) <= 0.) then
        vecL(HY_GRAV) = Vc(HY_GRAV)
        vecR(HY_GRAV) = Vc(HY_GRAV)
     endif
     if ( 6.*(vecR(HY_GRAV)-vecL(HY_GRAV))*(Vc(HY_GRAV)-0.5*(vecL(HY_GRAV)+vecR(HY_GRAV))) &
          > (vecR(HY_GRAV) - vecL(HY_GRAV))**2  ) then
        vecL(HY_GRAV) = 3.*Vc(HY_GRAV) - 2.*vecR(HY_GRAV)
     endif
     if ( 6.*(vecR(HY_GRAV)-vecL(HY_GRAV))*(Vc(HY_GRAV)-0.5*(vecL(HY_GRAV)+vecR(HY_GRAV))) &
          < -(vecR(HY_GRAV) - vecL(HY_GRAV))**2  ) then
        vecR(HY_GRAV) = 3.*Vc(HY_GRAV) - 2.*vecL(HY_GRAV)
     endif
#endif
     !! End of Contact steepeing, Flattening, and Monotonicity constraint for PPM and WENO


     !! -------------------------------------------------------------------------------!
     !! (2)-f: Take initial guesses for the left and right states ---------------------!
     !! -------------------------------------------------------------------------------!
     !! PPM coefficients for parabolic interpolations
     delW(HY_DENS:HY_GAME) = vecR(HY_DENS:HY_GAME)-vecL(HY_DENS:HY_GAME)
     W6(HY_DENS:HY_GAME)   = 6.*(Vc0(HY_DENS:HY_GAME)-0.5*(vecR(HY_DENS:HY_GAME)+vecL(HY_DENS:HY_GAME)))

#ifdef GRAVITY
     delW(HY_GRAV) = vecR(HY_GRAV)-vecL(HY_GRAV)
     W6(HY_GRAV)   = 6.*(Vc(HY_GRAV)-0.5*(vecR(HY_GRAV)+vecL(HY_GRAV)))
#endif

     !! Right states ===================================================================
     lambdaMax =max(lambda(HY_FASTRGHT),0.)
     Wp(HY_DENS:HY_MAGZ) = vecR(HY_DENS:HY_MAGZ) - lambdaMax*hdtn &
          *(delW(HY_DENS:HY_MAGZ) - (1.0 - lambdaMax*hdtn*4./3.)*W6(HY_DENS:HY_MAGZ) )

     !! gamc and game
     lambdaMax = max(lambda(HY_ENTROPY),0.)
     Wp(HY_GAMC:HY_GAME) = vecR(HY_GAMC:HY_GAME) - lambdaMax*hdtn &
          *(delW(HY_GAMC:HY_GAME) - (1.0 - lambdaMax*hdtn*4./3.)*W6(HY_GAMC:HY_GAME) )

     !! gravity component
#ifdef GRAVITY
     lambdaMax =max(lambda(HY_FASTRGHT),0.)
     temp1 = vecR(HY_GRAV) - lambdaMax*hdtn &
          *(delW(HY_GRAV) - (1.0 - lambdaMax*hdtn*4./3.)*W6(HY_GRAV) )
     if (lambda(HY_FASTRGHT) <= 0.) temp1 = 0.

     lambdaMax = max(lambda(HY_FASTLEFT),0.)
     temp2 = vecR(HY_GRAV) - lambdaMax*hdtn &
          *(delW(HY_GRAV) - (1.0 - lambdaMax*hdtn*4./3.)*W6(HY_GRAV) )
     if (lambda(HY_FASTLEFT) <= 0.) temp2 = 0.

     if (lambda(HY_FASTLEFT) > 0. .and. lambda(HY_FASTRGHT) > 0.) then
        factor = 0.5
     else
        factor = 1.
     endif
        Wp(HY_GRAV) = factor*(temp2 + temp1)        
#endif

     !! Left states  ===================================================================
     lambdaMin = -min(lambda(HY_FASTLEFT),0.)
     Wm(HY_DENS:HY_MAGZ) = vecL(HY_DENS:HY_MAGZ) + lambdaMin*hdtn &
          *(delW(HY_DENS:HY_MAGZ) + (1.0 - lambdaMin*hdtn*4./3.)*W6(HY_DENS:HY_MAGZ) )

     !! gamc and game
     lambdaMin = -min(lambda(HY_ENTROPY),0.)
     Wm(HY_GAMC:HY_GAME) = vecL(HY_GAMC:HY_GAME) + lambdaMin*hdtn &
          *(delW(HY_GAMC:HY_GAME) + (1.0 - lambdaMin*hdtn*4./3.)*W6(HY_GAMC:HY_GAME) )

     !! gravity component
#ifdef GRAVITY
     lambdaMin = -min(lambda(HY_FASTLEFT),0.)
     temp1 = vecL(HY_GRAV) + lambdaMin*hdtn &
          *(delW(HY_GRAV) + (1.0 - lambdaMin*hdtn*4./3.)*W6(HY_GRAV) )
     if (lambda(HY_FASTLEFT) >= 0.) temp1 = 0.

     lambdaMin = -min(lambda(HY_FASTRGHT),0.)
     temp2 = vecL(HY_GRAV) + lambdaMin*hdtn &
          *(delW(HY_GRAV) + (1.0 - lambdaMin*hdtn*4./3.)*W6(HY_GRAV) )
     if (lambda(HY_FASTRGHT) >= 0.) temp2 = 0.

     if (lambda(HY_FASTLEFT) < 0. .and. lambda(HY_FASTRGHT) < 0.) then
        factor = 0.5
     else
        factor = 1.
     endif
        Wm(HY_GRAV) = factor*(temp1 + temp2)
#endif

     !! End of initial guesses
  endif
  !! End of high-order polynomial interpolations (MUSCL-Hancock, PPM or WENO) for interface values


  !! -------------------------------------------------------------------------------!
  !! (3) Advance the above interpolated interface values by 1/2 time step using     !
  !!     characteristic tracing method for PPM or WENO                              !
  !! -------------------------------------------------------------------------------!

  do n=1,HY_WAVENUM
     if (lambda(n) < 0.) then
        select case (hy_transOrder)
        case(1)
           !! First order upwind for transverse flux derivative works fine but diffusive
           TransFlux(HY_DENS:hyEnd)=Vp(HY_DENS:hyEnd)-Vc(HY_DENS:hyEnd)
        case(2)
           !! Note: This second order upwind is not recommended for use
           TransFlux(HY_DENS:hyEnd)=0.5*(  -Vpp(HY_DENS:hyEnd)&
                                         +4.*Vp(HY_DENS:hyEnd)&
                                         -3.*Vc(HY_DENS:hyEnd))
        case(3)
           !! Third order upwind is default and rommended for use in most applications
           TransFlux(HY_DENS:hyEnd)=(  -Vpp(HY_DENS:hyEnd)&
                                     +6.*Vp(HY_DENS:hyEnd)&
                                     -3.*Vc(HY_DENS:hyEnd)&
                                     -2.*Vm(HY_DENS:hyEnd))/6.
        end select
    else
       select case (hy_transOrder)
       case(1)
          TransFlux(HY_DENS:hyEnd)=Vc(HY_DENS:hyEnd)-Vm(HY_DENS:hyEnd)
       case(2)
          TransFlux(HY_DENS:hyEnd)=0.5*( 3.*Vc(HY_DENS:hyEnd)&
                                        -4.*Vm(HY_DENS:hyEnd)&
                                          +Vmm(HY_DENS:hyEnd))
        case(3)
           TransFlux(HY_DENS:hyEnd)=( 2.*Vp(HY_DENS:hyEnd)&
                                     +3.*Vc(HY_DENS:hyEnd)&
                                     -6.*Vm(HY_DENS:hyEnd)&
                                       +Vmm(HY_DENS:hyEnd))/6.
        end select
     endif


     !! Apply slope limiter on characteristic variables
     if (hy_charLimiting) then
        if (hy_RiemannSolver == ROE) then
           if (lambda(n) < 0.) then
              if (hy_order == 2) then
                 !! Apply monotone slope limiting for normal flux
                 vecL(HY_DENS:HY_MAGZ) = .5*(-1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)*delbar(n)
                 sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)
              elseif (hy_order >= 3) then
                 ! PPM step 10
                 constC = hdtn*(lambda(HY_FASTLEFT) - lambda(n))
                 constD = 1./3. *(dtn)**2 * (lambda(HY_FASTLEFT)**2 - lambda(n)**2)
                 vecL(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                      (constC*(delW(HY_DENS:HY_MAGZ)+W6(HY_DENS:HY_MAGZ))+constD*W6(HY_DENS:HY_MAGZ)))&
                      *reig(HY_DENS:HY_MAGZ,n)
                 sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)
              endif

           elseif (lambda(n) > 0.) then
              if (hy_order == 2) then
                 !! Apply monotone slope limiting for normal flux term
                 vecR(HY_DENS:HY_MAGZ) = .5*(1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)*delbar(n)
                 sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
              elseif (hy_order >= 3) then
                 constA = hdtn*(lambda(HY_FASTRGHT) - lambda(n))
                 constB = 1./3. *(dtn)**2 * (lambda(HY_FASTRGHT)**2 - lambda(n)**2 )
                 vecR(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                      (constA*(delW(HY_DENS:HY_MAGZ)-W6(HY_DENS:HY_MAGZ))+constB*W6(HY_DENS:HY_MAGZ)))&
                      *reig(HY_DENS:HY_MAGZ,n)
                 sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
              endif

           endif
        else
           !! Left states for HLL* type solvers
           !! For more detail, see "Athena: A new code for astrophysical MHD"
           !! by Stone, Gardiner, Teuben, Hawley, Simon, arXiv:0804.0402v1 [astro-ph] 2 Apr 2008
           !! Apply monotone slope limiting for normal flux
           if (hy_order == 2) then
              vecL(HY_DENS:HY_MAGZ) = .5*(-1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)*delbar(n)
              sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)

              !! Right states for HLL* type solvers
              !! Apply monotone slope limiting for normal flux term
              vecR(HY_DENS:HY_MAGZ) = .5*(1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)*delbar(n)
              sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
           elseif (hy_order >= 3) then
              ! PPM step 10
              constC = hdtn*(lambda(HY_FASTLEFT) - lambda(n))
              constD = 1./3. *(dtn)**2 * (lambda(HY_FASTLEFT)**2 - lambda(n)**2)
              vecL(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                   (constC*(delW(HY_DENS:HY_MAGZ)+W6(HY_DENS:HY_MAGZ))+constD*W6(HY_DENS:HY_MAGZ)))&
                   *reig(HY_DENS:HY_MAGZ,n)
              sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)

              constA = hdtn*(lambda(HY_FASTRGHT) - lambda(n))
              constB = 1./3. *(dtn)**2 * (lambda(HY_FASTRGHT)**2 - lambda(n)**2 )
              vecR(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                   (constA*(delW(HY_DENS:HY_MAGZ)-W6(HY_DENS:HY_MAGZ))+constB*W6(HY_DENS:HY_MAGZ)))&
                   *reig(HY_DENS:HY_MAGZ,n)
              sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
           endif
        endif


        !! Transverse flux -------------------------------------------
        !! Apply upwinding differencing for transverse flux
        vec(HY_DENS:HY_MAGZ)  = lambda(n)*reig(HY_DENS:HY_MAGZ,n)&
             *dot_product(leig(n,HY_DENS:HY_MAGZ),TransFlux(HY_DENS:HY_MAGZ))
        if (hy_fullyLimit) then
           !! Total wave contributions for transverse flux
           vec(HY_DENS:HY_MAGZ)  = lambda(n)*reig(HY_DENS:HY_MAGZ,n)*delbar(n)
        endif
        sig(HY_DENS:HY_MAGZ) = sig(HY_DENS:HY_MAGZ) + vec(HY_DENS:HY_MAGZ)

     else
        !! Apply slope limiter on primitive variables
        if (hy_RiemannSolver == ROE) then
           if (lambda(n) < 0.) then
              if (hy_order == 2) then
                 !! Apply monotone slope limiting for normal flux
                 vecL(HY_DENS:HY_MAGZ) = .5*(-1.- dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)&
                      *dot_product(leig(n,HY_DENS:HY_MAGZ), delbar(HY_DENS:HY_MAGZ))
                 sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)
              elseif (hy_order >= 3) then
                 ! PPM
                 ! some coefficients
                 constC = hdtn*(lambda(HY_FASTLEFT) - lambda(n))
                 constD = 1./3. *(dtn)**2 * (lambda(HY_FASTLEFT)**2 - lambda(n)**2)
                 vecL(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                      (constC*(delW(HY_DENS:HY_MAGZ)+W6(HY_DENS:HY_MAGZ))+constD*W6(HY_DENS:HY_MAGZ)))&
                      *reig(HY_DENS:HY_MAGZ,n)
                 sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)
              endif

           elseif (lambda(n) > 0.) then
              if (hy_order == 2) then
                 !! Apply monotone slope limiting for normal flux
                 vecR(HY_DENS:HY_MAGZ) = .5*(1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)&
                      *dot_product(leig(n,HY_DENS:HY_MAGZ), delbar(HY_DENS:HY_MAGZ))
                 sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
              elseif (hy_order >= 3) then
                 ! PPM
                 ! some coefficients
                 constA = hdtn*(lambda(HY_FASTRGHT) - lambda(n))
                 constB = 1./3. *(dtn)**2 * (lambda(HY_FASTRGHT)**2 - lambda(n)**2 )
                 vecR(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                      (constA*(delW(HY_DENS:HY_MAGZ)-W6(HY_DENS:HY_MAGZ))+constB*W6(HY_DENS:HY_MAGZ)))&
                      *reig(HY_DENS:HY_MAGZ,n)
                 sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
              endif

           endif
        else
           !! Left states for HLL* type solvers
           !! Apply monotone slope limiting for normal flux
           if (hy_order == 2) then
              vecL(HY_DENS:HY_MAGZ) = .5*(-1.- dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)&
                   *dot_product(leig(n,HY_DENS:HY_MAGZ), delbar(HY_DENS:HY_MAGZ))
              sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)

              !! Right states for HLL* type solvers
              !! Apply monotone slope limiting for normal flux
              vecR(HY_DENS:HY_MAGZ) = .5*(1.-dtn*lambda(n))*reig(HY_DENS:HY_MAGZ,n)&
                   *dot_product(leig(n,HY_DENS:HY_MAGZ), delbar(HY_DENS:HY_MAGZ))
              sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)

           elseif (hy_order >= 3) then
              ! PPM step 10
              constC = hdtn*(lambda(HY_FASTLEFT) - lambda(n))
              constD = 1./3. *(dtn)**2 * (lambda(HY_FASTLEFT)**2 - lambda(n)**2)
              vecL(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                   (constC*(delW(HY_DENS:HY_MAGZ)+W6(HY_DENS:HY_MAGZ))+constD*W6(HY_DENS:HY_MAGZ)))&
                   *reig(HY_DENS:HY_MAGZ,n)
              sigL(HY_DENS:HY_MAGZ) = sigL(HY_DENS:HY_MAGZ) + vecL(HY_DENS:HY_MAGZ)

              constA = hdtn*(lambda(HY_FASTRGHT) - lambda(n))
              constB = 1./3. *(dtn)**2 * (lambda(HY_FASTRGHT)**2 - lambda(n)**2 )
              vecR(HY_DENS:HY_MAGZ) = dot_product(leig(n,HY_DENS:HY_MAGZ),&
                   (constA*(delW(HY_DENS:HY_MAGZ)-W6(HY_DENS:HY_MAGZ))+constB*W6(HY_DENS:HY_MAGZ)))&
                   *reig(HY_DENS:HY_MAGZ,n)
              sigR(HY_DENS:HY_MAGZ) = sigR(HY_DENS:HY_MAGZ) + vecR(HY_DENS:HY_MAGZ)
           endif

        endif

        !! Transverse flux -------------------------------------------
        !! Apply upwinding differencing for transverse flux
        vec(HY_DENS:HY_MAGZ)  = lambda(n)*reig(HY_DENS:HY_MAGZ,n)&
             *dot_product(leig(n,HY_DENS:HY_MAGZ),TransFlux(HY_DENS:HY_MAGZ))
        if (hy_fullyLimit) then
           !! Total wave contributions for transverse flux
           vec(HY_DENS:HY_MAGZ)  = lambda(n)*reig(HY_DENS:HY_MAGZ,n)&
                     *dot_product(leig(n,HY_DENS:HY_MAGZ),delbar(HY_DENS:HY_MAGZ))
        endif
        sig(HY_DENS:HY_MAGZ) = sig(HY_DENS:HY_MAGZ) + vec(HY_DENS:HY_MAGZ)

     endif
  enddo


  !! Transverse flux for gamc, game & gravity
  sig(HY_GAMC:HY_GAME) = lambda(HY_ENTROPY)*TransFlux(HY_GAMC:HY_GAME)


  !! Riemann states in normal direction
  if (hy_order == 2) then
     Wm(HY_DENS:HY_MAGZ) = Vc0(HY_DENS:HY_MAGZ)+sigL(HY_DENS:HY_MAGZ)-hdtn*aBn(HY_DENS:HY_MAGZ)*dnBn
     Wp(HY_DENS:HY_MAGZ) = Vc0(HY_DENS:HY_MAGZ)+sigR(HY_DENS:HY_MAGZ)-hdtn*aBn(HY_DENS:HY_MAGZ)*dnBn
  elseif (hy_order >= 3) then
     Wm(HY_DENS:HY_MAGZ) = Wm(HY_DENS:HY_MAGZ)+sigL(HY_DENS:HY_MAGZ)-hdtn*aBn(HY_DENS:HY_MAGZ)*dnBn
     Wp(HY_DENS:HY_MAGZ) = Wp(HY_DENS:HY_MAGZ)+sigR(HY_DENS:HY_MAGZ)-hdtn*aBn(HY_DENS:HY_MAGZ)*dnBn
  endif
  !! End of advancing the interpolated interface values by 1/2 time step for PPM or WENO

end Subroutine DataReconstructNormalDir
