!!****if* source/physics/Hydro/HydroMain/unsplit/hy_uhd_getTransportCoefficients.F90
!!
!! NAME
!!
!!  hy_uhd_getTransportCoefficients
!!
!!
!! SYNOPSIS
!!
!!  hy_uhd_getTransportCoefficients(temp,dens,magp,massFrac,Abar,Zbar,TransCoeff)
!!
!! DESCRIPTION
!!
!!  Calculates transport coefficients
!!
!!
!! ARGUMENTS
!!
!! temp       - local temperature
!! den        - local density
!! magp       - local magnetic pressure
!! Abar       - average species mass
!! Zbar       - average species charge
!! massFrac   - species composition
!! TransCoeff - array for transport coefficients
!!

subroutine getTransportCoefficients(temp,dens,magp,massFrac,Abar,Zbar,TransCoeff)

  use Hydro_data, ONLY : hy_proton_mass, hy_pe_mass_ratio, &
                         hy_CoulombLog_const1, hy_CoulombLog_const2, &
                         hy_Pi_units, hy_K_units, hy_eta_units, &
                         hy_Pi_coeff, hy_K_coeff, hy_eta_coeff, hy_delta_coeff, &
                         hy_use_Braginskii, hy_Braginskii_Z_dependence, &
                         hy_viscosity_0, hy_viscosity_1, hy_viscosity_2, &
                         hy_viscosity_3, hy_viscosity_4, &
                         hy_conductivity_para,  hy_resistivity_para, &
                         hy_conductivity_perp,  hy_resistivity_perp, &
                         hy_conductivity_caret, hy_resistivity_Hall, &
                         hy_max_viscosity, hy_max_conductivity, hy_max_resistivity, &
                         hy_NA, hy_kB

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! Argument List --------------------------------------------------------
  real, intent(IN) :: temp, dens, magp, Abar, Zbar
  real, dimension(NSPECIES), intent(IN) :: massFrac
  real, dimension(HY_TRANSCOEFF_NUM),intent(OUT) :: TransCoeff
  !! ----------------------------------------------------------------------

  real :: n_i !ion density
  real :: n_e !electron density
  real :: delta_e !dimensionless electron skin depth
  real :: delta_i !dimensionless ion skin depth
  real :: CoulombLog
  real :: K_para_Zdepend, K_perp_Zdepend, eta_para_Zdepend !Braginskii coefficients
  logical :: is_isotropic
  real :: max_K

  !Use prescribed transport parameters
  if( .not. hy_use_Braginskii ) then
     TransCoeff(HY_PI0)     = hy_viscosity_0
     TransCoeff(HY_PI1)     = hy_viscosity_1
     TransCoeff(HY_PI2)     = hy_viscosity_2
     TransCoeff(HY_PI3)     = hy_viscosity_3
     TransCoeff(HY_PI4)     = hy_viscosity_4
     TransCoeff(HY_K_PARA)  = hy_conductivity_para
     TransCoeff(HY_K_PERP)  = hy_conductivity_perp
     TransCoeff(HY_K_CARET) = hy_conductivity_caret
     TransCoeff(HY_ETA_PARA)= hy_resistivity_para
     TransCoeff(HY_ETA_PERP)= hy_resistivity_perp
     TransCoeff(HY_ETA_HALL)= hy_resistivity_Hall

  !apply factor relating conductivity to thermal diffusivity using heat capacity for ideal gas
  !D_K = K / (rho*Cv)
     max_K = 0.
     if( TransCoeff(HY_K_PARA)  > max_K ) max_K = TransCoeff(HY_K_PARA)
     if( TransCoeff(HY_K_PERP)  > max_K ) max_K = TransCoeff(HY_K_PERP)
     if( TransCoeff(HY_K_CARET) > max_K ) max_K = TransCoeff(HY_K_CARET)
     max_K = max_K/( dens*2.5*hy_NA*hy_kB )
     if( max_K > hy_max_conductivity ) hy_max_conductivity = max_K

     return
  endif

  !Determine local electron density
  n_i = dens / (Abar*hy_proton_mass)
  n_e = Zbar*n_i

  !Calculate Coulomb logarithm
  CoulombLog = min(hy_CoulombLog_const1+1.5*log(temp), hy_CoulombLog_const2+log(temp)) - 0.5*log(n_e)

  !!if (myPE == MASTER_PE) then
     !print *, "Coords=", i,j,k
     !print *, "abar=", abar, "zbar=", zbar
     !!WRITE(99,'(4(G14.6))',ADVANCE='NO') temp, dens, magp, CoulombLog
     !!print *, "temp=", temp, "dens=", dens, "magp=", magp, "CoulombLog=", CoulombLog
  !!endif


  !If Coulomb logarithm is positive, then this is probably a plasma.
  if( CoulombLog > 0.1 ) then
     
     !Determine if plasma is sufficiently anisotropic to make strong field approximations of Braginskii's eqns valid
     is_isotropic = .false.
     if (magp == 0.) then
        is_isotropic = .true.
     else
        !electron skin depth
        delta_e = hy_delta_coeff*sqrt(magp*temp**3/(CoulombLog*n_e**2))
        !ion skin depth
        delta_i = delta_e*sqrt(2.*hy_pe_mass_ratio/Abar)*Zbar
        if( delta_i < 1. ) then
           is_isotropic = .true.
        endif
     endif

     !Determine Z-dependent Braginskii coefficients
     if( hy_Braginskii_Z_dependence ) then
        K_para_Zdepend   = 12.47  - 13.92 *exp(-0.4023*Zbar**0.5967 )
        K_perp_Zdepend   = 3.250  + 26080 *exp(-9.825 *Zbar**0.09648)
        eta_para_Zdepend = 0.2949 + 0.5285*exp(-0.8845*Zbar**0.5525 )
     else !Otherwise, assume Z=1
        K_para_Zdepend   = 3.1616
        K_perp_Zdepend   = 4.664
        eta_para_Zdepend = 0.5129
     end if
     
     if (.not. is_isotropic) then !plasma is anisotropic
        
        !Transport Processes in a Plasma: S.I. Braginskii, Rev of Plasma Phys 1, 205 (1965)

        TransCoeff(HY_PI0)     = hy_Pi_coeff*temp**2.5/CoulombLog
        TransCoeff(HY_PI1)     = TransCoeff(HY_PI0)*0.3/delta_i**2
        TransCoeff(HY_PI2)     = TransCoeff(HY_PI1)*4.
        TransCoeff(HY_PI3)     = TransCoeff(HY_PI2)*(5./12.)*delta_i
        TransCoeff(HY_PI4)     = TransCoeff(HY_PI3)*2.
        TransCoeff(HY_PI0)     = 0.96*TransCoeff(HY_PI0)

        TransCoeff(HY_K_PARA)  = hy_K_coeff*temp**2.5/CoulombLog
        TransCoeff(HY_K_PERP)  = TransCoeff(HY_K_PARA)*K_perp_Zdepend/delta_e**2
        TransCoeff(HY_K_CARET) = TransCoeff(HY_K_PARA)*2.5/delta_e
        TransCoeff(HY_K_PARA)  = TransCoeff(HY_K_PARA)*K_para_Zdepend

        TransCoeff(HY_ETA_PERP)= hy_eta_coeff*Zbar*CoulombLog*temp**(-1.5)
        TransCoeff(HY_ETA_PARA)= TransCoeff(HY_ETA_PERP)*eta_para_Zdepend
        TransCoeff(HY_ETA_HALL)= TransCoeff(HY_ETA_PARA)*delta_e
        
     else !is_isotropic == .true., plasma is isotropic
        
        TransCoeff(HY_PI0)     = hy_Pi_coeff*0.96*temp**2.5/CoulombLog
        TransCoeff(HY_PI1)     = TransCoeff(HY_PI0)
        TransCoeff(HY_PI2)     = TransCoeff(HY_PI0)
        TransCoeff(HY_PI3)     = 0.
        TransCoeff(HY_PI4)     = 0.

        TransCoeff(HY_K_PARA)  = hy_K_coeff*K_para_Zdepend*temp**2.5/(CoulombLog)
        TransCoeff(HY_K_PERP)  = TransCoeff(HY_K_PARA)
        TransCoeff(HY_K_CARET) = 0.

        TransCoeff(HY_ETA_PERP)= hy_eta_coeff*Zbar*CoulombLog*temp**(-1.5)
        TransCoeff(HY_ETA_PARA)= TransCoeff(HY_ETA_PERP)
        TransCoeff(HY_ETA_HALL)= TransCoeff(HY_ETA_PARA)*delta_e
        
     end if !is_isotropic

  else !CoulombLog <= 0. -> This is probably a metal, give it appropriate transport parameters.

     !Properties for PbLi: B.Shulz, Fusion Eng. Design 14, 199 (1991).

     TransCoeff(HY_PI0)     = hy_Pi_units*1.87e-4*exp(1400./temp)
     TransCoeff(HY_PI1)     = TransCoeff(HY_PI0)
     TransCoeff(HY_PI2)     = TransCoeff(HY_PI0)
     TransCoeff(HY_PI3)     = 0.
     TransCoeff(HY_PI4)     = 0.

     TransCoeff(HY_K_PARA)  = hy_K_units*(1.95 + 0.0195*temp)
     TransCoeff(HY_K_PERP)  = TransCoeff(HY_K_PARA)
     TransCoeff(HY_K_CARET) = 0.

     TransCoeff(HY_ETA_PARA)= hy_eta_units*(1.023e-8 + 4.26e-12*temp)
     TransCoeff(HY_ETA_PERP)= TransCoeff(HY_ETA_PARA)
     TransCoeff(HY_ETA_HALL)= 0.

     !TODO REMOVE
     delta_e = 0.
     delta_i = 0.
     is_isotropic = .true.

  end if !CoulombLog

  !!if (myPE == MASTER_PE) then
     !!WRITE(99,'(13(G14.6))') delta_e, delta_i, TransCoeff
     !!print *, "delta_e=", delta_e, "delta_i=", delta_i, "is_isotropic=", is_isotropic, "TransCoeff=", TransCoeff
  !!endif

  !apply transport caps
  if( hy_viscosity_0 >= 0. .and. TransCoeff(HY_PI0) > hy_viscosity_0 ) then
     TransCoeff(HY_PI0) = hy_viscosity_0
  endif
  if( hy_viscosity_1 >= 0. .and. TransCoeff(HY_PI1) > hy_viscosity_1 ) then
     TransCoeff(HY_PI1) = hy_viscosity_1
  endif
  if( hy_viscosity_2 >= 0. .and. TransCoeff(HY_PI2) > hy_viscosity_2 ) then
     TransCoeff(HY_PI2) = hy_viscosity_2
  endif
  if( hy_viscosity_3 >= 0. .and. TransCoeff(HY_PI3) > hy_viscosity_3 ) then
     TransCoeff(HY_PI3) = hy_viscosity_3
  endif
  if( hy_viscosity_4 >= 0. .and. TransCoeff(HY_PI4) > hy_viscosity_4 ) then
     TransCoeff(HY_PI4) = hy_viscosity_4
  endif
  if( hy_conductivity_para >= 0. .and. TransCoeff(HY_K_PARA) > hy_conductivity_para ) then
     TransCoeff(HY_K_PARA) = hy_conductivity_para
  endif
  if( hy_conductivity_perp >= 0. .and. TransCoeff(HY_K_PERP) > hy_conductivity_perp ) then
     TransCoeff(HY_K_PERP) = hy_conductivity_perp
  endif
  if( hy_conductivity_caret>= 0. .and. TransCoeff(HY_K_CARET)> hy_conductivity_caret) then
     TransCoeff(HY_K_CARET) = hy_conductivity_caret
  endif
  if( hy_resistivity_para  >= 0. .and. TransCoeff(HY_ETA_PERP) > hy_resistivity_para ) then
     TransCoeff(HY_ETA_PERP) = hy_resistivity_para
  endif
  if( hy_resistivity_perp  >= 0. .and. TransCoeff(HY_ETA_PARA) > hy_resistivity_perp ) then
     TransCoeff(HY_ETA_PARA) = hy_resistivity_perp
  endif
  if( hy_resistivity_Hall  >= 0. .and. TransCoeff(HY_ETA_HALL) > hy_resistivity_Hall ) then
     TransCoeff(HY_ETA_HALL) = hy_resistivity_Hall
  endif

  !check for maximum transport
  if( TransCoeff(HY_PI0) > hy_max_viscosity ) hy_max_viscosity = TransCoeff(HY_PI0)
  if( TransCoeff(HY_PI1) > hy_max_viscosity ) hy_max_viscosity = TransCoeff(HY_PI1)
  if( TransCoeff(HY_PI2) > hy_max_viscosity ) hy_max_viscosity = TransCoeff(HY_PI2)
  if( TransCoeff(HY_PI3) > hy_max_viscosity ) hy_max_viscosity = TransCoeff(HY_PI3)
  if( TransCoeff(HY_PI4) > hy_max_viscosity ) hy_max_viscosity = TransCoeff(HY_PI4)
  if( TransCoeff(HY_ETA_PARA) > hy_max_resistivity  ) hy_max_resistivity  = TransCoeff(HY_ETA_PARA)
  if( TransCoeff(HY_ETA_PERP) > hy_max_resistivity  ) hy_max_resistivity  = TransCoeff(HY_ETA_PERP)
  if( TransCoeff(HY_ETA_HALL) > hy_max_resistivity  ) hy_max_resistivity  = TransCoeff(HY_ETA_HALL)

  !apply factor relating conductivity to thermal diffusivity using heat capacity for ideal gas
  !D_K = K / (rho*Cv)
  max_K = 0.
  if( TransCoeff(HY_K_PARA)  > max_K ) max_K = TransCoeff(HY_K_PARA)
  if( TransCoeff(HY_K_PERP)  > max_K ) max_K = TransCoeff(HY_K_PERP)
  if( TransCoeff(HY_K_CARET) > max_K ) max_K = TransCoeff(HY_K_CARET)
  max_K = max_K/( dens*2.5*hy_NA*hy_kB )
  if( max_K > hy_max_conductivity ) hy_max_conductivity = max_K

  return
end subroutine getTransportCoefficients
