!!****if* source/physics/Hydro/HydroMain/unsplit/Anisotropic_MHD/hy_uhd_shockDetect
!!
!! NAME
!!
!!  hy_uhd_shockDetect
!!
!! SYNOPSIS
!!
!!  hy_uhd_shockDetect( integer (IN) :: blockID )
!!
!! DESCRIPTION
!!
!!  This routine detects strongly compressive motions in simulation
!!  by calculating undivided pressure gradients and divergence of
!!  velocity fields. Two parameters beta and delta have been set 
!!  to detect very strong shocks. If such shocks exit then the unsplit
!!  scheme applies its robust flux differencings using limited slopes
!!  in data reconstruction step (see hy_uhd_dataReconstruct.F90).
!!  Different shock strengths can also be detected by lowering/increasing
!!  beta and delta values.
!!
!! ARGUMENTS
!!
!!  blockID  - local block ID
!!
!! REFERENCE 
!!
!!  Balsara and Spicer, JCP, 149:270--292, 1999.
!!
!!***

!!REORDER(4): U, scrch_Ctr

!#define DEBUG_HY_SHOCK_DETECT

Subroutine hy_uhd_shockDetect( blockID )


  use Grid_interface,    ONLY : Grid_getBlkIndexLimits, &
                                Grid_getBlkPtr,         &
                                Grid_releaseBlkPtr,     &
                                Grid_getDeltas
  use Hydro_data,        ONLY : hy_cfl, hy_cfl_original,&
                                hy_fullyLimit, hy_myPE, &
                                hy_order_original,      &
                                hy_order
  use Driver_data,       ONLY : dr_nStep
  use Logfile_interface, ONLY : Logfile_open,Logfile_close

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"

  !! ---- Argument List ----------------------------------
  integer, INTENT(IN) :: blockID
  !! -----------------------------------------------------


  integer :: i,j,k
  integer, dimension(LOW:HIGH,MDIM) :: blkLimits,blkLimitsGC

  real :: divv,gradPx,gradPy,gradPz
  real, pointer, dimension(:,:,:,:) :: U
  logical :: SW1, SW2
  real :: minP,minC,beta,delta
  integer :: logUnit
  logical :: logUnitLocal=.true.
  real, save, dimension(6) :: globalMinMaxs
  real, dimension(MDIM) :: del


  ! Initialize two switches
  SW1 = .false.
  SW2 = .false.

  ! Two parameters that can be adjusted to detect shocks
  ! with different strengths:
  ! (a) The lower the values the weaker shocks detected and
  ! (b) The larger the values the stronger shocks detected.
  beta = 0.5 !10.
  delta= 0.1 !2.

  ! Initialize global maxs
  globalMinMaxs(:) = tiny(1.)
  globalMinMaxs(2) = huge(1.)
  call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)
  call Grid_getBlkPtr(blockID,U,CENTER)
  call Grid_getDeltas(blockID,del)

  do k=blkLimits(LOW,KAXIS),blkLimits(HIGH,KAXIS)
     do j=blkLimits(LOW,JAXIS),blkLimits(HIGH,JAXIS)
        do i=blkLimits(LOW,IAXIS),blkLimits(HIGH,IAXIS)

#if NDIM == 1
           minP = min(U(PRES_VAR,i,  j,  k  ),&
                      U(PRES_VAR,i+1,j,  k  ),&
                      U(PRES_VAR,i-1,j,  k  ))

           minC = min(sqrt((U(GAMC_VAR,i,  j,  k  )  &
                           *U(PRES_VAR,i,  j,  k  )  &
                        +2.*U(MAGP_VAR,i,  j,  k  )) &
                           /U(DENS_VAR,i,  j,  k  )),&
                      sqrt((U(GAMC_VAR,i-1,j,  k  )  &
                           *U(PRES_VAR,i-1,j,  k  )  &
                        +2.*U(MAGP_VAR,i-1,j,  k  )) &
                           /U(DENS_VAR,i-1,j,  k  )),&
                      sqrt((U(GAMC_VAR,i+1,j,  k  )  &
                           *U(PRES_VAR,i+1,j,  k  )  &
                        +2.*U(MAGP_VAR,i+1,j,  k  )) &
                           /U(DENS_VAR,i+1,j,  k  )))

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  ))

           gradPx = 0.5*(U(PRES_VAR,i+1,j,  k  ) - U(PRES_VAR,i-1,j,  k  ))
           gradPy = 0.
           gradPz = 0.

           if ( abs(gradPx) > beta*minP ) then
              SW1 = .true.
           endif

           if (-delta*minC > divv) then
              SW2 = .true.
           endif

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  )/del(DIR_X))

           if (SW1 .and. SW2) then
              if (hy_cfl > 0.6) hy_cfl = 0.6
              hy_fullyLimit = .true.
              if (abs(minC)   > globalMinMaxs(1)) globalMinMaxs(1) = minC
              if ((divv < 0.) .and. (abs(divv) < globalMinMaxs(2))) globalMinMaxs(2) = divv !compression
              if ((divv > 0.) .and. (abs(divv) > globalMinMaxs(3))) globalMinMaxs(3) = divv !rarefaction
              if (abs(gradPx) > globalMinMaxs(4)) globalMinMaxs(4) = gradPx/del(DIR_X)

           else
              hy_cfl = hy_cfl_original
              hy_fullyLimit = .false.
           endif



#elif NDIM == 2
           minP = min(U(PRES_VAR,i,  j,  k  ),&
                      U(PRES_VAR,i-1,j,  k  ),&
                      U(PRES_VAR,i+1,j,  k  ),&
                      U(PRES_VAR,i,  j-1,k  ),&
                      U(PRES_VAR,i,  j+1,k  ))

           minC = min(sqrt((U(GAMC_VAR,i,  j,  k  )  &
                           *U(PRES_VAR,i,  j,  k  )  &
                        +2.*U(MAGP_VAR,i,  j,  k  )) &
                           /U(DENS_VAR,i,  j,  k  )),&
                      sqrt((U(GAMC_VAR,i-1,j,  k  )  &
                           *U(PRES_VAR,i-1,j,  k  )  &
                        +2.*U(MAGP_VAR,i-1,j,  k  )) &
                           /U(DENS_VAR,i-1,j,  k  )),&
                      sqrt((U(GAMC_VAR,i+1,j,  k  )  &
                           *U(PRES_VAR,i+1,j,  k  )  &
                        +2.*U(MAGP_VAR,i+1,j,  k  )) &
                           /U(DENS_VAR,i+1,j,  k  )),&
                      sqrt((U(GAMC_VAR,i,  j-1,k  )  &
                           *U(PRES_VAR,i,  j-1,k  )  &
                        +2.*U(MAGP_VAR,i,  j-1,k  )) &
                           /U(DENS_VAR,i,  j-1,k  )),&
                      sqrt((U(GAMC_VAR,i,  j+1,k  )  &
                           *U(PRES_VAR,i,  j+1,k  )  &
                        +2.*U(MAGP_VAR,i,  j+1,k  )) &
                           /U(DENS_VAR,i,  j+1,k  )) )

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  )&
                        +U(VELY_VAR,i,  j+1,k  ) - U(VELY_VAR,i,  j-1,k  ))

           gradPx = 0.5*(U(PRES_VAR,i+1,j,  k  ) - U(PRES_VAR,i-1,j,  k  ))
           gradPy = 0.5*(U(PRES_VAR,i,  j+1,k  ) - U(PRES_VAR,i,  j-1,k  ))
           gradPz = 0.

           if ( abs(gradPx)+abs(gradPy) > beta*minP ) then
              SW1 = .true.
           endif

           if (-delta*minC > divv) then
              SW2 = .true.
           endif

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  )/del(DIR_X)&
                        +U(VELY_VAR,i,  j+1,k  ) - U(VELY_VAR,i,  j-1,k  )/del(DIR_Y))

           if (SW1 .and. SW2) then
              if (hy_cfl > 0.45) hy_cfl = 0.45
              hy_fullyLimit = .true.

              if (abs(minC)   > globalMinMaxs(1)) globalMinMaxs(1) = minC
              if ((divv < 0.) .and. (abs(divv) < globalMinMaxs(2))) globalMinMaxs(2) = divv !compression
              if ((divv > 0.) .and. (abs(divv) > globalMinMaxs(3))) globalMinMaxs(3) = divv !rarefaction
              if (abs(gradPx) > globalMinMaxs(4)) globalMinMaxs(4) = gradPx/del(DIR_X)
              if (abs(gradPy) > globalMinMaxs(5)) globalMinMaxs(5) = gradPy/del(DIR_Y)
           else
              hy_cfl = hy_cfl_original
              hy_fullyLimit = .false.
           endif

#elif NDIM == 3
           minP = min(U(PRES_VAR,i,  j,  k  ),&
                      U(PRES_VAR,i-1,j,  k  ),&
                      U(PRES_VAR,i+1,j,  k  ),&
                      U(PRES_VAR,i,  j-1,k  ),&
                      U(PRES_VAR,i,  j+1,k  ),&
                      U(PRES_VAR,i,  j,  k-1),&
                      U(PRES_VAR,i,  j,  k+1))

           minC = min(sqrt((U(GAMC_VAR,i,  j,  k  )  &
                           *U(PRES_VAR,i,  j,  k  )  &
                        +2.*U(MAGP_VAR,i,  j,  k  )) &
                           /U(DENS_VAR,i,  j,  k  )),&
                      sqrt((U(GAMC_VAR,i-1,j,  k  )  &
                           *U(PRES_VAR,i-1,j,  k  )  &
                        +2.*U(MAGP_VAR,i-1,j,  k  )) &
                           /U(DENS_VAR,i-1,j,  k  )),&
                      sqrt((U(GAMC_VAR,i+1,j,  k  )  &
                           *U(PRES_VAR,i+1,j,  k  )  &
                        +2.*U(MAGP_VAR,i+1,j,  k  )) &
                           /U(DENS_VAR,i+1,j,  k  )),&
                      sqrt((U(GAMC_VAR,i,  j-1,k  )  &
                           *U(PRES_VAR,i,  j-1,k  )  &
                        +2.*U(MAGP_VAR,i,  j-1,k  )) &
                           /U(DENS_VAR,i,  j-1,k  )),&
                      sqrt((U(GAMC_VAR,i,  j+1,k  )  &
                           *U(PRES_VAR,i,  j+1,k  )  &
                        +2.*U(MAGP_VAR,i,  j+1,k  )) &
                           /U(DENS_VAR,i,  j+1,k  )),&
                      sqrt((U(GAMC_VAR,i,  j,  k-1)  &
                           *U(PRES_VAR,i,  j,  k-1)  &
                        +2.*U(MAGP_VAR,i,  j,  k-1)) &
                           /U(DENS_VAR,i,  j,  k-1)),&
                      sqrt((U(GAMC_VAR,i,  j,  k+1)  &
                           *U(PRES_VAR,i,  j,  k+1)  &
                        +2.*U(MAGP_VAR,i,  j,  k+1)) &
                           /U(DENS_VAR,i,  j,  k+1)) )

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  )&
                        +U(VELY_VAR,i,  j+1,k  ) - U(VELY_VAR,i,  j-1,k  )&
                        +U(VELZ_VAR,i,  j,  k+1) - U(VELZ_VAR,i,  j,  k-1))

           gradPx = 0.5*(U(PRES_VAR,i+1,j,  k  ) - U(PRES_VAR,i-1,j,  k  ))
           gradPy = 0.5*(U(PRES_VAR,i,  j+1,k  ) - U(PRES_VAR,i,  j-1,k  ))
           gradPz = 0.5*(U(PRES_VAR,i,  j,  k+1) - U(PRES_VAR,i,  j,  k-1))

           if ( abs(gradPx)+abs(gradPy)+abs(gradPz) > beta*minP ) then
              SW1 = .true.
           endif

           if (-delta*minC > divv) then
              SW2 = .true.
           endif

           divv   = 0.5*(U(VELX_VAR,i+1,j,  k  ) - U(VELX_VAR,i-1,j,  k  )/del(DIR_X)&
                        +U(VELY_VAR,i,  j+1,k  ) - U(VELY_VAR,i,  j-1,k  )/del(DIR_Y)&
                        +U(VELZ_VAR,i,  j,  k+1) - U(VELZ_VAR,i,  j,  k-1)/del(DIR_Z))

           if (SW1 .and. SW2) then

!!$              if (hy_RiemannSolver == HYBRL) then !local hybrid method
!!$                 !local hybrid method which applies (a diffusive) HLL solver when scrch_Ctr = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i,  j,  k  ) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i+1,j,  k  ) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i-1,j,  k  ) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i,  j+1,k  ) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i,  j-1,k  ) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i,  j,  k+1) = 1.
!!$                 scrch_Ctr(FLAG_SCRATCH_CENTER_VAR,i,  j,  k-1) = 1.
!!$              elseif (hy_RiemannSolver == HYBRG) then
!!$                 !global hybrid method that uses HLL solver when there is a detection of shock in the domain
!!$                 hy_RiemannSolverLoc = HLL
!!$              endif


!!$              if (.not. hy_hybridRiemannOnly) then
                 if (hy_cfl > 0.25) hy_cfl = 0.25
                 hy_fullyLimit = .true.

!!$                 ! order reducing
!!$                 if (hy_order_original == 2) then
!!$                    hy_order = 1
!!$                 elseif (hy_order_original > 2) then
!!$                    hy_order = 2
!!$                 endif

!!$                 if (myPE .EQ. MASTER_PE) print*, &
!!$                      "[hy_uhd_detectShock] Strong shocks detected! Now increasing stability limit."

                 if (abs(minC)   > globalMinMaxs(1)) globalMinMaxs(1) = minC
                 if ((divv < 0.) .and. (abs(divv) < globalMinMaxs(2))) globalMinMaxs(2) = divv !compression
                 if ((divv > 0.) .and. (abs(divv) > globalMinMaxs(3))) globalMinMaxs(3) = divv !rarefaction
                 if (abs(gradPx) > globalMinMaxs(4)) globalMinMaxs(4) = gradPx/del(DIR_X)
                 if (abs(gradPy) > globalMinMaxs(5)) globalMinMaxs(5) = gradPy/del(DIR_Y)
                 if (abs(gradPz) > globalMinMaxs(6)) globalMinMaxs(6) = gradPz/del(DIR_Z)
!!$              endif

           else

              hy_cfl = hy_cfl_original
              hy_fullyLimit = .false.

!!$              if (hy_order .ne. hy_order_original) then
!!$                 hy_order = hy_order_original
!!$!$                 if (myPE .EQ. MASTER_PE) print*, &
!!$!$                      "[hy_uhd_detectShock] Done with strong shock regimes."//&
!!$!$                      " Now relaxing stability limit."
!!$              endif

           endif


#endif

        enddo
     enddo
  enddo

#ifdef DEBUG_HY_SHOCK_DETECT
  if (hy_fullyLimit) then
     call Logfile_open(logUnit,logUnitLocal)
     write(logUnit,*)'Shock detected: printing global max at nstep=',dr_nStep
     write(logUnit,*)'SoundVel+Alfven,divvComp,divvRare,gradPx,gradPy,gradPz=',globalMinMaxs(:)
     call Logfile_close(hy_myPE,logUnitLocal)
  endif
#endif

  ! Release block pointer
  call Grid_releaseBlkPtr(blockID,U,CENTER)

End Subroutine hy_uhd_shockDetect
