!!****if* source/physics/Hydro/HydroMain/unsplit/Anisotropic_MHD/hy_uhd_shockInstabilityFix
!!
!! NAME
!!
!!  hy_uhd_shockInstabilityFix
!!
!! SYNOPSIS
!!
!!  hy_uhd_shockInstabilityFix( integer(IN) :: blockID,
!!                              integer(IN) :: blkLimits(2,MDIM),
!!                              integer(IN) :: blkLimitsGC(2,MDIM),
!!                              real(IN)    :: Lambda(:,:,:,:,:),
!!                              real(OUT)   :: LambdaFix(:,:,:,:,:))
!!
!! ARGUMENTS
!!
!!  blockID     - a current block ID
!!  blkLimits   - an array that holds the lower and upper indices of the section
!!                of block without the guard cells
!!  blkLimitsGC - an array that holds the lower and upper indices of the section
!!                of block with the guard cells
!!  Lambda      - unmodified eigenvalues
!!  LambdaFix   - modified sets of eigenvalues
!!
!! DESCRIPTION
!!
!!  This routine adds numerical dissipations to transverse fluxes in order to cure
!!  shock instabilities such as carbuncle instabilities, odd-even decouplings, etc.
!!  The use of this algorithm is available with Roe Riemann solver.
!!  Please note that using this shockInstability method may require a reduced CFL
!!  limit for stability.
!!
!! REFERENCE
!!
!!  T. Hanawa, H. Mikami, and T. Matsumoto, JCP 227 (2008), 7952 -- 7976
!!
!!***

!!REORDER(4): U

Subroutine hy_uhd_shockInstabilityFix(blockID,blkLimits,blkLimitsGC,Lambda,LambdaFix)

  implicit none

#include "constants.h"
#include "Flash.h"
#include "UHD.h"


  !! Arguments type declaration ------------------------------
  integer, intent(IN)  :: blockID
  integer, dimension(LOW:HIGH,MDIM),intent(IN) :: blkLimits, blkLimitsGC

#ifdef FIXEDBLOCKSIZE
  real, intent(IN) :: Lambda(NDIM,HY_WAVENUM,&
                  GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC)
  real, dimension(NDIM,2, &
                  GRID_ILO_GC:GRID_IHI_GC, &
                  GRID_JLO_GC:GRID_JHI_GC, &
                  GRID_KLO_GC:GRID_KHI_GC),&
                  intent(OUT):: LambdaFix
#else
  real, intent(IN) :: Lambda(NDIM,HY_WAVENUM,&
       blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
       blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
       blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS))
  real, dimension(NDIM,2,&
                  blkLimitsGC(LOW,IAXIS):blkLimitsGC(HIGH,IAXIS), &
                  blkLimitsGC(LOW,JAXIS):blkLimitsGC(HIGH,JAXIS), &
                  blkLimitsGC(LOW,KAXIS):blkLimitsGC(HIGH,KAXIS)),&
                  intent(OUT) :: LambdaFix
#endif
  !! ---------------------------------------------------------
  integer :: i,j,k,dir
  integer :: i0,imax,j0,jmax,k0,kmax,ibeg,iend,jbeg,jend,kbeg,kend
  real :: lambdaFluxX1,lambdaFluxX2,lambdaFluxX3,&
          lambdaFluxX4,lambdaFluxX5,lambdaFluxX6,lambdaFluxXmax,&
          lambdaFluxY1,lambdaFluxY2,lambdaFluxY3,&
          lambdaFluxY4,lambdaFluxY5,lambdaFluxY6,lambdaFluxYmax,&
          lambdaFluxZ1,lambdaFluxZ2,lambdaFluxZ3,&
          lambdaFluxZ4,lambdaFluxZ5,lambdaFluxZ6,lambdaFluxZmax


  i0   = blkLimits(LOW, IAXIS)
  imax = blkLimits(HIGH,IAXIS)
  j0   = blkLimits(LOW, JAXIS)
  jmax = blkLimits(HIGH,JAXIS)
  k0   = blkLimits(LOW, KAXIS)
  kmax = blkLimits(HIGH,KAXIS)

  if (NDIM == 1) then
     jbeg= 4
     jend=-2
     kbeg= 4
     kend=-2
  elseif (NDIM == 2) then
     jbeg= j0
     jend= jmax
     kbeg= 4
     kend=-2
  else
     jbeg= j0
     jend= jmax
     kbeg= k0
     kend= kmax
  endif


  do k=kbeg-3,kend+3
     do j=jbeg-3,jend+3
        do i=i0-3,imax+3

#if NDIM >= 2

           ! Calculate gradients in y-direction for cure of Flux in x-direction
           lambdaFluxX1 = max(max(Lambda(DIR_Y,HY_FASTLEFT,i-1,j-1,k  )-Lambda(DIR_Y,HY_FASTLEFT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWLEFT,i-1,j-1,k  )-Lambda(DIR_Y,HY_SLOWLEFT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWRGHT,i-1,j-1,k  )-Lambda(DIR_Y,HY_SLOWRGHT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_FASTRGHT,i-1,j-1,k  )-Lambda(DIR_Y,HY_FASTRGHT,i-1,j,  k  ),0.))

           lambdaFluxX2 = max(max(Lambda(DIR_Y,HY_FASTLEFT,i-1,j,  k  )-Lambda(DIR_Y,HY_FASTLEFT,i-1,j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWLEFT,i-1,j,  k  )-Lambda(DIR_Y,HY_SLOWLEFT,i-1,j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWRGHT,i-1,j,  k  )-Lambda(DIR_Y,HY_SLOWRGHT,i-1,j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_FASTRGHT,i-1,j,  k  )-Lambda(DIR_Y,HY_FASTRGHT,i-1,j+1,k  ),0.))

           lambdaFluxX3 = max(max(Lambda(DIR_Y,HY_FASTLEFT,i,  j-1,k  )-Lambda(DIR_Y,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWLEFT,i,  j-1,k  )-Lambda(DIR_Y,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWRGHT,i,  j-1,k  )-Lambda(DIR_Y,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_FASTRGHT,i,  j-1,k  )-Lambda(DIR_Y,HY_FASTRGHT,i,  j,  k  ),0.))

           lambdaFluxX4 = max(max(Lambda(DIR_Y,HY_FASTLEFT,i,  j,  k  )-Lambda(DIR_Y,HY_FASTLEFT,i,  j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWLEFT,i,  j,  k  )-Lambda(DIR_Y,HY_SLOWLEFT,i,  j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWRGHT,i,  j,  k  )-Lambda(DIR_Y,HY_SLOWRGHT,i,  j+1,k  ),0.),&
                              max(Lambda(DIR_Y,HY_FASTRGHT,i,  j,  k  )-Lambda(DIR_Y,HY_FASTRGHT,i,  j+1,k  ),0.))

           lambdaFluxXmax = max(lambdaFluxX1,lambdaFluxX2,lambdaFluxX3,lambdaFluxX4)



           ! Calculate gradients in x-direction for cure of Flux in y-direction
           lambdaFluxY1 = max(max(Lambda(DIR_X,HY_FASTLEFT,i-1,j-1,k  )-Lambda(DIR_X,HY_FASTLEFT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWLEFT,i-1,j-1,k  )-Lambda(DIR_X,HY_SLOWLEFT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWRGHT,i-1,j-1,k  )-Lambda(DIR_X,HY_SLOWRGHT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_FASTRGHT,i-1,j-1,k  )-Lambda(DIR_X,HY_FASTRGHT,i,  j-1,k  ),0.))

           lambdaFluxY2 = max(max(Lambda(DIR_X,HY_FASTLEFT,i,  j-1,k  )-Lambda(DIR_X,HY_FASTLEFT,i+1,j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWLEFT,i,  j-1,k  )-Lambda(DIR_X,HY_SLOWLEFT,i+1,j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWRGHT,i,  j-1,k  )-Lambda(DIR_X,HY_SLOWRGHT,i+1,j-1,k  ),0.),&
                              max(Lambda(DIR_X,HY_FASTRGHT,i,  j-1,k  )-Lambda(DIR_X,HY_FASTRGHT,i+1,j-1,k  ),0.))

           lambdaFluxY3 = max(max(Lambda(DIR_X,HY_FASTLEFT,i-1,j,  k  )-Lambda(DIR_X,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWLEFT,i-1,j,  k  )-Lambda(DIR_X,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWRGHT,i-1,j,  k  )-Lambda(DIR_X,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_FASTRGHT,i-1,j,  k  )-Lambda(DIR_X,HY_FASTRGHT,i,  j,  k  ),0.))

           lambdaFluxY4 = max(max(Lambda(DIR_X,HY_FASTLEFT,i,  j,  k)-Lambda(DIR_X,HY_FASTLEFT,i+1,j,  k),0.),&
                              max(Lambda(DIR_X,HY_SLOWLEFT,i,  j,  k)-Lambda(DIR_X,HY_SLOWLEFT,i+1,j,  k),0.),&
                              max(Lambda(DIR_X,HY_SLOWRGHT,i,  j,  k)-Lambda(DIR_X,HY_SLOWRGHT,i+1,j,  k),0.),&
                              max(Lambda(DIR_X,HY_FASTRGHT,i,  j,  k)-Lambda(DIR_X,HY_FASTRGHT,i+1,j,  k),0.))

           lambdaFluxYmax = max(lambdaFluxY1,lambdaFluxY2,lambdaFluxY3,lambdaFluxY4)


           ! Store maxima for Flux in x-direction
           LambdaFix(DIR_X,1,i,j,k) = lambdaFluxXmax
           LambdaFix(DIR_X,2,i,j,k) = lambdaFluxY3


           ! Store maxima for Flux in y-direction
           LambdaFix(DIR_Y,1,i,j,k) = lambdaFluxYmax
           LambdaFix(DIR_Y,2,i,j,k) = lambdaFluxX3


#if NDIM == 3

           ! Calculate gradients in x & y-directions for cure of Flux in z-direction
           lambdaFluxZ1   = max(lambdaFluxX1,lambdaFluxY1)
           lambdaFluxZ2   = max(lambdaFluxX2,lambdaFluxY2)
           lambdaFluxZ3   = max(lambdaFluxX3,lambdaFluxY3)
           lambdaFluxZ4   = max(lambdaFluxX4,lambdaFluxY4)
           lambdaFluxZmax = max(lambdaFluxZ1,lambdaFluxZ2,lambdaFluxZ3,lambdaFluxZ4)


           ! Calculate gradients in y & z -directions for cure of Flux in x-direction
           lambdaFluxX1 = max(lambdaFluxX1,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i-1,j,  k-1)-Lambda(DIR_Z,HY_FASTLEFT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i-1,j,  k-1)-Lambda(DIR_Z,HY_SLOWLEFT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i-1,j,  k-1)-Lambda(DIR_Z,HY_SLOWRGHT,i-1,j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i-1,j,  k-1)-Lambda(DIR_Z,HY_FASTRGHT,i-1,j,  k  ),0.))

           lambdaFluxX2 = max(lambdaFluxX2,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i-1,j,  k  )-Lambda(DIR_Z,HY_FASTLEFT,i-1,j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i-1,j,  k  )-Lambda(DIR_Z,HY_SLOWLEFT,i-1,j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i-1,j,  k  )-Lambda(DIR_Z,HY_SLOWRGHT,i-1,j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i-1,j,  k  )-Lambda(DIR_Z,HY_FASTRGHT,i-1,j,  k+1),0.))

           lambdaFluxX3 = max(lambdaFluxX3,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k  ),0.))

           lambdaFluxX4 = max(lambdaFluxX4,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k  )-Lambda(DIR_Z,HY_FASTLEFT,i,  j,k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k  )-Lambda(DIR_Z,HY_SLOWLEFT,i,  j,k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k  )-Lambda(DIR_Z,HY_SLOWRGHT,i,  j,k+1),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k  )-Lambda(DIR_Z,HY_FASTRGHT,i,  j,k+1),0.))


           lambdaFluxXmax = max(lambdaFluxX1,lambdaFluxX2,lambdaFluxX3,lambdaFluxX4)


           ! Calculate gradients in x & z-directions for cure of Flux in y-direction
           lambdaFluxY1 = max(lambdaFluxY1,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j-1,k-1)-Lambda(DIR_Z,HY_FASTLEFT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j-1,k-1)-Lambda(DIR_Z,HY_SLOWLEFT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j-1,k-1)-Lambda(DIR_Z,HY_SLOWRGHT,i,  j-1,k  ),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j-1,k-1)-Lambda(DIR_Z,HY_FASTRGHT,i,  j-1,k  ),0.))

           lambdaFluxY2 = max(lambdaFluxY2,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j-1,k  )-Lambda(DIR_Z,HY_FASTLEFT,i,  j-1,k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j-1,k  )-Lambda(DIR_Z,HY_SLOWLEFT,i,  j-1,k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j-1,k  )-Lambda(DIR_Z,HY_SLOWRGHT,i,  j-1,k+1),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j-1,k  )-Lambda(DIR_Z,HY_FASTRGHT,i,  j-1,k+1),0.))

           lambdaFluxY3 = max(lambdaFluxY3,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k  ),0.))

           lambdaFluxY4 = max(lambdaFluxY4,&
                              max(Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k  )-Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k  )-Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k  )-Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k+1),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k  )-Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k+1),0.))


           lambdaFluxYmax = max(lambdaFluxY1,lambdaFluxY2,lambdaFluxY3,lambdaFluxY4)


           ! Store maxima information for Flux in x-direction
           LambdaFix(DIR_X,1,i,j,k) = lambdaFluxXmax
           LambdaFix(DIR_X,2,i,j,k) = &
                          max(max(Lambda(DIR_X,HY_FASTLEFT,i-1,j,  k  )-Lambda(DIR_X,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWLEFT,i-1,j,  k  )-Lambda(DIR_X,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_SLOWRGHT,i-1,j,  k  )-Lambda(DIR_X,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_X,HY_FASTRGHT,i-1,j,  k  )-Lambda(DIR_X,HY_FASTRGHT,i,  j,  k  ),0.))


           ! Store maxima for Flux in y-direction
           LambdaFix(DIR_Y,1,i,j,k) = lambdaFluxYmax
           LambdaFix(DIR_Y,2,i,j,k) = &
                          max(max(Lambda(DIR_Y,HY_FASTLEFT,i,  j-1,k  )-Lambda(DIR_Y,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWLEFT,i,  j-1,k  )-Lambda(DIR_Y,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_SLOWRGHT,i,  j-1,k  )-Lambda(DIR_Y,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Y,HY_FASTRGHT,i,  j-1,k  )-Lambda(DIR_Y,HY_FASTRGHT,i,  j,  k  ),0.))

           ! Store maxima for Flux in z-direction
           LambdaFix(DIR_Z,1,i,j,k) = lambdaFluxZmax
           LambdaFix(DIR_Z,2,i,j,k) = &
                          max(max(Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWLEFT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_SLOWRGHT,i,  j,  k  ),0.),&
                              max(Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k-1)-Lambda(DIR_Z,HY_FASTRGHT,i,  j,  k  ),0.))

#endif
#endif


        enddo
     enddo
  enddo

End Subroutine hy_uhd_shockInstabilityFix
