!!****if* source/physics/Hydro/HydroMain/unsplit/MHD_StaggeredMesh/hy_uhd_unsplit
!!
!! NAME
!!
!!  hy_uhd_unsplit
!!
!! SYNOPSIS
!!
!!  hy_uhd_unsplit( integer (IN) :: myPE, 
!!                  integer (IN) :: blockCount,
!!                  integer (IN) :: blockList(blockCount),
!!                  real    (IN) :: dt,
!!                  real    (IN) :: dtOld  )
!!
!! DESCRIPTION
!!
!!  Performs MHD update in a directionally unsplit fashion over a set
!!  of blocks. Divergence cleaning control for the magnetic fields is
!!  handled by using the staggered mesh algorithm with a call to
!!  hy_uhd_staggeredDivb. Before calling this routine, electric fields are
!!  to be calculated with a call to hy_uhd_getElectricFields.
!!  blockList is an integer array of size blockCount that contains 
!!  the blocks over which to update.
!!  dt gives the timestep over which to advance, and timeEndAdv gives the
!!  simulation time at the end of the update.
!! 
!!  This routine performs a guardcell fill and for each block: 
!!   - applies an eos to the guard cells; 
!!   - computes fluxes using a call to hy_uhd_getFaceFlux
!!   - if we're not doing flux correction (as controlled by the flux_correct
!!     runtime parameter), then we update all the cell values from the fluxes 
!!     (with a call to hy_uhd_unsplitUpdate), otherwise, we update just cells 
!!     not on the boundaries, and save fluxes for cells on the boundary;
!!   - and finally, we apply an eos to the block.
!! 
!!  After the main block loop, if doing flux correction, we have
!!  the Grid correct boundary fluxes for all blocks where approriate,
!!  and do another loop over blocks, updating the cell values for
!!  cells on the block boundaries using the corrected fluxes, and
!!  apply an eos on the block. 
!!  The same is true for the electric fields correction where
!!  there are block boundaries that are sharing different levels of
!!  refinements.
!!
!!
!! REFERENCES
!!
!!  * Lee, D., Ph.D. Dissertation, Univ. of MD, 2006
!!  * Lee, D. and Deane, A., "An Unsplit Staggered Mesh Scheme for Multidimensional
!!                            Magnetohydrodynamics", 228 (2009), 952-975, JCP
!!
!!
!! ARGUMENTS
!!
!!  myPE       -  local processor number
!!  blockCount -  the number of blocks in blockList
!!  blockList  -  array holding local IDs of blocks on which to advance
!!  dt         -  timestep
!!  dtOld      -  old timestep (needed for temporal extrapolations of gravity)
!!
!!***

!!REORDER(4): U, fl[xyz]

#ifdef DEBUG_ALL
#define DEBUG_MHD
#endif

Subroutine hy_uhd_unsplit ( myPE, blockCount, blockList, dt, dtOld )

  use Hydro_data, ONLY : hy_fluxCorrect,    &
                         hy_killdivb,       &
                         hy_gref,           &
                         hy_useGravity,     &
                         hy_order,          &
                         hy_units,          &
                         hy_gcMaskSize,     &
                         hy_gcMask,         &
                         hy_unsplitEosMode, &
                         hy_forceHydroLimit,&
                         hy_facevar2ndOrder,&
                         hy_flux_dt

  use Driver_interface, ONLY : Driver_abortFlash

  use hy_uhd_interface, ONLY : hy_uhd_getRiemannState,  &
                               hy_uhd_getFaceFlux,      &
                               hy_uhd_unsplitUpdate,    &
                               hy_uhd_getElectricFields,&
                               hy_uhd_staggeredDivb,    &
                               hy_uhd_unitConvert,      &
                               hy_uhd_energyFix,        &
                               hy_uhd_putGravityUnsplit,&
                               hy_uhd_eigenParameters,&
                               hy_uhd_eigenValue

  use Grid_interface, ONLY : Grid_getDeltas,         &
                             Grid_getBlkIndexLimits, &
                             Grid_fillGuardCells,    &
                             Grid_putFluxData,       &
                             Grid_getFluxData,       &
                             Grid_conserveFluxes,    &
                             Grid_conserveField,     &
                             Grid_getBlkPtr,         &
                             Grid_releaseBlkPtr

  use Eos_interface, ONLY : Eos_wrapped

  use Timers_interface, ONLY : Timers_start, Timers_stop

  implicit none

#include "constants.h"
#include "Flash.h"
#include "Eos.h"
#include "UHD.h"


  !! ---- Argument List ----------------------------------
  integer, INTENT(IN) :: myPE, blockCount
  integer, INTENT(IN), dimension(blockCount) :: blockList
  real,    INTENT(IN) :: dt, dtOld
  !! -----------------------------------------------------

  integer, dimension(MDIM) :: datasize
  integer, dimension(LOW:HIGH,MDIM) :: blkLimits,blkLimitsGC
  integer :: i,blockID,istat,level=0
  integer :: ix,iy,iz
  real, dimension(MDIM) :: del
  logical :: gcMask(hy_gcMaskSize)
  logical :: halfTimeAdvance
  integer, dimension(2,MDIM) :: eosRange

#ifdef FIXEDBLOCKSIZE
  real :: flx(NFLUXES,&
              GRID_ILO_GC:GRID_IHI_GC,     &
              GRID_JLO_GC:GRID_JHI_GC,     &
              GRID_KLO_GC:GRID_KHI_GC)
  real :: fly(NFLUXES,&
              GRID_ILO_GC:GRID_IHI_GC,     &
              GRID_JLO_GC:GRID_JHI_GC,     &
              GRID_KLO_GC:GRID_KHI_GC)
  real :: flz(NFLUXES,&
              GRID_ILO_GC:GRID_IHI_GC,     &
              GRID_JLO_GC:GRID_JHI_GC,     &
              GRID_KLO_GC:GRID_KHI_GC)

  real, dimension(NDIM,                        &
                  GRID_ILO_GC:GRID_IHI_GC,     &
                  GRID_JLO_GC:GRID_JHI_GC,     &
                  GRID_KLO_GC:GRID_KHI_GC) ::  &
                  vint

  real, dimension(3,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC) :: gravX,gravY,gravZ
  real, dimension(NDIM,HY_WAVENUM,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC) :: lambda
#else
  real, allocatable, dimension(:,:,:,:) :: flx,fly,flz
  real, allocatable, dimension(:,:,:,:) :: vint
  real, allocatable, dimension(:,:,:,:)   :: gravX, gravY, gravZ
  real, allocatable, dimension(:,:,:,:,:) :: lambda
#endif

!#ifdef HY_NEED_EXTRA_GCFILL
  logical :: doEos
  real, pointer, dimension(:,:,:,:) :: U
!#endif

  real :: csx,cax,cfx,csy,cay,cfy,as,af,bbN,uN
  real,dimension(MDIM)         :: beta
  real,dimension(HY_WAVENUM)   :: lambdaLoc
  real,dimension(HY_VARINUM2)  :: Vloc
  !! End of data declaration ***********************************************


#ifdef FLASH_GRID_PARAMESH2
  call Driver_abortFlash("The staggeredMesh MHD solver only works with PARAMESH 3 or 4!")
#endif


#ifdef FLASH_GRID_UG
  hy_fluxCorrect = .false.
#endif

  !push back flux dt values
  do i = 10, 2, -1
     hy_flux_dt(i) = hy_flux_dt(i-1)
  enddo
  !set maximum time step due to flux to a large value
  hy_flux_dt(1) = HUGE(1.)

  !! Guardcell filling routine
  call Grid_fillGuardCells(myPE,CENTER_FACES,ALLDIR,&
       maskSize=hy_gcMaskSize, mask=hy_gcMask,makeMaskConsistent=.true.)


  !! ***************************************************************************
  !! First part of advancement                                                 *
  !! ***************************************************************************
  !! Loop over the blocks
  do i=1,blockCount

     blockID = blockList(i)

     if ( hy_units .NE. "NONE" .and. hy_units .NE. "none" ) then
        call hy_uhd_unitConvert(blockID,FWDCONVERT)
     endif

     call Grid_getDeltas(blockID,del)

     call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)


     !! Call to Eos for guardcell regions --------------------------------------
     !! Low end region in x-direction
     call Timers_start("eos gc")
     eosRange = blkLimitsGC
     eosRange(LOW, IAXIS) = blkLimitsGC(LOW,IAXIS)
     eosRange(HIGH,IAXIS) = blkLimits(LOW,IAXIS)-1
     call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)

     !! High end region in x-direction
     eosRange = blkLimitsGC
     eosRange(LOW, IAXIS) = blkLimits(HIGH,IAXIS)+1
     eosRange(HIGH,IAXIS) = blkLimitsGC(HIGH,IAXIS)
     call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)


     if (NDIM >= 2) then
        !! Low end region in y-direction
        eosRange = blkLimitsGC
        eosRange(LOW, JAXIS) = blkLimitsGC(LOW,JAXIS)
        eosRange(HIGH,JAXIS) = blkLimits(LOW,JAXIS)-1
        call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)
        !! High end region in y-direction
        eosRange = blkLimitsGC
        eosRange(LOW, JAXIS) = blkLimits(HIGH,JAXIS)+1
        eosRange(HIGH,JAXIS) = blkLimitsGC(HIGH,JAXIS)
        call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)

        if (NDIM == 3) then
           !! Low end region in z-direction
           eosRange = blkLimitsGC
           eosRange(LOW, KAXIS) = blkLimitsGC(LOW,KAXIS)
           eosRange(HIGH,KAXIS) = blkLimits(LOW,KAXIS)-1
           call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)
           !! High end region in z-direction
           eosRange = blkLimitsGC
           eosRange(LOW, KAXIS) = blkLimits(HIGH,KAXIS)+1
           eosRange(HIGH,KAXIS) = blkLimitsGC(HIGH,KAXIS)
           call Eos_wrapped(hy_unsplitEosMode, eosRange, blockID)
        endif
     endif
     !! End of Eos call for guardcell regions ----------------------------------
     call Timers_stop("eos gc")

     datasize(1:MDIM)=blkLimitsGC(HIGH,1:MDIM)-blkLimitsGC(LOW,1:MDIM)+1
     
#ifndef FIXEDBLOCKSIZE
     allocate(flx(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(fly(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(flz(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(vint(MDIM,  dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravX(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravY(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravZ(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(lambda(NDIM,HY_WAVENUM,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
#endif

     !! Initialize arrays with zero
     flx = 0.
     fly = 0.
     flz = 0.
     vint= 0.

     !! ************************************************************************
     !! Get gravity
     gravX = 0.
     gravY = 0.
     gravZ = 0.

#ifdef GRAVITY
     if (hy_useGravity) then
        call hy_uhd_putGravityUnsplit(blockID,blkLimitsGC,dataSize,dt,dtOld,gravX,gravY,gravZ)
        gravX = gravX/hy_gref
        gravY = gravY/hy_gref
        gravZ = gravZ/hy_gref
     endif
#endif


     !! ************************************************************************
     !! Calculate Riemann (interface) states
     call Timers_start("get Riemann State")
     !! Note: gravX(1,:,:,:) - gravity at n
     !!       gravX(2,:,:,:) - gravity at n+1/2 (extrapolated)
     !!       gravX(3,:,:,:) - gravity at n+1 (exprapolated)
     call hy_uhd_getRiemannState(blockID,blkLimits,blkLimitsGC,dt,del, &
                                 gravX(1,:,:,:),gravY(1,:,:,:),gravZ(1,:,:,:),&
                                 gravX(2,:,:,:),gravY(2,:,:,:),gravZ(2,:,:,:),&
                                 lambda,normalFieldUpdate=.false.)
     call Timers_stop("get Riemann State")

     !! Don't do this part when pure hydro (NFACE_VARS = 0) is considered
#if NFACE_VARS > 0
#if NDIM > 1
     !! Note: Do this intermediate part for multidimensional problem only.
     !!       One dimensional problem doesn't need to do this part as 1D setup
     !!       doesn't need to use magnetic fields (facevars) at cell face centers.

     !! ************************************************************************
     !! Evolve facevars n+1/2 time step to get 2nd order accuracy
     if ((.not. hy_forceHydroLimit) .and. hy_facevar2ndOrder)then
        !! Intermediate Godunov fluxes
        call Timers_start("get Face Flux")
        call hy_uhd_getFaceFlux(blockID,blkLimits,blkLimitsGC,datasize,del,lambda,flx,fly,flz,vint)
        call Timers_stop("get Face Flux")
        if (hy_killdivb) then
           !! DivB=0 calls
           halfTimeAdvance = .true.
           !! Calculate Godunov fluxes one more time to calculate electric fields
           call Timers_start("get Electric Field")
           call hy_uhd_getElectricFields(blockID,blkLimits,blkLimitsGC,del,flx,fly,flz)
           call Timers_stop("get Electric Field")
        endif
     endif
     !! End of evolution of face-centered magnetic fields by n+1/2 time step
     !! ************************************************************************

#ifndef FIXEDBLOCKSIZE
     deallocate(flx)
     deallocate(fly)
     deallocate(flz)
     deallocate(vint)
     deallocate(gravX)
     deallocate(gravY)
     deallocate(gravZ)
     deallocate(lambda)
#endif
  enddo
  !! End of the first loop of evolution by n+1/2 time step
  !! ***************************************************************************



  !! ***************************************************************************
  !! Exchange guardcell data for face-centered magnetic fields that will be
  !! used in the data reconstruction-evolution calculation.
  if ( hy_facevar2ndOrder .and. (.not. hy_forceHydroLimit)) then

     if (hy_fluxCorrect .and. hy_killdivb) then
        !! Call the electric field correction routine
        call Timers_start("conserve Field")
        call Grid_conserveField(myPE)
        call Timers_stop("conserve Field")
     endif

     !! Calculate divergence-free magnetic fields at interfaces
     do i=1,blockCount
        blockID = blockList(i)
        call Grid_getDeltas(blockID,del)
        call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)

        halfTimeAdvance = .true.
        !! Evolve face-centered magnetic fields n+1/2 time step
        call Timers_start("staggered DivB")
        call hy_uhd_staggeredDivb(blockID,0.5*dt,del,blkLimits,halfTimeAdvance)
        call Timers_stop("staggered DivB")
     enddo

     !! Fill guardcells for only the face-centered field variable that are
     !! evolved to n+1/2 time step for date reconstruction
     gcMask  =.false.
     gcMask(MAGI_FACE_VAR:NFACE_VARS*NDIM:NFACE_VARS) = .true.

     !! Guardcell filling for facevars
     call Timers_start("fill gc")
     call Grid_fillGuardCells(myPE,FACES,ALLDIR,doEos=.false.,&
          maskSize=(hy_gcMaskSize-NUNK_VARS),mask=gcMask,makeMaskConsistent=.true.)
     call Timers_stop("fill gc")
  endif


  !! ***************************************************************************
  !! Second part of advancement                                                *
  !! ***************************************************************************
  !! Loop over the blocks
  do i=1,blockCount

     blockID = blockList(i)
     call Grid_getDeltas(blockID,del)
     call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)

     datasize(1:MDIM)=blkLimitsGC(HIGH,1:MDIM)-blkLimitsGC(LOW,1:MDIM)+1
#ifndef FIXEDBLOCKSIZE
     allocate(flx(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(fly(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(flz(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(vint(MDIM,  dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravX(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravY(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(gravZ(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
     allocate(lambda(NDIM,HY_WAVENUM,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
#endif

     ! Initialize arrays with zero
     flx = 0.
     fly = 0.
     flz = 0.
     vint= 0.


     !! ************************************************************************
     !! Get gravity
     gravX = 0.
     gravY = 0.
     gravZ = 0.

#ifdef GRAVITY
     if (hy_useGravity) then
        call hy_uhd_putGravityUnsplit(blockID,blkLimitsGC,dataSize,dt,dtOld,gravX,gravY,gravZ)
        gravX = gravX/hy_gref
        gravY = gravY/hy_gref
        gravZ = gravZ/hy_gref
     endif
#endif

     if ((hy_order >= 2) .and. hy_facevar2ndOrder) then
        !! *********************************************************************
        !! Calculate Riemann (interface) states with divergence-free fields at n+1/2
        call Timers_start("get Riemann State")
        !! Note: gravX(1,:,:,:) - gravity at n
        !!       gravX(2,:,:,:) - gravity at n+1/2 (extrapolated)
        !!       gravX(3,:,:,:) - gravity at n+1 (exprapolated)
        call hy_uhd_getRiemannState(blockID,blkLimits,blkLimitsGC,dt,del, &
                                    gravX(1,:,:,:),gravY(1,:,:,:),gravZ(1,:,:,:),&
                                    gravX(2,:,:,:),gravY(2,:,:,:),gravZ(2,:,:,:),&
                    lambda,normalFieldUpdate=.true.)
        call Timers_stop("get Riemann State")
     endif
#endif !end of #if NDIM > 1
#endif !end of #if NFACE_VARS > 0


     !! ************************************************************************
     !! Calculate high order Godunov fluxes
     call Timers_start("get Face Flux")
     call hy_uhd_getFaceFlux(blockID,blkLimits,blkLimitsGC,datasize,del,lambda,&
                             flx,fly,flz,vint,lastCall=.true.)
     call Timers_stop("get Face Flux")


     if (.not. hy_fluxCorrect) then
        !! ************************************************************************
        !! Unsplit update for conservative variables from n to n+1 time step
        call Timers_start("unsplit Update")
        call hy_uhd_unsplitUpdate(blockID,dt,del,datasize,blkLimits,&
                                  flx,fly,flz,gravX,gravY,gravZ)
        call Timers_stop("unsplit Update")

#if NFACE_VARS > 0
        !! ************************************************************************
        !! Update face-centered magnetic fields from n to n+1 time step
        if (.not. hy_forceHydroLimit) then
           if (NDIM >= 2) then
              !! DivB=0 calls
              if (hy_killdivb) then
                 halfTimeAdvance = .false.
                 call Timers_start("get Electric Field")
                 call hy_uhd_getElectricFields(blockID,blkLimits,blkLimitsGC,del,flx,fly,flz)
                 call Timers_stop("get Electric Field")
                 call Timers_start("staggered DivB")
                 call hy_uhd_staggeredDivb(blockID,dt,del,blkLimits,halfTimeAdvance)
                 call Timers_stop("staggered DivB")
              endif
           endif
        endif
#endif

        !! Correct energy if necessary
        call Timers_start("energy Fix")
        call hy_uhd_energyFix(blockID,blkLimits,dt,del,hy_unsplitEosMode)
        call Timers_stop("energy Fix")

        if ( hy_units .NE. "none" .and. hy_units .NE. "NONE" ) then
           !! Convert unit
           call hy_uhd_unitConvert(blockID,BWDCONVERT)
        endif

        !! Call to Eos
        call Timers_start("eos final")
        call Eos_wrapped(hy_unsplitEosMode, blkLimits, blockID)
        call Timers_stop("eos final")

     else !! if Flux correction is used.
        !! Flux conservation calls on AMR:
        !! Correct fluxes at each block boundary where coarse and fine
        !! blocks are neighboring each other.

        call Timers_start("put Flux Data")
        call Grid_putFluxData(blockID,IAXIS,flx,datasize)
        if (NDIM >= 2) then
           call Grid_putFluxData(blockID,JAXIS,fly,datasize)
           if (NDIM == 3) then
              call Grid_putFluxData(blockID,KAXIS,flz,datasize)
           endif
        endif
        call Timers_stop("put Flux Data")
     end if

     
#ifndef FIXEDBLOCKSIZE
     deallocate(flx)
     deallocate(fly)
     deallocate(flz)
     deallocate(vint)
     deallocate(gravX)
     deallocate(gravY)
     deallocate(gravZ)
     deallocate(lambda)
#endif
  end do
  !! End of leaf block do-loop before flux conserve call


  !! ***************************************************************************
  !! Third part of advancement                                                 *
  !! ***************************************************************************
  !! Do this part only if refining and flux correcting
  if (hy_fluxCorrect) then

     !! ************************************************************************
     !! Conservation of Fluxes at each block boundary
     call Timers_start("conserve Flux")
     call Grid_conserveFluxes(myPE,IAXIS,level)
     if (NDIM >= 2) then
        call Grid_conserveFluxes(myPE,JAXIS,level)
        if (NDIM == 3) then
           call Grid_conserveFluxes(myPE,KAXIS,level)
        endif
     endif
     call Timers_stop("conserve Flux")

#if NFACE_VARS > 0
     !! ************************************************************************
     !! Correct electric fields at fine-coarse boundaries
     if ((NDIM >= 2) .and. hy_killdivb .and. (.not. hy_forceHydroLimit)) then
        !! Call the electric field correction routine
        call Timers_start("conserve Field")
        call Grid_conserveField(myPE)
        call Timers_stop("conserve Field")
     endif
#endif


     !! ************************************************************************
     !! Perform updates for every blocks
     do i = 1,blockCount

        blockID = blockList(i)
        call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)
        call Grid_getDeltas(blockID,del)

        datasize(1:MDIM)=blkLimitsGC(HIGH,1:MDIM)-blkLimitsGC(LOW,1:MDIM)+1
#ifndef FIXEDBLOCKSIZE
        allocate(flx(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
        allocate(fly(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
        allocate(flz(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
        allocate(gravX(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
        allocate(gravY(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
        allocate(gravZ(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),stat=istat)
#endif

        !! Initialize arrays with zero
        flx = 0.
        fly = 0.
        flz = 0.

        !! *********************************************************************
        !! Get gravity
        gravX = 0.
        gravY = 0.
        gravZ = 0.

#ifdef GRAVITY
        if (hy_useGravity) then
           call hy_uhd_putGravityUnsplit(blockID,blkLimitsGC,dataSize,dt,dtOld,gravX,gravY,gravZ)
           gravX = gravX/hy_gref
           gravY = gravY/hy_gref
           gravZ = gravZ/hy_gref
        endif
#endif

        !! Get modified conserved flux values at block interfaces:
        !! This is important especially at the block interface at 
        !! different refinement levels of neighboring blocks
        call Timers_start("get Flux Data")
        call Grid_getFluxData(blockID,IAXIS,flx,datasize)
        if (NDIM >= 2) then
           call Grid_getFluxData(blockID,JAXIS,fly,datasize)
           if (NDIM == 3) then
              call Grid_getFluxData(blockID,KAXIS,flz,datasize)
           endif
        endif
        call Timers_stop("get Flux Data")


        !! *********************************************************************
        !! Unsplit update for conservative variables from n to n+1 time step
        call Timers_start("unsplit Update")
        call hy_uhd_unsplitUpdate(blockID,dt,del,datasize,blkLimits,&
                                  flx,fly,flz,gravX,gravY,gravZ)
        call Timers_stop("unsplit Update")

#if NFACE_VARS > 0
        !! *********************************************************************
        !! Update face-centered magnetic fields from n to n+1 time step
        if (NDIM >=2) then
           if (hy_killdivb) then
              halfTimeAdvance = .false.
              call Timers_start("staggered DivB")
              call hy_uhd_staggeredDivb(blockID,dt,del,blkLimits,halfTimeAdvance)
              call Timers_stop("staggered DivB")
           endif
        endif
#endif

        !! *********************************************************************
        !! Correct energy if necessary
        call Timers_start("energy Fix")
        call hy_uhd_energyFix(blockID,blkLimits,dt,del,hy_unsplitEosMode)
        call Timers_stop("energy Fix")

        !! *********************************************************************
        !! Convert unit if necessary
        if ( hy_units .NE. "none" .and. hy_units .NE. "NONE" ) then
           !! Convert unit
           call hy_uhd_unitConvert(blockID,BWDCONVERT)
        endif


        !! *********************************************************************
        !! Call to Eos
        call Timers_start("eos final")
        call Eos_wrapped(hy_unsplitEosMode, blkLimits, blockID)
        call Timers_stop("eos final")

#ifndef FIXEDBLOCKSIZE
        deallocate(flx)
        deallocate(fly)
        deallocate(flz)
        deallocate(gravX)
        deallocate(gravY)
        deallocate(gravZ)
#endif
        
     end do !! End of the loop over blocks
  end if !! End of the flux conservation routine


#ifdef FAST_VAR
  do i=1,blockCount

     blockID = blockList(i)
     call Grid_getBlkPtr(blockID,U,CENTER)
     call Grid_getDeltas(blockID,del)
     call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)

     do iz=blkLimits(LOW,KAXIS),blkLimits(HIGH,KAXIS)
        do iy=blkLimits(LOW,JAXIS),blkLimits(HIGH,JAXIS)
           do ix=blkLimits(LOW,IAXIS),blkLimits(HIGH,IAXIS)

              Vloc(HY_DENS)=U(DENS_VAR,ix,iy,iz)
              Vloc(HY_VELX:HY_VELZ)=U(VELX_VAR:VELZ_VAR,ix,iy,iz)
              Vloc(HY_MAGX:HY_MAGZ)=U(MAGX_VAR:MAGZ_VAR,ix,iy,iz)
              Vloc(HY_PRES)=U(PRES_VAR,ix,iy,iz)
              Vloc(HY_GAMC:HY_GAME)=U(GAMC_VAR:GAME_VAR,ix,iy,iz)

              call hy_uhd_eigenParameters(Vloc,DIR_X,.true.,&
                   uN,cfx,C_alfn=cax,C_slow=csx,A_f=af,A_s=as,B_beta=beta)
              call hy_uhd_eigenValue(lambdaLoc,uN,cfx,C_alfn=cax,C_slow=csx)

              call hy_uhd_eigenParameters(Vloc,DIR_Y,.true.,&
                   uN,cfy,C_alfn=cay,C_slow=csy,A_f=af,A_s=as,B_beta=beta)
              call hy_uhd_eigenValue(lambdaLoc,uN,cfy,C_alfn=cay,C_slow=csy)

              U(FAST_VAR,ix,iy,iz) = max(cfx,cfy)
              U(SLOW_VAR,ix,iy,iz) = max(csx,csy)

           enddo
        enddo
     enddo
  enddo
#endif

#if NFACE_VARS > 0
  !! ***********************************************************************
  !! Note: Need guardcell fills to compute current densities and divv.
  !!       Use this call only if needed as this is expensive after all!
#ifdef HY_NEED_EXTRA_GCFILL


#ifdef FLASH_GRID_PARAMESH
  gcMask  =.FALSE.
  gcMask(MAGX_VAR)=.true.
  gcMask(MAGY_VAR)=.true.
  gcMask(MAGZ_VAR)=.true.
  gcMask(VELX_VAR)=.true.
  gcMask(VELY_VAR)=.true.
  gcMask(VELY_VAR)=.true.

  call Grid_fillGuardCells(myPE,CENTER_FACES,ALLDIR,doEos=.false.,&
       maskSize=hy_gcMaskSize,mask=gcMask,makeMaskConsistent=.true.)
#endif

#ifdef FLASH_GRID_UG
  !! Note that, on UG, guardcell masking is not applicable
  call Grid_fillGuardCells(myPE,CENTER_FACES,ALLDIR)
#endif

  do i=1,blockCount

     blockID = blockList(i)
     call Grid_getBlkPtr(blockID,U,CENTER)
     call Grid_getDeltas(blockID,del)
     call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)

     do iz=blkLimits(LOW,KAXIS),blkLimits(HIGH,KAXIS)
        do iy=blkLimits(LOW,JAXIS),blkLimits(HIGH,JAXIS)
           do ix=blkLimits(LOW,IAXIS),blkLimits(HIGH,IAXIS)

              !! current density jx
#ifdef CURX_VAR
#if NDIM > 1
              U(CURX_VAR,ix,iy,iz)=&
                   0.5*(U(MAGZ_VAR,ix,iy+1,iz)-U(MAGZ_VAR,ix,iy-1,iz))/del(DIR_Y)
#if NDIM > 2
              U(CURX_VAR,ix,iy,iz)=U(CURX_VAR,ix,iy,iz)&
                   -0.5*(U(MAGY_VAR,ix,iy,iz+1)-U(MAGY_VAR,ix,iy,iz-1))/del(DIR_Z)
#endif
#endif
#endif

              !! current density jy
#ifdef CURY_VAR
              U(CURY_VAR,ix,iy,iz)=&
                   -0.5*(U(MAGZ_VAR,ix+1,iy,iz)-U(MAGZ_VAR,ix-1,iy,iz))/del(DIR_X)
#if NDIM > 2
              U(CURY_VAR,ix,iy,iz)=U(CURY_VAR,ix,iy,iz)&
                   +0.5*(U(MAGX_VAR,ix,iy,iz+1)-U(MAGX_VAR,ix,iy,iz-1))/del(DIR_Z)
#endif
#endif

              !! current density jz
#ifdef CURZ_VAR
              U(CURZ_VAR,ix,iy,iz)=&
                   0.5*(U(MAGY_VAR,ix+1,iy,iz)-U(MAGY_VAR,ix-1,iy,iz))/del(DIR_X)
#if NDIM > 1
              U(CURZ_VAR,ix,iy,iz)=U(CURZ_VAR,ix,iy,iz)&
                   -0.5*(U(MAGX_VAR,ix,iy+1,iz)-U(MAGX_VAR,ix,iy-1,iz))/del(DIR_Y)
#endif
#endif

#ifdef DIVV_VAR
              U(DIVV_VAR,ix,iy,iz) =&
                   0.5*((U(VELX_VAR,ix+1,iy,  iz  )-U(VELX_VAR,ix-1,iy,  iz  ))/del(DIR_X) &
                       +(U(VELY_VAR,ix,  iy+1,iz  )-U(VELY_VAR,ix,  iy-1,iz  ))/del(DIR_Y) &
                       +(U(VELZ_VAR,ix,  iy,  iz+1)-U(VELZ_VAR,ix,  iy,  iz-1))/del(DIR_Z))
#endif
           enddo
        enddo
     enddo

     call Grid_releaseBlkPtr(blockID,U,CENTER)
  enddo


#endif !end of #ifdef HY_NEED_EXTRA_GCFILL
#endif !end of #if NFACE_VARS > 0

End Subroutine hy_uhd_unsplit
