!!****if* source/physics/Hydro/HydroMain/unsplit/hy_uhd_unsplitUpdate
!!
!! NAME
!!
!!  hy_uhd_unsplitUpdate
!!
!! SYNOPSIS
!!
!!  hy_uhd_unsplitUpdate( integer(IN) :: blockID,
!!                        real(IN)    :: dt,
!!                        real(IN)    :: del(MDM),
!!                        integer(IN) :: dataSize(3),
!!                        integer(IN) :: blkLimits(2,MDIM),
!!                        real(IN)    :: xflux(:,:,:,:), 
!!                        real(IN)    :: yflux(:,:,:,:), 
!!                        real(IN)    :: zflux(:,:,:,:),
!!                        real(IN)    :: gravX(3,:,:,:),
!!                        real(IN)    :: gravY(3,:,:,:),
!!                        real(IN)    :: gravZ(3,:,:,:))
!!
!! ARGUMENTS
!!
!!   blockID      - current block ID
!!   dt           - timestep
!!   del          - deltas in {x,y,z} directions
!!   dataSize     - size of the current block
!!   blkLimits    - an array that holds the lower and upper indices of the section
!!                  of block without the guard cells
!!   xflux,yflux,zflux - cell face centered fluxes at each {=x,y,z} direction
!!   gravX,gravY,gravZ - gravity components in x,y,z directions
!!
!! DESCRIPTION
!!
!!   This routine updates the cell-centered conservative variables and intermediate
!!   internal energy to the next time step using an directionally unsplit scheme.
!!
!!***

!!REORDER(4): scrch_[XYZ], [xyz]flux, U

Subroutine hy_uhd_unsplitUpdate(blockID,dt,del,dataSize,blkLimits,&              
                                xflux,yflux,zflux,gravX,gravY,gravZ)

  use Hydro_data,      ONLY : hy_smalldens,hy_forceHydroLimit,hy_order,hy_irenorm,hy_numXN, &
                              hy_updateMagneticFields
  use hy_uhd_interface,ONLY : hy_uhd_updateSpeciesMassScalar
  use Grid_interface,  ONLY : Grid_getBlkPtr, Grid_releaseBlkPtr

  implicit none

#include "constants.h"
#include "Flash.h"
#include "Eos.h"
#include "UHD.h"

  !! ---- Arguments ---------------------------------
  integer,intent(IN) :: blockID
  real, intent(IN)   :: dt
  real, intent(IN)   :: del(MDIM)
  integer,dimension(MDIM),intent(IN) :: dataSize
  integer,intent(IN) :: blkLimits(LOW:HIGH,MDIM)
#ifdef FIXEDBLOCKSIZE
  real, intent(in) ::xflux(NFLUXES,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC)
  real, intent(in) ::yflux(NFLUXES,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC)
  real, intent(in) ::zflux(NFLUXES,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC)
  real, dimension(3,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC), intent(IN) :: gravX,gravY,gravZ

#else
  real, intent(in) :: xflux(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS))  
  real, intent(in) :: yflux(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS))  
  real, intent(in) :: zflux(NFLUXES,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS))
  real, dimension(3,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)),& 
                  intent(IN) :: gravX,gravY,gravZ
#endif
  !!---------------------------------------------------

  integer :: i,j,k,imin,imax,jmin,jmax,kmin,kmax
  real    :: dx, dy, dz
  real, dimension(HY_VARINUM) :: U0
  real, dimension(NFLUXES)    :: FL,FR,GL,GR,HL,HR
!!$  real, pointer, dimension(:,:,:,:) :: U, scrch_X,scrch_Y,scrch_Z
  real, pointer, dimension(:,:,:,:) :: U, scrch_Ctr!,scrch_Y,scrch_Z
  real    :: IntEner,tempPres
  integer :: sp, spn

#if (NSPECIES+NMASS_SCALARS) > 0
#ifdef FIXEDBLOCKSIZE
  real, dimension(hy_numXN,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC) :: SpOld
  real, dimension(6,GRID_IHI_GC,GRID_JHI_GC,GRID_KHI_GC) :: Uold
#else
  real, dimension(hy_numXN,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)) :: SpOld
  real, dimension(6,dataSize(IAXIS),dataSize(JAXIS),dataSize(KAXIS)) :: Uold
#endif
#endif

  !! Set ranges for update
  imin  = blkLimits(LOW, IAXIS)
  imax  = blkLimits(HIGH,IAXIS)

  dx = del(DIR_X)

  if (NDIM >= 2) then
     jmin  = blkLimits(LOW, JAXIS)
     jmax  = blkLimits(HIGH,JAXIS)
     dy = del(DIR_Y)
     if (NDIM == 3) then
        kmin  = blkLimits(LOW, KAXIS)
        kmax  = blkLimits(HIGH,KAXIS)
        dz = del(DIR_Z)
     endif
  endif


  ! Get block pointers
  call Grid_getBlkPtr(blockID,U,CENTER)
  call Grid_getBlkPtr(blockID,scrch_Ctr,SCRATCH_CTR)
!!$  if (NDIM > 1) call Grid_getBlkPtr(blockID,scrch_Y,SCRATCH_CTR)
!!$  if (NDIM > 2) call Grid_getBlkPtr(blockID,scrch_Z,SCRATCH_CTR)
#ifdef FLASH_USM_MHD
  if (hy_forceHydroLimit) U(MAGX_VAR:MAGZ_VAR,:,:,:) = 0.
#endif

#if (NSPECIES+NMASS_SCALARS) > 0
  do sp =  SPECIES_BEGIN, MASS_SCALARS_END !SPECIES_END
     spn= sp-NPROP_VARS
     SpOld(spn,:,:,:) = U(sp,:,:,:)
  enddo

  Uold(1,  :,:,:) = U(DENS_VAR,:,:,:)
  Uold(2,  :,:,:) = U(PRES_VAR,:,:,:)
  Uold(3:5,:,:,:) = U(VELX_VAR:VELZ_VAR,:,:,:)
  Uold(6,  :,:,:) = U(GAME_VAR,:,:,:)
#endif


#if NDIM == 1
  k=1
  j=1 ! only for 1D
  do i = imin,imax

     !! NOTE: (1)  xflux, yflux -- corrected fluxes
     !!       (2) scrch_Ctr,scrch_Y -- stored fluxes
     !! Fluxes for the 5 (or 8 for MHD) conservative variables and internal energy:
     !! density, momenta, magnetic fields, and total energy 
     !! plus two auxilary fluxes for internal energy.

#ifndef FLASH_GRID_UG
     if ((i==imin).or.(i==imax)) then
        FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
        FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
     else
        FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,  j,  k  )
        FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,  k  )
     endif
#else
     FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
     FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
#endif
     GL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
     HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
     GR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
     HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.


     !! Update intenal energy rho*eint
     IntEner  = U(DENS_VAR,i,j,k)*U(EINT_VAR,i,j,k)

     !! Note that XN04 holds XP05 pressure from hy_uhd_getFaceFlux.F90
     tempPres = .5*( scrch_Ctr(XN04_SCRATCH_CENTER_VAR,i,j,k) &
                    +scrch_Ctr(XN05_SCRATCH_CENTER_VAR,i,j,k))

     call updateInternalEnergy&
          (IntEner,tempPres,&
          FL(HY_EINT_FLUX:HY_PRES_FLUX),&
          FR(HY_EINT_FLUX:HY_PRES_FLUX),&
          GL(HY_EINT_FLUX:HY_PRES_FLUX),&
          GR(HY_EINT_FLUX:HY_PRES_FLUX),&
          HL(HY_EINT_FLUX:HY_PRES_FLUX),&
          HR(HY_EINT_FLUX:HY_PRES_FLUX),&
          dx,dy,dz,dt)

     U(EINT_VAR,i,j,k) = IntEner ! rho*eint


     !! Update conserved quantities
     U0(HY_DENS) = U(DENS_VAR,i,j,k)                                      !density
     U0(HY_XMOM:HY_ZMOM) = U(VELX_VAR:VELZ_VAR,i,j,k)*U(DENS_VAR,i,j,k)   !momenta
#ifndef FLASH_USM_MHD
     U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)                    !total gas energy 
#else
     U0(HY_MAGX:HY_MAGZ) = U(MAGX_VAR:MAGZ_VAR,i,j,k)                     !magnetic fields
     U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)+&                  !total plasma energy 
          .5*dot_product(U(MAGX_VAR:MAGZ_VAR,i,j,k),&
                         U(MAGX_VAR:MAGZ_VAR,i,j,k))
#endif

     call updateConservedVariable(U0(HY_DENS:HY_DENS+HY_VARINUM-1),&
                                  FL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  GL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  HL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  FR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  GR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  HR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                  gravX(:,i,j,k),gravY(:,i,j,k),gravZ(:,i,j,k),dx,dy,dz,dt)

     U(DENS_VAR,i,j,k) = max(U0(HY_DENS),hy_smalldens)                   !density
#ifdef FLASH_USM_MHD
     if( hy_updateMagneticFields ) then
        U(MAGX_VAR:MAGZ_VAR,i,j,k) = U0(HY_MAGX:HY_MAGZ)                 !magnetic fields
     endif
#endif
     U(ENER_VAR,i,j,k) = U0(HY_ENER)                                     !total plasma energy
     !! We will update velocity fields after species & mass scalar update

#if (NSPECIES+NMASS_SCALARS) > 0
     !! Note that the velocity fields here are old velocities at time step n, not n+1
     call hy_uhd_updateSpeciesMassScalar&
          (hy_order,U(DENS_VAR,i,j,k),&
          SpOld(1:hy_numXN,i-3:i+3,j,k),&
          Uold(1:6,i-3:i+3,j,k),&
          FL(HY_DENS_FLUX),FR(HY_DENS_FLUX),&
          GL(HY_DENS_FLUX),GR(HY_DENS_FLUX),&
          HL(HY_DENS_FLUX),HR(HY_DENS_FLUX),&
          dx,dy,dz,dt, U(SPECIES_BEGIN:MASS_SCALARS_END,i,j,k))
#endif
     !! Update velocity fields after species & mass scalar update
     U(VELX_VAR:VELZ_VAR,i,j,k) = U0(HY_XMOM:HY_ZMOM)/U0(HY_DENS)        !velocities
  enddo

#endif

#if NDIM == 2
  k=1
  do j=jmin,jmax
     do i=imin,imax

        !! NOTE: (1)  xflux, yflux -- corrected fluxes
        !!       (2) scrch_Ctr,scrch_Y -- stored fluxes
        !! Fluxes for the 5 (or 8 for MHD) conservative variables and internal energy:
        !! density, momenta, magnetic fields, and total energy 
        !! plus two auxilary fluxes for internal energy.
#ifndef FLASH_GRID_UG
        if (((i==imin).or.(i==imax)) .and. (j>=jmin+1).and.(j<=jmax-1)) then
           !! column bands in the left and right of the 2D domain
           FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
           GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
           GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
           HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
           HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.

        elseif (((j==jmin).or.(j==jmax)) .and. (i>=imin+1).and.(i<=imax-1)) then
           !! column bands in the top and bottom of the 2D domain
           FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
           FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
           GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
           HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
           HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.

        elseif (((i==imin).or.(i==imax)) .and. ((j==jmin).or.(j==jmax))) then
           !! 4 corner regions
           FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
           GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
           HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
           HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.

        else
           !! All the other interior cells
           FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
           FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
           GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
           GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
           HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
           HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
        endif
#else
        FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
        FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
        GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
        GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
        HL(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
        HR(HY_DENS_FLUX:HY_PRES_FLUX) = 0.
#endif
        !! Update intenal energy rho*eint
        IntEner  = U(DENS_VAR,i,j,k)*U(EINT_VAR,i,j,k)

        !! Note that XN04,YN04 hold XP05,YP05 pressure from hy_uhd_getFaceFlux.F90
        tempPres = .25*( scrch_Ctr(XN04_SCRATCH_CENTER_VAR,i,j,k) &
                        +scrch_Ctr(XN05_SCRATCH_CENTER_VAR,i,j,k) &
                        +scrch_Ctr(YN04_SCRATCH_CENTER_VAR,i,j,k) &
                        +scrch_Ctr(YN05_SCRATCH_CENTER_VAR,i,j,k) )

        call updateInternalEnergy&
             (IntEner,tempPres,&
             FL(HY_EINT_FLUX:HY_PRES_FLUX),&
             FR(HY_EINT_FLUX:HY_PRES_FLUX),&
             GL(HY_EINT_FLUX:HY_PRES_FLUX),&
             GR(HY_EINT_FLUX:HY_PRES_FLUX),&
             HL(HY_EINT_FLUX:HY_PRES_FLUX),&
             HR(HY_EINT_FLUX:HY_PRES_FLUX),&
             dx,dy,dz,dt)

        U(EINT_VAR,i,j,k) = IntEner ! internal energy rho*eint


        !! Update conserved quantities
        U0(HY_DENS) = U(DENS_VAR,i,j,k)                                      !density
        U0(HY_XMOM:HY_ZMOM) = U(VELX_VAR:VELZ_VAR,i,j,k)*U(DENS_VAR,i,j,k)   !momenta
#ifndef FLASH_USM_MHD
        U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)                    !total gas energy 
#else
        U0(HY_MAGX:HY_DENS+HY_VARINUM-1) = U(MAGX_VAR:MAGZ_VAR,i,j,k)        !magnetic fields
        U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)+&                  !total plasma energy 
             .5*dot_product(U(MAGX_VAR:MAGZ_VAR,i,j,k),&
                            U(MAGX_VAR:MAGZ_VAR,i,j,k))
#endif
        call updateConservedVariable(U0(HY_DENS:HY_DENS+HY_VARINUM-1),&
                                     FL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     GL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     HL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     FR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     GR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     HR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                     gravX(:,i,j,k),gravY(:,i,j,k),gravZ(:,i,j,k),dx,dy,dz,dt)

        U(DENS_VAR,i,j,k) = max(U0(HY_DENS),hy_smalldens)                    !density
#ifdef FLASH_USM_MHD
        if( hy_updateMagneticFields ) then
           U(MAGX_VAR:MAGZ_VAR,i,j,k) = U0(HY_MAGX:HY_MAGZ)                  !magnetic fields
        endif
#endif
        U(ENER_VAR,i,j,k) = U0(HY_ENER)                                      !total plasma energy
        !! We will update velocity fields after species & mass scalar update

#if (NSPECIES+NMASS_SCALARS) > 0
        !! Note that the velocity fields here are old velocities at time step n, not n+1
        call hy_uhd_updateSpeciesMassScalar&
             (hy_order,U(DENS_VAR,i,j,k),&
             SpOld(1:hy_numXN,i-3:i+3,j-3:j+3,k),&
             Uold(1:6,i-3:i+3,j-3:j+3,k),&
             FL(HY_DENS_FLUX),FR(HY_DENS_FLUX),&
             GL(HY_DENS_FLUX),GR(HY_DENS_FLUX),&
             HL(HY_DENS_FLUX),HR(HY_DENS_FLUX),&
             dx,dy,dz,dt, U(SPECIES_BEGIN:MASS_SCALARS_END,i,j,k))
#endif

        !! Update velocity fields after species & mass scalar update
        U(VELX_VAR:VELZ_VAR,i,j,k) = U0(HY_XMOM:HY_ZMOM)/U0(HY_DENS)        !velocities

     enddo
  enddo
#endif

#if NDIM == 3
  do k=kmin,kmax
     do j=jmin,jmax
        do i=imin,imax
#ifndef FLASH_GRID_UG
           !! NOTE: (1)  xflux, yflux -- corrected fluxes
           !!       (2) scrch_Ctr,scrch_Y -- stored fluxes
           !! Fluxes for the 5 (or 8 for MHD) conservative variables and internal energy:
           !! density, momenta, magnetic fields, and total energy 
           !! plus two auxilary fluxes for internal energy.
           !! All 26 special cases neighboring boundaries
           if (((k==kmin).or.(k==kmax)).and. (((i==imin).or.(i==imax)) &
                .and. (j>=jmin+1).and.(j<=jmax-1))) then
              !! strips of data in the top and bottom of the 3D cube at i=imin & imax (all four of them)
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k+1)

           elseif (((k==kmin).or.(k==kmax)).and. (((j==jmin).or.(j==jmax)) &
                .and. (i>=imin+1).and.(i<=imax-1))) then
              !! strips of data in the top and bottom of the 3D cube at j=jmin & jmax (all four of them)
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k+1)

           elseif (((i==imin).or.(i==imax)).and. (((j==jmin).or.(j==jmax)) &
                .and. (k>=kmin+1).and.(k<=kmax-1))) then
              !! columns of data on the sides of the 3D cube at j=jmin & jmax (all four of them)
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k+1)

           elseif ( ((i==imin).or.(i==imax)).and.&
                ((j>=jmin+1).and.(j<=jmax-1).and.(k>=kmin+1).and.(k<=kmax-1))) then
              !! two faces in x-direction
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k+1)

           elseif ( ((j==jmin).or.(j==jmax)).and.&
                ((i>=imin+1).and.(i<=imax-1).and.(k>=kmin+1).and.(k<=kmax-1))) then
              !! two faces in y-direction
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k+1)

           elseif ( ((k==kmin).or.(k==kmax)).and.&
                ((j>=jmin+1).and.(j<=jmax-1).and.(i>=imin+1).and.(i<=imax-1))) then
              !! two faces in z-direction
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k+1)

           elseif (((i==imin).or.(i==imax)).and.&
                ((j==jmin).or.(j==jmax)).and.&
                ((k==kmin).or.(k==kmax))) then
              !! 8 corner regions need updated fluxes in all direction
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k+1)

           else
              !! All the other interior cells
              FL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              FR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(XP01_SCRATCH_CENTER_VAR:XP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i+1,j,k)
              GL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              GR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(YP01_SCRATCH_CENTER_VAR:YP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j+1,k)
              HL(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k)
              HR(HY_DENS_FLUX:HY_PRES_FLUX) = scrch_Ctr(ZP01_SCRATCH_CENTER_VAR:ZP01_SCRATCH_CENTER_VAR+HY_SCRATCH_NUM-1,i,j,k+1)
           endif
#else
           FL(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           FR(HY_DENS_FLUX:HY_PRES_FLUX) = xflux(HY_DENS_FLUX:HY_PRES_FLUX,i+1,j,  k  )
           GL(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           GR(HY_DENS_FLUX:HY_PRES_FLUX) = yflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j+1,k  )
           HL(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k  )
           HR(HY_DENS_FLUX:HY_PRES_FLUX) = zflux(HY_DENS_FLUX:HY_PRES_FLUX,i,  j,  k+1)
#endif
           !! Update intenal energy rho*eint
           IntEner  = U(DENS_VAR,i,j,k)*U(EINT_VAR,i,j,k)

           !! Note that XN04,YN04,ZN04 hold XP05,YP05,ZP05 pressure from hy_uhd_getFaceFlux.F90
           tempPres = ( scrch_Ctr(XN04_SCRATCH_CENTER_VAR,i,j,k) &
                       +scrch_Ctr(XN05_SCRATCH_CENTER_VAR,i,j,k) &
                       +scrch_Ctr(YN04_SCRATCH_CENTER_VAR,i,j,k) &
                       +scrch_Ctr(YN05_SCRATCH_CENTER_VAR,i,j,k) &
                       +scrch_Ctr(ZN04_SCRATCH_CENTER_VAR,i,j,k) &
                       +scrch_Ctr(ZN05_SCRATCH_CENTER_VAR,i,j,k))/6.

           call updateInternalEnergy&
                (IntEner,tempPres,&
                FL(HY_EINT_FLUX:HY_PRES_FLUX),&
                FR(HY_EINT_FLUX:HY_PRES_FLUX),&
                GL(HY_EINT_FLUX:HY_PRES_FLUX),&
                GR(HY_EINT_FLUX:HY_PRES_FLUX),&
                HL(HY_EINT_FLUX:HY_PRES_FLUX),&
                HR(HY_EINT_FLUX:HY_PRES_FLUX),&
                dx,dy,dz,dt)

           U(EINT_VAR,i,j,k) = IntEner ! rho*eint


           !! Update conserved quantities
           U0(HY_DENS) = U(DENS_VAR,i,j,k)                                      !density
           U0(HY_XMOM:HY_ZMOM) = U(VELX_VAR:VELZ_VAR,i,j,k)*U(DENS_VAR,i,j,k)   !momenta
#ifndef FLASH_USM_MHD
           U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)                    !total gas energy 
#else
           U0(HY_MAGX:HY_DENS+HY_VARINUM-1) = U(MAGX_VAR:MAGZ_VAR,i,j,k)        !magnetic fields
           U0(HY_ENER) = U(DENS_VAR,i,j,k)*U(ENER_VAR,i,j,k)+&                  !total plasma energy 
                .5*dot_product(U(MAGX_VAR:MAGZ_VAR,i,j,k),&
                               U(MAGX_VAR:MAGZ_VAR,i,j,k))
#endif
           call updateConservedVariable(U0(HY_DENS:HY_DENS+HY_VARINUM-1),&
                                        FL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        GL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        HL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        FR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        GR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        HR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1),&
                                        gravX(:,i,j,k),gravY(:,i,j,k),gravZ(:,i,j,k),dx,dy,dz,dt)

           U(DENS_VAR,i,j,k) = max(U0(HY_DENS),hy_smalldens)                    !density
#ifdef FLASH_USM_MHD
           if( hy_updateMagneticFields ) then
              U(MAGX_VAR:MAGZ_VAR,i,j,k) = U0(HY_MAGX:HY_MAGZ)                  !magnetic fields
           endif
#endif
           U(ENER_VAR,i,j,k) = U0(HY_ENER)                                      !total plasma energy
           !! We will update velocity fields after species & mass scalar update

#if (NSPECIES+NMASS_SCALARS) > 0
           !! Note that the velocity fields here are old velocities at time step n, not n+1
           call hy_uhd_updateSpeciesMassScalar&
                (hy_order,U(DENS_VAR,i,j,k),&
                SpOld(1:hy_numXN,i-3:i+3,j-3:j+3,k-3:k+3),&
                Uold(1:6,i-3:i+3,j-3:j+3,k-3:k+3),&
                FL(HY_DENS_FLUX),FR(HY_DENS_FLUX),&
                GL(HY_DENS_FLUX),GR(HY_DENS_FLUX),&
                HL(HY_DENS_FLUX),HR(HY_DENS_FLUX),&
                dx,dy,dz,dt, U(SPECIES_BEGIN:MASS_SCALARS_END,i,j,k))
#endif
           !! Update velocity fields after species & mass scalar update
           U(VELX_VAR:VELZ_VAR,i,j,k) = U0(HY_XMOM:HY_ZMOM)/U0(HY_DENS)        !velocities

        enddo !end of i loop
     enddo !end of j loop
  enddo !end of k loop
  !! ---------------------------------------------------------------
#endif

#ifdef FLASH_USM_MHD
  if (hy_forceHydroLimit) U(MAGX_VAR:MAGZ_VAR,:,:,:) = 0.
#endif

  ! Release block pointers
  call Grid_releaseBlkPtr(blockID,U,CENTER)
  call Grid_releaseBlkPtr(blockID,scrch_Ctr,SCRATCH_CTR)
!!$  if (NDIM > 1) call Grid_releaseBlkPtr(blockID,scrch_Y,SCRATCH_CTR)
!!$  if (NDIM > 2) call Grid_releaseBlkPtr(blockID,scrch_Z,SCRATCH_CTR)

End Subroutine hy_uhd_unsplitUpdate



!! ==================================================================
Subroutine updateConservedVariable(U,xFluxL,yFluxL,zFluxL,&
                                     xFluxR,yFluxR,zFluxR,&
                                     gravX, gravY, gravZ, &
                                     dx,dy,dz,dt)
  use Hydro_data, ONLY : hy_useGravity, hy_useGravPotUpdate, hy_max_flux, hy_flux_dt
  implicit none
  real, dimension(HY_VARINUM), intent(INOUT) :: U
  real, dimension(HY_VARINUM), intent(IN) :: xFluxL,yFluxL,zFluxL,xFluxR,yFluxR,zFluxR
  real, dimension(3),intent(IN) :: gravX,gravY,gravZ
  real, intent(IN) :: dx,dy,dz,dt
  real, dimension(3) :: momentaOld
  real, dimension(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1) :: netFlux
  real :: densOld, dt_temp, magp_sq, magp_flux_sq
  integer :: I

  !! Store old states at n
  densOld = U(HY_DENS)
  momentaOld(1:3) = U(HY_XMOM:HY_ZMOM)

  !! Check net flux to cell
  netFlux = &
       (xFluxR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1)-xFluxL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1))/dx
#if NDIM >= 2
  netflux = netflux + &
       (yFluxR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1)-yFluxL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1))/dy
#if NDIM == 3
  netflux = netflux + &
       (zFluxR(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1)-zFluxL(HY_DENS_FLUX:HY_DENS_FLUX+HY_VARINUM-1))/dz
#endif
#endif

  ! Calculate maximum time step, save for next step
  !do I = 1, HY_VARINUM
  !   dt_temp = abs( hy_max_flux * U(HY_DENS+I-1) / netFlux(I) )
  !   if( dt_temp < hy_flux_dt ) then
  !      !print *, "Reducing hy_flux_dt:", dt_temp, I, netFlux(I), U(HY_DENS+I-1)
  !      hy_flux_dt = dt_temp
  !   endif
  !enddo

  ! Calculate maximum time step based on change in density
  dt_temp = abs( hy_max_flux * U(HY_DENS) / netFlux(HY_DENS_FLUX) )
  if( dt_temp < hy_flux_dt(1) ) then
     !print *, "Reducing hy_flux_dt(DENS):", dt_temp, I, netFlux(HY_DENS), U(HY_DENS)
     hy_flux_dt(1) = dt_temp
  endif

  ! Calculate maximum time step based on change in energy
  dt_temp = abs( hy_max_flux * U(HY_ENER) / netFlux(HY_ENER_FLUX) )
  if( dt_temp < hy_flux_dt(1) ) then
     !print *, "Reducing hy_flux_dt(ENER):", dt_temp, I, netFlux(HY_ENER), U(HY_ENER)
     hy_flux_dt(1) = dt_temp
  endif

  ! Calculate maximum time step based on change in magnetic pressure
  !magp_sq = U(HY_MAGX)**2 + U(HY_MAGY)**2 + U(HY_MAGZ)**2
  !magp_flux_sq = netFlux(HY_MAGX_FLUX)**2 + netFlux(HY_MAGY_FLUX)**2 + netFlux(HY_MAGZ_FLUX)**2
  !if( magp_sq > 0. .and. mag_flux_sq > 0. ) then
  !   dt_temp = abs( hy_max_flux * magp_sq / magp_flux_sq )
  !   if( dt_temp < hy_flux_dt(1) ) then
        !print *, "Reducing hy_flux_dt(ENER):", dt_temp, I, netFlux(HY_ENER), U(HY_ENER)
        hy_flux_dt(1) = dt_temp
  !   endif
  !endif

  !! Update conservative variables from n to n+1 step 
  U(HY_DENS:HY_DENS+HY_VARINUM-1)=U(HY_DENS:HY_DENS+HY_VARINUM-1)-dt*netFlux

  !print *,"updtConsVar", netflux(HY_ENER_FLUX), U(HY_ENER),dt,hy_flux_dt

#ifdef GRAVITY
  if (hy_useGravity) then
     !! Energy should be updated before momenta. This is also found in Athena (see integrate_1d.c)

     !! Note: gravX(1) - gravity at n
     !!       gravX(2) - gravity at n+1/2 (extrapolated)
     !!       gravX(3) - gravity at n+1 (exprapolated)

     !! Update total energy (rho*E) with gravity source
     if (.not. hy_useGravPotUpdate) then
        U(HY_ENER) = U(HY_ENER) + dt*0.5*(dot_product(momentaOld(1:3),   (/gravX(1),gravY(1),gravZ(1)/)) + &
                                          dot_product(U(HY_XMOM:HY_ZMOM),(/gravX(3),gravY(3),gravZ(3)/)))

     !! Include gravity source terms to momentum equations
        U(HY_XMOM:HY_ZMOM) = U(HY_XMOM:HY_ZMOM)+dt*0.5*(    densOld*(/gravX(1),gravY(1),gravZ(1)/) &
                                                        +U(HY_DENS)*(/gravX(3),gravY(3),gravZ(3)/) )

     else
     !! Update total energy (rho*E) with gravity source
     !! Here the grav*(1) is at gravity at n+1/2 from Poisson solve
        U(HY_ENER) = U(HY_ENER) + dt*dot_product(U(HY_XMOM:HY_ZMOM),(/gravX(1),gravY(1),gravZ(1)/))

     !! Include gravity source terms to momentum equations
        U(HY_XMOM:HY_ZMOM) = U(HY_XMOM:HY_ZMOM)+dt*(U(HY_DENS)*(/gravX(1),gravY(1),gravZ(1)/))
     endif

  endif
#endif

End Subroutine updateConservedVariable


!! ==================================================================
Subroutine updateInternalEnergy(eint,pres,FL,FR,GL,GR,HL,HR,dx,dy,dz,dt)

  implicit none

  real, intent(INOUT) :: eint ! rho*eint
  real, dimension(2), intent(IN) :: FL,FR,GL,GR,HL,HR
  real, intent(IN) :: pres,dx,dy,dz,dt

  eint = eint + dt/dx*(FL(1)-FR(1) + pres*(FL(2)-FR(2)))
#if NDIM >= 2
  eint = eint + dt/dy*(GL(1)-GR(1) + pres*(GL(2)-GR(2)))
#if NDIM == 3
  eint = eint + dt/dz*(HL(1)-HR(1) + pres*(HL(2)-HR(2)))
#endif
#endif

End Subroutine updateInternalEnergy
!! ==================================================================
